/**
 *
 */

#include "aqsis_Mesh.h"
#include "aqsis_File.h"

#include <set>
#include <sstream>



/**
 * Constructor.
 *
 * \param mesh  Mesh object that the AQMeshBase encloses.
 */
AQMeshBase::AQMeshBase(Mesh *mesh, const char *name) :
        m_mesh(mesh),
        m_vidx(0),
        m_has_st(false),
        m_has_cs(false),
        m_firstvlr(false),
        m_name(name)
{
        m_subdiv = (
                (m_mesh->flag & ME_SUBSURF) &&
                (m_mesh->subsurftype == ME_CC_SUBSURF) );
}



/**
 * Destructor.
 */
AQMeshBase::~AQMeshBase()
{
}



/**
 * Adds a face to the mesh.
 *
 * \param v     VlakRen object containing the face data.
 */
void AQMeshBase::addData(VlakRen* v)
{
        // check that we have a mesh
        assert(v->ob->type == OB_MESH);

        // push the face material into the vector of materials (one per face)
        //  this is used for meshes that have multiple materials in order to
        //  split them later
        m_materials.push_back(v->mat);

        // if this is the first VlakRen that has been processed, we set the
        //  flags indicating whether or not the mesh has texture coordinates
        //  and color-per-vertex information based upon it.  these flags then
        //  apply to the entire mesh.
        if (!m_firstvlr)
        {
                m_firstvlr = true;
                m_has_st   = (v->tface);
                m_has_cs   = (v->vcol);
        }

        // first, check whether or not this is a triangle or a quadrilateral
        bool is_quad = (v->v4);

        // add each vertex in the VlakRen object.  there are at least 3
        //  vertices (a triangle), but sometimes there are 4 (a quadrilateral)
        postVertex(v->v1);
        postVertex(v->v2);
        postVertex(v->v3);
        if (is_quad)
                postVertex(v->v4);

        // find texture coordinates for the face, and whether or not we have
        //  color-per-vertex information
        TFace* uvc = v->tface;          // s,t texture coordinates
        bool has_st = (uvc);            // do we have s,t texture coordinates?
        bool has_cs = (v->vcol);        // do we have color-per-vertex info?
        // check that the local information available on just this face is the
        //  same as the information we expect to be available for the entire
        //  mesh
        assert(has_st == m_has_st);
        assert(has_cs == m_has_cs);

        // set the number of vertices for this face
        int nverts = (is_quad) ? 4 : 3;
        m_nverts.push_back( nverts );

        // set the vertex indices
        m_iverts.push_back(m_vert_idx.find(v->v1)->second);
        m_iverts.push_back(m_vert_idx.find(v->v2)->second);
        m_iverts.push_back(m_vert_idx.find(v->v3)->second);
        if (is_quad)
                m_iverts.push_back(m_vert_idx.find(v->v4)->second);

        // add normals if the mesh isn't a subdivision surface
        if (!isSubdivMesh())
        {
                if ( v->flag & R_SMOOTH )
                {
                        postNormal(v->v1->n);
                        postNormal(v->v2->n);
                        postNormal(v->v3->n);
                        if (is_quad)
                                postNormal(v->v4->n);
                }
                else
                {
                        for (int i=0; i < nverts; i++)
                                postNormal(v->n);
                }
        }

        // add s,t texture coordinates
        if (has_st)
        {
                m_s.push_back(uvc->uv[0][0]);
                m_t.push_back(1.0f - uvc->uv[0][1]);
                m_s.push_back(uvc->uv[1][0]);
                m_t.push_back(1.0f - uvc->uv[1][1]);
                m_s.push_back(uvc->uv[2][0]);
                m_t.push_back(1.0f - uvc->uv[2][1]);
                if (is_quad)
                {
                        m_s.push_back(uvc->uv[3][0]);
                        m_t.push_back(1.0f - uvc->uv[3][1]);
                }
        }

        // add vertex colors
        if (has_cs)
        {
                postCs(v->vcol[0]);
                postCs(v->vcol[1]);
                postCs(v->vcol[2]);
                if (is_quad)
                        postCs(v->vcol[3]);
        }
}



/**
 * Write the mesh out as a subdivision surface.
 */
void AQMeshBase::writeSubdivisionMesh()
{
        assert(isSubdivMesh());

        RtToken tags[]      = { "interpolateboundary" };
        RtInt   nargs[]     = { 1,1 };
        RtInt   intargs[]   = { 0 };
        RtFloat floatargs[] = { 0 };

        RtToken tokens[4];
        void*   values[4];
        RtInt   count = 1;
        tokens[0] = RI_P;
        values[0] = &m_P[0];

        if (m_has_st)
        {
                tokens[count] = "facevarying float s";
                values[count++] = &m_s[0];
                tokens[count] = "facevarying float t";
                values[count++] = &m_t[0];
        }

        if (m_has_cs)
        {
                tokens[count] = "facevarying color Cs";
                values[count++] = &m_Cs[0];
        }
        
        RiSubdivisionMeshV("catmull-clark", m_nverts.size(),
                &m_nverts[0], &m_iverts[0], 1, tags, nargs,
                intargs, floatargs, count, tokens, values);
        
}



/**
 * Write the mesh out as a PointsPolygon surface.
 */
void AQMeshBase::writePointsPolygons()
{
        assert(!isSubdivMesh());

        RtToken tokens[5];
        void*   values[5];
        RtInt   count = 0;
        
        tokens[count]   = RI_P;
        values[count++] = &m_P[0];

        tokens[count]   = "facevarying normal N";
        values[count++] = &m_N[0];

        if (m_has_st)
        {
                tokens[count]   = "facevarying float s";
                values[count++] = &m_s[0];
                tokens[count]   = "facevarying float t";
                values[count++] = &m_t[0];
        }

        if (m_has_cs)
        {
                tokens[count]   = "facevarying color Cs";
                values[count++] = &m_Cs[0];
        }

        RiPointsPolygonsV(m_nverts.size(), &m_nverts[0], &m_iverts[0],
                count, tokens, values);
}



/**
 * Writes the mesh out.
 *
 * The mesh is written using either RiPointsPolygonsV or RiSubdivisionMeshV
 * depending upon whether it it is a subdivision mesh or an ordinary mesh.
 */
void AQMeshBase::write()
{
        // un-comment the line below to dump out some de-bugging information
        //  about this stage of the mesh just prior to writing it out
        //debugDump();
        
        if (isSubdivMesh())
                writeSubdivisionMesh();
        else
                writePointsPolygons();
}



/**
 * Equality operator.
 *
 * This operator has a "special" use, in that it is used to check whether or
 * not two meshes can be considered the same for the purposes of deformation
 * motion blur.  Don't use it for anything else, without good reason, since
 * it performs additional checks that are appropriate for this use case.
 *
 * \param mb    Other AQMeshBase to check for equality with.
 *
 * \return true if the two AQMeshBase objects are equal, false otherwise.
 */
bool AQMeshBase::operator==(const AQMeshBase& mb) const
{
        // we should *always* have the same number of points
        assert( m_P.size() == mb.m_P.size() );
        assert( m_nverts   == mb.m_nverts   );

        // we should always have consistent texture coordinates and color-
        //  per-vertex information
        assert( m_has_st    == mb.m_has_st    );
        assert( m_has_cs    == mb.m_has_cs    );
        assert( m_Cs.size() == mb.m_Cs.size() );
        assert( m_s.size()  == mb.m_s.size()  );
        assert( m_t.size()  == mb.m_t.size()  );

        // note, for the sake of efficiency, we do not check the shading-
        // related members:
        //      - normals
        //      - s,t texture coordinates
        //      - varying surface color
        return (
                ( m_P      == mb.m_P      ) &&
                ( m_iverts == mb.m_iverts )
        );
}



/**
 * Dumps some debugging information about this AQMeshBase to cout.
 */
void AQMeshBase::debugDump() const
{
        cout << "----------------------- AQMeshBase dump start" << endl;
        cout << "this               = " << this << endl;
        cout << "name               = " << m_name << endl;
        cout << "m_has_st           = " << m_has_st << endl;
        cout << "m_has_cs           = " << m_has_cs << endl;
        cout << "m_subdiv           = " << m_subdiv << endl;
        cout << "m_P.size()         = " << m_P.size() << endl;
        cout << "m_N.size()         = " << m_N.size() << endl;
        cout << "m_s.size()         = " << m_s.size() << endl;
        cout << "m_t.size()         = " << m_t.size() << endl;
        cout << "m_Cs.size()        = " << m_Cs.size() << endl;
        cout << "m_nverts.size()    = " << m_nverts.size() << endl;
        cout << "m_iverts.size()    = " << m_iverts.size() << endl;
        cout << "m_materials.size() = " << m_materials.size() << endl;
        cout << "----------------------- AQMeshBase dump end" << endl;
}



/**
 * Splits the mesh out into a sequence of sub-meshes, each of which has only
 * a single material.
 *
 * In RenderMan, only one surface shader may be applied to any given surface.
 * For meshes that have multiple materials, we must therefore split them
 * out into sub-meshes which only have one material each.  The current mesh
 * becomes obsolete if there is more than one material (ie: it may be
 * deleted).
 *
 * If a mesh is split, it becomes defunct (it can be deleted), and all new
 * meshes are placed into the output vector.  The new meshes in the output
 * vector only have a *single* element in their materials vector, which is
 * the material for the entire object.
 *
 * \param submeshes     Vector to which the split meshes should be added.
 */
void AQMeshBase::materialSplit(vector<AQMeshBase*>& submeshes)
{
        assert(m_materials.size() > 0);
        assert(m_materials.size() == m_nverts.size());

        // first, we count the number of unique materials, and assign a
        //  (possibly new) AQMeshBase to each material
        map<Material*, AQMeshBase*> matmap;
	unsigned int i;
        for (i=0; i < m_materials.size(); i++)
        {
                // add the material, along with a new mesh base, if it's not
                //  present already
                if (matmap.find(m_materials[i]) == matmap.end())
                {
                        stringstream strName;
                        strName << &m_name[0] << ".submateral"
                                << matmap.size() << ends;
                
                        matmap[m_materials[i]] = new AQMeshBase(m_mesh, 
                                &strName.str()[0]);
                        assert(matmap.find(m_materials[i]) != matmap.end());
                        AQMeshBase* newMesh = matmap[m_materials[i]];
                        assert(newMesh);
                        newMesh->m_has_st = m_has_st;
                        newMesh->m_has_cs = m_has_cs;
                        newMesh->m_subdiv = m_subdiv;
                        newMesh->m_firstvlr = false;
                        // set the current material as the only element of the
                        //  materials vector
                        newMesh->m_materials.push_back(m_materials[i]);
                }
        }
        // bail quickly if the number of unique materials is only 1
        if (matmap.size() == 1)
        {
                // delete the allocated AQMeshBase
                delete matmap[m_materials[0]];
                matmap.clear();
                // clear the list of materials to save a little memory
                m_materials.clear();
        
                return;
        }

        // process each face of the mesh
        unsigned int vidx = 0;          // vertex index
        unsigned int Nptr = 0;          // normals index
        unsigned int Csptr = 0;         // color-per-vertex index
        for (i=0; i < m_nverts.size(); i++)
        {
                // find the appropriate sub-mesh for the material of the
                //  current face
                assert(matmap.find(m_materials[i]) != matmap.end());
                AQMeshBase* newMesh = matmap[m_materials[i]];
                assert(newMesh);

                // process each vertex of the current face
                for (int j=0; j < m_nverts[i]; j++)
                {
                        // post the current vertex for the new mesh
                        int Pidx = 3*m_iverts[vidx];
                        float x = m_P[Pidx], y = m_P[Pidx+1], z = m_P[Pidx+2];
                        newMesh->postVertex(vidx, x, y, z);
                        int ivert = newMesh->m_idmap.find(vidx)->second;
                        newMesh->m_iverts.push_back(ivert);

                        // copy normal at the current vertex for the new mesh
                        assert(isSubdivMesh() == newMesh->isSubdivMesh());
                        if (!isSubdivMesh())
                        {
                                newMesh->m_N.push_back(m_N[Nptr++]);
                                newMesh->m_N.push_back(m_N[Nptr++]);
                                newMesh->m_N.push_back(m_N[Nptr++]);
                        }

                        // copy texture coordinates at the current vertex
                        assert(m_has_st == newMesh->m_has_st);
                        if (m_has_st)
                        {
                                newMesh->m_s.push_back(m_s[vidx]);
                                newMesh->m_t.push_back(m_t[vidx]);
                        }

                        // copy vertex color at the current vertex
                        assert(m_has_cs == newMesh->m_has_cs);
                        if (m_has_cs)
                        {
                                newMesh->m_Cs.push_back(m_Cs[Csptr++]);
                                newMesh->m_Cs.push_back(m_Cs[Csptr++]);
                                newMesh->m_Cs.push_back(m_Cs[Csptr++]);
                        }

                        // increment the vertex pointer
                        ++vidx;
                }

                // push the current face onto the new mesh
                newMesh->m_nverts.push_back(m_nverts[i]);

        }

        // push all of the new sub-meshes onto the output vector
        map<Material*, AQMeshBase*>::iterator it;
        for (it = matmap.begin(); it != matmap.end(); ++it)
        {
                submeshes.push_back(it->second);
                //it->second->debugDump();
        }
        
}



/**
 * AQMesh constructor.
 *
 * \param obj   Blender Object from which to construct this AQMesh.
 */
AQMesh::AQMesh(Object *obj) :
        AQObject(obj),
        m_hasSubFrames(false)
{
        assert(obj);
        assert(obj->type == OB_MESH);

        m_mesh = reinterpret_cast<Mesh*>(obj->data);
        m_subFrames.push_back(new AQMeshBase(m_mesh, obj->id.name));
}



/**
 * AQMesh destructor.
 */
AQMesh::~AQMesh()
{
        if (hasSubFrames())
        {
                // compose a list of unique AQMeshBase objects from 
                //  m_subFrames that need to be deleted.  note that this is
                //  required since m_subFrames may contain duplicate pointers
                set<AQMeshBase*> uniques;
                for (unsigned int i=0; i < m_subFrames.size(); i++)
                        uniques.insert(m_subFrames[i]);
                // delete the unique AQMeshBase objects
                set<AQMeshBase*>::iterator _i;
                for (_i=uniques.begin(); _i != uniques.end(); ++_i)
                        delete *_i;
        }
        else
        {
                delete m_subFrames[0];
        }
}



/**
 * Add data to the mesh.
 *
 * \param v     Data chunk.
 */
void AQMesh::addData(VlakRen* v)
{
        AQObject::addData(v);

        assert(v);
        assert(m_subFrames.back());
        m_subFrames.back()->addData(v);
}



/**
 * Increments the current sub-frame of the animation.
 */
void AQMesh::incrementSubFrame()
{
        AQObject::incrementSubFrame();

        // here, we compare the sub-frame that has just finished with the
        //  one before it (if there is one).  if the two sub-frame meshes
        //  are identical, then the most recent one is deleted and replaced by
        //  a pointer to the previous one.  this means that non-deforming
        //  meshes will end up with only a single AQMeshBase pointer in
        //  their entire m_subFrames array, which is more memory-efficient.
        unsigned int nSubFrames = getNSubFrames();
        assert(nSubFrames == m_subFrames.size());
        if ((!m_hasSubFrames) && (nSubFrames > 1))
        {
                // check if the sub-frame we've just finished is equal to the
                //  one before it
                if (*m_subFrames[nSubFrames-1] == *m_subFrames[nSubFrames-2])
                {
                        // if the most recent sub-frame is a repeat of the
                        //  one before it (ie: it hasn't deformed since then),
                        //  then delete the most recent sub-frame and
                        //  replace it with a pointer to the one before it.
                        delete m_subFrames[nSubFrames-1];
                        m_subFrames[nSubFrames-1] = m_subFrames[nSubFrames-2];
                } else {
                        // here, the sub-frame is NOT equal to the one
                        //  before it, and we have a case of deformation
                        //  motion blur
                        m_hasSubFrames = true;
                }
        }

        m_subFrames.push_back(new AQMeshBase(m_mesh, 
                getBlenderObject()->id.name)); 
}



/**
 * Splits a mesh with multiple materials into sub-meshes with only one
 * material each.
 *
 * RenderMan can only apply a single surface shader to an object.  Hence, if a
 * mesh contains multiple materials then it must be split into a set of meshes
 * that only have one material each.  This method is designed to achieve the
 * required split.
 *
 * If the current AQMesh has only one material, then no new elements are added
 * to the vector.  Otherwise, new meshes are added to the submeshes vector,
 * and the current AQMesh becomes defunct (ie: it may be deleted, since all of
 * the information is now in the submeshes vector).
 *
 * \param submeshes     Vector to contain new meshes.
 */
void AQMesh::materialSplit(vector<AQMesh*>& submeshes)
{
        bool hasSplit = false;

        // find the unique sub-frames of the mesh.  don't check the last
        //  sub-frame since it's always empty as a result of the final
        //  incrementSubFrame() call
        map<AQMeshBase*, vector<AQMeshBase*> > uniques;
	unsigned int i;
        for (i=0; i < (m_subFrames.size()-1); i++)
        {
                if (uniques.find(m_subFrames[i]) == uniques.end())
                        uniques[m_subFrames[i]].clear();
        }

        // split out each unique sub-frame of the mesh, placing the resulting
        //  sub-frames into the second part of uniques
        unsigned int nSubMeshes = 0;
        map<AQMeshBase*, vector<AQMeshBase*> >::iterator it;
        for (it = uniques.begin(); it != uniques.end(); ++it)
        {
                AQMeshBase* curMeshBase = it->first;
                assert(curMeshBase);
        
                // try to split the current sub-frame
                curMeshBase->materialSplit(it->second);
                if (it->second.empty())
                {
                        // if the first split wasn't necessary, then bail
                        //  quickly at this point
                        assert(hasSplit == false);
                        return;
                }
                else
                {
                        hasSplit = true;
                        if (nSubMeshes == 0)
                        {
                                nSubMeshes = it->second.size();
                        }
                        else
                        {
                                assert(it->second.size() == nSubMeshes);
                        }
                }
        }

        // create a new set of AQMesh objects
        for (i = 0; i < nSubMeshes; i++)
        {
                AQMesh* newMesh = new AQMesh(getBlenderObject());
                newMesh->m_hasSubFrames = m_hasSubFrames;
                newMesh->m_subFrames.clear();
                newMesh->m_subframe = m_subFrames.size()-1;
                newMesh->m_material = 
                        uniques[m_subFrames[0]][i]->getFirstMaterial();

                for (unsigned int j=0; j < (m_subFrames.size()-1); j++)
                {
                        AQMeshBase* curMeshBase = uniques[m_subFrames[j]][i];
                        newMesh->m_subFrames.push_back(curMeshBase);
                }
                // add an empty AQMeshBase object on the end to simulate the
                //  standard case
                newMesh->m_subFrames.push_back(new AQMeshBase(m_mesh,
                        getName()));

                // add this new AQMesh to the output vector
                submeshes.push_back(newMesh);
        }
}



/**
 * Exports the mesh.
 *
 * \mblur       Flag to indicate that motion blur is being used.
 */
void AQMesh::internWriteObject(bool mblur)
{
        AQObject::internWriteObject(mblur);

        assert(m_subFrames.size() > 0);
        assert(static_cast<int>(m_subFrames.size()) >= getNSubFrames());

        // for multiple sub-frames and motion blur, we have to export multiple
        //  sub-frames within a motion block.  for a single sub-frame, and no
        //  motion blur, we just export the first sub-frame.
        if ((mblur) && (hasSubFrames()))
        {
                // if we're using motion blur and have multiple sub-frames
                //  that are different from each other, then we need to
                //  export all appropriate sub-frames
                RtInt nSubFrames = getNSubFrames();
                vector<RtFloat> times;
		int i;
                for (i=0; i < nSubFrames; i++)
                {
                        RtFloat time = (RtFloat)i / (RtFloat)(nSubFrames-1);
                        times.push_back(time);
                }
                RiMotionBeginV(nSubFrames, &times[0]);
                        for (i=0; i < nSubFrames; i++)
                                m_subFrames[i]->write();
                RiMotionEnd();
        }
        else
        {
                // if we're not using motion blur, or only have a single
                //  sub-frame, then only the first sub-frame should be
                //  exported
                m_subFrames[0]->write();
        }
}



/**
 * AQMMMesh destructor.
 */
AQMMMesh::~AQMMMesh()
{
        // delete all sub-meshes
        for (unsigned int i=0; i < m_subMeshes.size(); i++)
        {
                delete m_subMeshes[i];
        }
        m_subMeshes.clear();
}



/**
 * Exports the multiple-material mesh.
 */
void AQMMMesh::writeObject(const string& ribpath, bool mblur)
{
        // store the original RIB context handle
        RtContextHandle oldRI = RiGetContext();

        // start a new RIB file for the geometry of this object
        RiBegin(&getFileName(ribpath,false)[0]);

        // comment this file to let the user know exactly what it is
        RiArchiveRecord(RI_COMMENT, AQ_EXPORTER_NOTIFY, RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                " The RenderMan (R) Interface Procedures and Protocol are:",
                RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "       Copyright 1988, 1989, Pixar", RI_NULL);
        RiArchiveRecord(RI_COMMENT,
                "        All Rights Reserved", RI_NULL);
        stringstream strComment;
        strComment << " Object: " << getName() << ends;
        RiArchiveRecord(RI_COMMENT, &strComment.str()[0], RI_NULL);
        RiArchiveRecord(RI_COMMENT, " This is a mesh with multiple materials.",
                RI_NULL);

        // export the real sub-meshes
        assert(m_subMeshes.size() > 1);
        for (unsigned int i=0; i < m_subMeshes.size(); i++)
        {
                assert(m_subMeshes[i]);
                AQMesh *mesh = m_subMeshes[i];
                if (R.r.AQ_shadebug == AQ_SHADEBUG_OFF)
                {
                        Material *fmat = mesh->getMaterial();
                        std::string matname = fmat->id.name;
                        RiReadArchive(&aqsisFileRender_t::getMaterialRIBName(
                                fmat, ribpath, R.r.AQ_relpath)[0], 0, RI_NULL);
                }
                mesh->internWriteObject(mblur);
        }

        // close the current rib file
        RiEnd();

        // reset the original RIB context handle
        RiContext(oldRI);

}


