/**
 *
 */
#ifndef __AQSIS_MESH_H
#define __AQSIS_MESH_H
#ifdef __cplusplus
extern "C" {
#include "DNA_mesh_types.h"
}
#include <assert.h>
#include <iostream>
#include <map>
#include <set>
#include "ri.h"
#include "aqsis_Object.h"

using namespace std;



/**
 * Base mesh object, that represents a snapshot of a mesh at a single 
 * sub-frame.
 */
class AQMeshBase
{
        //------------------------------------------------------ Public Methods
        public:
                AQMeshBase(Mesh *mesh, const char *name);
                ~AQMeshBase();
                void addData(VlakRen* v);
                void write();
                bool operator==(const AQMeshBase& mb) const;
                void debugDump() const;
                void materialSplit(vector<AQMeshBase*>& submeshes);
                Material* getFirstMaterial() const
                {
                        assert(m_materials.size() > 0);
                        return m_materials[0];
                }
        //--------------------------------------------------- Protected Methods
        protected:
                /**
                 * Posts a vertex to the list if the vertex hasn't been added
                 * already.  This version is used when creating the original
                 * mesh from VlakRen data.
                 */
                void postVertex(VertRen* ver)
                {
                        assert(ver);
                        assert(m_idmap.size() == 0);  // don't mix with other
                        if (m_vert_idx.find(ver) == m_vert_idx.end())
                        {
                                m_vert_idx[ver] = m_vidx++;
                                m_P.push_back(ver->co[0]);
                                m_P.push_back(ver->co[1]);
                                m_P.push_back(ver->co[2]);
                        }
                }
                /**
                 * Posts a vertex to the list if the vertex hasn't been added
                 * already.  This version is used when splitting multiple-
                 * material meshes.
                 */
                void postVertex(int originalIndex, float x, float y, float z)
                {
                        assert(m_vert_idx.size() == 0); // don't mix with other
                        if (m_idmap.find(originalIndex) == m_idmap.end())
                        {
                                m_idmap[originalIndex] = m_vidx++;
                                m_P.push_back(x);
                                m_P.push_back(y);
                                m_P.push_back(z);
                        }
                }
                /**
                 * Posts a normal vector to their list.
                 */
                void postNormal(float n[3])
                {
                        assert(n);
                        m_N.push_back(n[0]); 
                        m_N.push_back(n[1]);
                        m_N.push_back(n[2]);
                }
                /**
                 * Posts a vertex color to their list.
                 */
                void postCs(unsigned int vcol)
                {
                        m_Cs.push_back( ((vcol >> 24) & 255) / 255.0 );
                        m_Cs.push_back( ((vcol >> 16) & 255) / 255.0 );
                        m_Cs.push_back( ((vcol >>  8) & 255) / 255.0 );
                }
                /**
                 * Returns true if the mesh is a subdivision surface.
                 */
                bool isSubdivMesh() const
                {
                        return m_subdiv;
                }
                void writeSubdivisionMesh();
                void writePointsPolygons();
        //----------------------------------------------------- Private Members
        private:
                Mesh*           m_mesh;         ///< the original mesh
                vector<RtFloat> m_P;            ///< vertices
                vector<RtFloat> m_N;            ///< normals
                vector<RtFloat> m_s;            ///< s texture coordinates
                vector<RtFloat> m_t;            ///< t texture coordinates
                vector<RtFloat> m_Cs;           ///< varying surface color
                vector<RtInt>   m_nverts;       ///< n vertices per polygon
                vector<RtInt>   m_iverts;       ///< vertex indices
                map<VertRen*, int> m_vert_idx;  ///< for removing duplicates
                map<int, int>   m_idmap;        ///< for removing duplicates
                vector<Material*> m_materials;  ///< one material per face
                int             m_vidx;         ///< vertex index counter
                bool            m_has_st;       ///< we have s,t coordinates
                bool            m_has_cs;       ///< we have color-per-vertex
                bool            m_firstvlr;     ///< first vlr processed?
                bool            m_subdiv;       ///< subdiv mesh flag
                string          m_name;         ///< name of the mesh object
};



/**
 * Aqsis mesh object.
 */
class AQMesh : public AQObject
{
        // exception: let AQMMMesh access AQMesh's protected method
        //  internWriteObject
        friend class AQMMMesh;
        //------------------------------------------------------ Public Methods
        public:
                // constructor and destructor
                AQMesh(Object *obj);
                virtual ~AQMesh();
                // addition of export data
                virtual void addData(VlakRen* v);
                virtual void incrementSubFrame();
                // split a mesh with multiple materials
                void materialSplit(vector<AQMesh*>& submeshes);
        //--------------------------------------------------- Protected Methods
        protected:
                // export functionality
                virtual void internWriteObject(bool mblur);
                // does this mesh have multiple sub-frames?
                bool hasSubFrames() { return m_hasSubFrames; }
        //----------------------------------------------------- Private Members
        private:
                Mesh* m_mesh;           ///< mesh
                bool m_hasSubFrames;    ///< the mesh has multiple sub-frames?
                vector<AQMeshBase*> m_subFrames; ///< mesh sub-frames
};



/**
 * Mesh object with multiple materials.
 */
class AQMMMesh : public AQObject
{
        //------------------------------------------------------ Public Methods
        public:
                // constructor and destructor
                AQMMMesh(const vector<AQMesh*>& submeshes, Material* mat) :
                        AQObject(submeshes[0]->getBlenderObject()),
                        m_subMeshes(submeshes)
                { 
                        m_material = mat;
                }
                virtual ~AQMMMesh();
                virtual void writeObject(const string& ribpath, bool mblur);
        //----------------------------------------------------- Private Members
        private:
                // list of AQMesh objects
                vector<AQMesh*> m_subMeshes;
};



#endif // __cplusplus
#endif // __AQSIS_MESH_H

