#ifdef WIN32 
#pragma warning(disable : 4786)
#endif

#include <assert.h>
#include "aqsis_Render.h"

#include <math.h>
#include <set>

using namespace std;



/**
 * Indicates to the exporter that the current (sub)frame is the last one
 * being exported.
 *
 * To permit motion blur, the exporter must have access to transforms and
 * object data from multiple frames.  This method marks that the last sub-
 * frame is being exported, for any tasks that require data to be grabbed
 * only once.
 *
 * @param last	true if the current frame is the last frame of export, false
 *			otherwise.
 */
void aqsisRender_t::lastSubFrame(bool last)
{
	// set the boolean flag member variable
	lastPass = last;
}



/**
 * Export a sub-frame of the animation.
 *
 * Multiple sub-frames make up a single rendering frame.  Multiple sub-frames
 * are used so that the exporter is able to acquire motion blur information
 * from Blender.
 */
bool aqsisRender_t::exportSubFrame() 
{

	// fetch camera details if this is the last rendering pass
	if (lastPass)
	{
                maincam_obj = G.scene->camera;
                assert(maincam_obj);
		mainCamLens = 35.0;
		if (maincam_obj->type == OB_CAMERA)
                {
			mainCamLens = ((Camera*)maincam_obj->data)->lens;

                        // if the camera transformation has not yet been
                        //  recorded then record it.  this can happen if, for
                        //  example, the camera is in a layer that is not
                        //  currently active
                        if (objTransforms.find(maincam_obj) ==
                            objTransforms.end())
                        {
                                addTransformationMatrix(maincam_obj);
                        }
                }
	}

        // fetch all object data
	if (!getAllMatTexObs())
	{
		// error found, clear for next call
		clearAll();
		return false;
	}

        // tell all the exporter objects that the current sub-frame is
        //  being incremented
        std::map<Object*, AQObject*>::iterator i;
        for (i = m_objects.begin(); i != m_objects.end(); ++i)
        {
                AQObject* aqobj = i->second;
                assert(aqobj);
                aqobj->incrementSubFrame();
        } 

        // clear the object data for the current frame
        objectData.clear();

	// all finished OK - ready for the next sub-frame
	return true;
}



/**
 * Commences actual rendering export.
 *
 * This method invokes the renderer.
 * This method should be called once all sub-frame data has been acquired.
 */
bool aqsisRender_t::doRender()
{
        // perform a sanity-check of the export data
        checkData();

        // special case: split meshes that have multiple materials into
        //  individual sub-meshes
        std::map<Object*, AQObject*>::iterator it;
        std::map<Object*, AQObject*> newMeshes;
        for (it = m_objects.begin(); it != m_objects.end(); ++it)
        {
                Object*   obj   = it->first;
                AQObject* aqobj = it->second;
                assert(obj);
                assert(aqobj);
                
                if (obj->type == OB_MESH)
                {
                        std::vector<AQMesh*> submeshes;
                        AQMesh* aqmesh = dynamic_cast<AQMesh*>(aqobj);
                        aqmesh->materialSplit(submeshes);
                        if (!submeshes.empty())
                        {
                                newMeshes[obj] = new AQMMMesh(submeshes,
                                        aqmesh->getMaterial());
                                assert(newMeshes[obj]);
                                submeshes.clear();
                        }
                }
        }
        for (it = newMeshes.begin(); it != newMeshes.end(); ++it)
        {
                Object*   obj      = it->first;
                AQObject* aqobj    = it->second;
                assert(obj);
                assert(aqobj);

                delete m_objects[obj];
                m_objects.erase(m_objects.find(obj));
                m_objects[obj] = aqobj;
        }
        newMeshes.clear();

	// initialize export
	if (!initExport())
	{
		// error found, clear for next call
		clearAll();
		return false;
	}

	// start actual data export
        if ( R.r.AQ_shadebug == AQ_SHADEBUG_OFF )
	        writeMaterialsAndModulators();

	// write out shadow map frames
	if ( (R.r.mode & R_SHADOW) && (R.r.AQ_shadebug == AQ_SHADEBUG_OFF) )
	{
		for (int i=0; i < R.totlamp; i++)
		{
			LampRen* lamp = R.la[i];
                        assert(lamp);
	
			if ((lamp->type == LA_SPOT) && (lamp->mode & LA_SHAD))
				writeShadowMap(lamp);
		}
	}

	// write out the main frame
	writeFrame();

	// clear for the next call, before render to free some memory
	clearAll();

        // finishExport() finishes the export procedure and invokes the
        //  renderer
	if(!finishExport())
	{
		G.afbreek = 1;	//stop render and anim if doing so
		return false;
	}
	else 
	{
		return true;
	}
}



/**
 * Gets all object data.
 */
bool aqsisRender_t::getAllMatTexObs()
{
	VlakRen* vlr;

	for (int i=0; i < R.totvlak; i++)
	{
		// fetch the current VlakRen (render face) object
		if ((i & 255) == 0)
			vlr = R.blovl[ i >> 8 ];
		else
			vlr++;

		//------------------------------------- OBJECT RENDER FACE DATA
                Object *obj = vlr->ob;
                // add the object to the internal map of objects if it isn't
                //  yet present
                AQObject* aqobj = NULL;
                if (m_objects.find(obj) == m_objects.end())
                {
                        if (obj->type == OB_MESH) {
                                m_objects[obj] = new AQMesh(obj);
                                aqobj = m_objects[obj];
                                assert(aqobj);

                                //std::cout << "Added new AQObject for: "
                                //          << obj->id.name << std::endl;
                        }
                }
                else
                {
                        if (obj->type == OB_MESH) {
                                aqobj = m_objects[obj];
                                assert(aqobj);
                        }
                }
                // add rendering face data to the specified object
                if (aqobj)
                        aqobj->addData(vlr);

		//-------------------------------------- MATERIALS AND TEXTURES
		Material *matr = vlr->mat;
		if (matr)
		{
			// if the default, nameless material is applied, call
			//  it "blender_default"
			if (strlen(matr->id.name) == 0)
				used_materials["blender_default"] = matr;
			else
				used_materials[matr->id.name] = matr;

			// textures, all active channels
			for (int m=0; m < 8; m++)
			{
				// skip inactive channels
				if (matr->septex & (1<<m))
					continue;
				MTex* mx = matr->mtex[m];
				if (mx == NULL)
					continue;
				Tex* tx = mx->tex;
				if (tx == NULL)
					continue;
				short txtp = tx->type;
				// skip any texture types that we don't know
				//  about
				if (
					(txtp != TEX_STUCCI) &&
					(txtp != TEX_CLOUDS) &&
					(txtp != TEX_WOOD)   &&
					(txtp != TEX_MARBLE) &&
					(txtp != TEX_IMAGE)
				) continue;
				// in the case of an image texture, check that
				//  there is an actual image; otherwise ignore
				if ((txtp & TEX_IMAGE) && (!tx->ima))
					continue;
					
				// add the name of the texture to the list of
				//  used textures
				used_textures[tx->id.name] = 
					make_pair(matr, mx);
			}
		}

	}

	return true;

}



/**
 * Pushes another transformation matrix onto the stack for a given object.
 */
void aqsisRender_t::addTransformationMatrix(Object *obj)
{
        objTransforms[obj].push(AQTransform(obj));
}



/**
 * Explicitly adds a duplication matrix to the stack for a given object.
 */
void aqsisRender_t::addDupliMtx(Object* obj)
{
	addTransformationMatrix(obj);
}



/**
 * Checks whether or not data from a specific object has already been
 * recorded; if not, then record it.
 *
 * This method allows for the detection of objects that share the same object
 * data.  If two objects share the same object data, then only one of them
 * need be recorded, and the other one can be represented as a linked object
 * with its own transformation.
 */
bool aqsisRender_t::objectKnownData(Object* obj)
{ 
        assert(obj);

        // only process known object types:
        //      - cameras
        //      - meshes that aren't particle systems
        // note that lights aren't included in this - they get their transform
        // data separately
        bool canProcess = false;
        // flag true processing
        if ((obj->type == OB_MESH) || (obj->type == OB_CAMERA))
                canProcess = true;
        // check any false cases
        if (isParticleSystem(obj))
                canProcess = false;

        // bail if we can't process this object
        if (!canProcess)
                return false;

	if (objectData.find(obj->data) != objectData.end())
	{
		// object is known, so find the original object
		Object *orgob = objectData[obj->data];
                assert(orgob);

		// then save matrix of linked object in the list, and set up a
		//  link to the original object
		addTransformationMatrix(obj);
                if (lastPass)
                {
                        dup_srcob[obj] = orgob;
                }

                //std::cout << "Added dupli matrix for known object: "
                //          << obj->id.name << std::endl;

		return true;
	}
	else
	{
		// object not known yet, so set it as the main object
		addTransformationMatrix(obj);
		objectData[obj->data] = obj;

                //std::cout << "Added original matrix for unknown object: "
                //          << obj->id.name << std::endl;

		return false;
	}

}



/**
 * Clears all exporter data.  This method should return the exporter to a
 * fresh state.
 */
void aqsisRender_t::clearAll()
{
	used_materials.clear();
	used_textures.clear();
	dup_srcob.clear();
	objectData.clear();
	layerLights.clear();
	shadowDict.clear();
	objTransforms.clear();

        // delete all AQObjects from the scene
        std::map<Object*, AQObject*>::iterator i;
        for (i = m_objects.begin(); i != m_objects.end(); ++i)
        {
                delete i->second;
        }
        m_objects.clear();
	
	lastPass = false;
}



/**
 * Returns true if obj is a particle system mesh.
 *
 * Particle system meshes should not be rendered as normal meshes.
 *
 * \param obj   Object to check.
 *
 * \return true if obj is a particle system and false otherwise.
 */
bool aqsisRender_t::isParticleSystem(Object* obj)
{
        assert(obj);
        return ((obj->type & OB_MESH) && (give_parteff(obj)));
}



/**
 * Checks the rendering data and reports on its current validity.
 *
 * This method can be used to check the data available to the exporter is
 * complete.  The exporter receives data (potentially) from several frames,
 * and different types of data are stored in different places.  This diversity
 * of required information has a high potential for bugs to be introduced,
 * and this method is here to provide a consistency check and to aid in
 * debugging.
 *
 * \returns true if all data is consistent, and false otherwise.
 */
bool aqsisRender_t::checkData()
{
        bool failed = false;

        cout << endl << "aqsisRender_t::checkData() start" << endl;

        // check that there is a camera object.  the abscence of a camera
        //  isn't enough to make the scene invalid, but it should be flagged
        if (!maincam_obj)
        {
                cout << "WARNING: maincam_obj is NULL" << endl;
        }

        // sanity check that mainCamLens > 0.0
        if (mainCamLens <= 0)
        {
                cout << "WARNING: mainCamLens <= 0.0" << endl;
        }

        // check that the camera has transformation data
        if (objTransforms.find(maincam_obj) == objTransforms.end())
        {
                cout << "ERROR: Main camera has no transform" << endl;
                failed = true;
        }

        // iterate through m_objects, the map of Blender objects to AQObjects
        //  and make sure that:
        //   - it has no null entries
        //   - it has no duplicate entries
        //   - all objects contained have appropriate transform data available
        set<AQObject*> aqObjSet;
        map<Object*, AQObject*>::iterator i1;
        for (i1 = m_objects.begin(); i1 != m_objects.end(); ++i1)
        {
                Object*   obj   = i1->first;
                AQObject* aqobj = i1->second;

                if (obj == NULL)
                {
                        cout << "ERROR: NULL Object* in m_objects" << endl;
                        failed = true;
                }

                if (aqobj == NULL)
                {
                        cout << "ERROR: NULL AQObject* in m_objects" << endl;
                        failed = true;
                }
                else
                {
                        // check that the AQObject hasn't already been
                        //  encountered
                        if (aqObjSet.find(aqobj) != aqObjSet.end())
                        {
                                cout << "ERROR: Duplicate AQObject* in "
                                     << "m_objects. Name = " 
                                     << obj->id.name << endl;
                                failed = true;
                        }
                        else
                        {
                                aqObjSet.insert(aqobj);
                        }
                }

                map<Object*, AQTransformStack>::iterator t;
                t = objTransforms.find(obj);
                if (t == objTransforms.end())
                {
                        cout << "ERROR: No transform data available for "
                             << "Object* with name = " << obj->id.name << endl;
                        failed = true;
                }

                // print the object
                if (obj)
                {
                        cout << "OBJECT(" << objStringType(obj) << ",";
                        if (t != objTransforms.end())
                                cout << t->second.size();
                        else
                                cout << "N";
                        cout << "):" << obj->id.name << endl;
                }
                else
                {
                        cout << "OBJECT: NULL!" << endl;
                }
        }

        // check the list of linked objects
        //      - linkOb isn't in the list of unique objects
        //      - srcOb is in the list of unique objects
        //      - linkOb has transform data available
        map<Object*, Object*>::iterator i2;
        for (i2 = dup_srcob.begin(); i2 != dup_srcob.end(); ++i2)
        {
                Object* linkOb = i2->first;
                Object* srcOb  = i2->second;

                // make sure that linkOb is *not* in m_objects
                if (m_objects.find(linkOb) != m_objects.end())
                {
                        cout << "ERROR: Object* with name = "
                             << linkOb->id.name << " found in both lists of "
                             << "duplicated objects and unique objects"
                             << endl;
                        failed = true;
                }

                // make sure that srcOb *is* in m_objects
                if (m_objects.find(srcOb) == m_objects.end())
                {
                        cout << "ERROR: Link source with name = "
                             << srcOb->id.name << " not in "
                             << "original objects" << endl;
                        cout << "       This object is the source for object "
                             << "with name = " << linkOb->id.name << endl;
                        failed = true;
                }

                // check that transformation data is available for linkOb
                map<Object*, AQTransformStack>::iterator tlink;
                map<Object*, AQTransformStack>::iterator slink;
                tlink = objTransforms.find(linkOb);
                slink = objTransforms.find(srcOb);
                if (tlink == objTransforms.end())
                {
                        cout << "ERROR: Link object with name = "
                             << linkOb->id.name << " has no transform data"
                             << endl;
                        failed = true;
                }

                // print the link
                if (linkOb && srcOb)
                {
                        cout << "LINK(" << objStringType(linkOb) << ",";
                        if (tlink != objTransforms.end())
                                cout << tlink->second.size();
                        else
                                cout << "N";
                        cout << "):" << linkOb->id.name << " -> "
                             << "OBJECT(" << objStringType(srcOb) << ",";
                        if (slink != objTransforms.end())
                                cout << slink->second.size();
                        else
                                cout << "N";
                        cout << "):" << srcOb->id.name << endl;
                }
                else
                {
                        cout << "LINK: One member is NULL!" << endl;
                }
        }

        cout << "aqsisRender_t::checkData() end. failed = " << failed << endl;

        return failed;
}



/**
 * Returns the type of the specified object as a string.
 *
 * Can be any of:
 *   OB_EMPTY, OB_MESH, OB_CURVE, OB_SURF, OB_FONT, OB_MBALL, OB_LAMP,
 *   OB_CAMERA, OB_IKA, OB_WAVE, OB_LATTICE, OB_ARMATURE, or unknown.
 */
string aqsisRender_t::objStringType(const Object *obj)
{
        assert(obj);
        short type = obj->type;

        if (type == OB_EMPTY)
                return string("OB_EMPTY");
        else if (type == OB_MESH)
                return string("OB_MESH");
        else if (type == OB_CURVE)
                return string("OB_CURVE");
        else if (type == OB_SURF)
                return string("OB_SURF");
        else if (type == OB_FONT)
                return string("OB_FONT");
        else if (type == OB_MBALL)
                return string("OB_MBALL");
        else if (type == OB_LAMP)
                return string("OB_LAMP");
        else if (type == OB_CAMERA)
                return string("OB_CAMERA");
        else if (type == OB_IKA)
                return string("OB_IKA");
        else if (type == OB_WAVE)
                return string("OB_WAVE");
        else if (type == OB_LATTICE)
                return string("OB_LATTICE");
        else if (type == OB_ARMATURE)
                return string("OB_ARMATURE");
        else
                return string("unknown");
}

