/**
 * Class definition for AQTransform
 */

#include <assert.h>
#include <string.h>
#include <iostream>

#include "aqsis_Transform.h"

#include "BLI_arithb.h"         // basic arithmetic on matrices
#include "MTC_matrixops.h"      // more matrix operations
#include "ri.h"                 // RenderMan interface

using namespace std;



/**
 * AQTransform default constructor.
 *
 * This constructs an identity transform.
 */
AQTransform::AQTransform()
{
        // zero all entries
        memset(m_matrix, 0, 16*sizeof(float));
        // set diagonal entries to 1
        m_matrix[0][0] = m_matrix[1][1] = m_matrix[2][2] = m_matrix[3][3] = 1;
}



/**
 * AQTransform constructor.
 *
 * AQTransforms are constructed from a Blender Object.  The object
 * transformation matrix is copied from obj->obmat.
 *
 * @param obj   Object from which to copy the transformation matrix.
 */
AQTransform::AQTransform(const Object *obj)
{
        // copy the matrix
        memcpy(m_matrix, obj->obmat, 16*sizeof(float));
        
        // record the type of the source object
        m_srcObType = obj->type;
}



/**
 * AQTransform constructor.
 *
 * Constructs an AQTransform for a lamp.
 *
 * @param lamp  Lamp from which to construct the transformation matrix.
 */
AQTransform::AQTransform(const LampRen *lamp)
{
        float lampmat[3][3];
        memcpy(lampmat, lamp->mat, 9*sizeof(float));

        // set the type to lamp
        m_srcObType = OB_LAMP;

        // fetch the transform
        MTC_Mat4One(m_matrix);
        MTC_Mat4CpyMat3nc(m_matrix, lampmat);
        
        // set the scale factor to unity
        Normalise(m_matrix[0]);
        Normalise(m_matrix[1]);
        Normalise(m_matrix[2]);

        // set the position
        m_matrix[3][0] = lamp->co[0];
        m_matrix[3][1] = lamp->co[1];
        m_matrix[3][2] = lamp->co[2];
}



/**
 * Copy constructor.
 */
AQTransform::AQTransform(const AQTransform &xform)
{
        // delegate to the assignment operator
        (*this) = xform;
}



/**
 * Assignment operator.
 */
const AQTransform& AQTransform::operator=(const AQTransform& xform)
{
        memcpy(m_matrix, xform.m_matrix, 16*sizeof(float));
        m_srcObType = xform.m_srcObType;

        return (*this);
}




/**
 * Writes the transform as a camera viewing transformation.
 */
void AQTransform::writeCamera() const
{
        // check that the source object was a camera
        if (m_srcObType != OB_CAMERA)
        {
                cout << "Warning: Exporting a non-camera object "
                     << "transformation matrix as a camera." << endl;
        }

        // set the scale of the object transformation to unity
        float unscaled[4][4];
        memcpy(unscaled, m_matrix, 16*sizeof(float));
        Normalise(unscaled[0]);
        Normalise(unscaled[1]);
        Normalise(unscaled[2]);

        // invert the transformation
        float inverse[4][4];
        Mat4Invert(inverse, unscaled);
        inverse[0][2] = -inverse[0][2];
        inverse[1][2] = -inverse[1][2];
        inverse[2][2] = -inverse[2][2];
        inverse[3][2] = -inverse[3][2];

        // write out the transformation
        RiTransform(inverse);
}



/**
 * Writes the transform as an object transform.
 */
void AQTransform::writeObject() const
{
        // check that the source object was not a camera or light, since if it
        //  was, there's probably something amiss
        if ((m_srcObType == OB_LAMP) || (m_srcObType == OB_CAMERA))
        {
                cout << "Warning: Exporting a camera or lamp using an object "
                     << "transformation matrix." << endl;
        }

        float outmat[4][4];
        memcpy(outmat, m_matrix, 16*sizeof(float));
        RiTransform(outmat);
}



/**
 * Writes the transform as a light transform.
 *
 * Note: This method should be used when the light is actually lighting the
 *       scene.  writeLightShadowMap() should be used to position the light
 *       as a camera for generating a shadow map.
 */
void AQTransform::writeLight() const
{
        // check that the source object was a lamp
        if (m_srcObType != OB_LAMP)
        {
                cout << "Warning: Exporting a non-lamp using a lamp "
                     << "transformaton matrix." << endl;
        }

        float m[4][4];
        memcpy(m, m_matrix, 16*sizeof(float));

        i_rotate(180.0, 'x', m);

        RiTransform(m);
}



/**
 * Writes the transform as a light shadow map viewing transform.
 *
 * Note: This method sets the light up in the correct position to act as a
 *       camera to create a shadow map.  writeLight() should be used when
 *       the light is actually illuminating the scene.
 */
void AQTransform::writeLightShadowMap() const
{
        // check that the source object was a lamp
        if (m_srcObType != OB_LAMP)
        {
                cout << "Warning: Exporting a non-lamp using a lamp "
                     << "transformation matrix." << endl;
        }

        float m[4][4];
        memcpy(m, m_matrix, 16*sizeof(float));

        // invert the transform
        float inverse[4][4];
        Mat4Invert(inverse, m);
        inverse[0][2] = -inverse[0][2];
        inverse[1][2] = -inverse[1][2];
        inverse[2][2] = -inverse[2][2];
        inverse[3][2] = -inverse[3][2];

        // write out
        RiTransform(inverse);
}



/**
 * Copy constructor.
 */
AQTransformStack::AQTransformStack(const AQTransformStack& ostack)
{
        // fall back to the assignment operator
        (*this) = ostack;
}



/**
 * Assignment operator
 */
const AQTransformStack& AQTransformStack::operator=(
        const AQTransformStack& ostack)
{
        m_stack = ostack.m_stack;

        return (*this);
}



/**
 * Start the Motion block with a MotionBegin()
 */
void AQTransformStack::initMotion() const
{
        int ntimes = m_stack.size();
        assert(ntimes > 1);
        std::vector<RtFloat> times(ntimes);
        RtFloat fss = (RtFloat)(ntimes-1);
        for (int i=0; i < ntimes; i++)
                times[i] = (RtFloat)i / fss;
        RiMotionBeginV(ntimes, &times[0]);
}



/**
 * Finishes the motion block with a MotionEnd()
 */
void AQTransformStack::endMotion() const
{
        assert(m_stack.size() > 1);
        RiMotionEnd();
}



/**
 * Writes a transform stack as an object motion blur.
 *
 * @param mblur         true if motion blur is enabled
 */
void AQTransformStack::writeObjectMotion(bool mblur) const
{
        assert(m_stack.size() > 0);

        if (mblur & (m_stack.size() > 1))
        {
                initMotion();
                        for (unsigned int i=0; i < m_stack.size(); i++)
                                m_stack[i].writeObject();
                endMotion();
        }
        else
        {
                // if motion blur is not enabled, then we write out only the
                //  first transformation on the stack
                m_stack[0].writeObject();
        }
}



/**
 * Writes a transform stack as a camera motion blur.
 *
 * @param mblur         true if motion blur is enabled
 */
void AQTransformStack::writeCameraMotion(bool mblur) const
{
        assert(m_stack.size() > 0);

        if (mblur & (m_stack.size() > 1))
        {
                initMotion();
                        for (int i=0; i < (int)m_stack.size(); i++)
                                m_stack[i].writeCamera();
                endMotion();
        }
        else
        {
                // if motion blur is not enabled, then write out only the
                //  first transformation on the stack
                m_stack[0].writeCamera();
        }
}

