/**
 * Class declarations for Transformation-related classes.
 */
#ifndef __AQSIS_TRANSFORM_H
#define __AQSIS_TRANSFORM_H

#ifdef __cplusplus
extern "C" {
#endif

#include "DNA_object_types.h"	/* for struct Object  */
#include "render_types.h"       /* for struct LampRen */

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus

#include <vector>

using namespace std;



/**
 * Transformation matrix wrapper.
 *
 * This class wraps transformation matrices.  The purpose of this class is to
 * provide a single wrapper that can be passed around, instantiated easily, 
 * etc., and can handle all required transformation functions.  A 
 * transformation thus becomes an atomic object.
 *
 * Example:
 *   This class can be instantiated on a Blender object during rendering, and
 *   then used to export the transformation of that object:
 *
 *      Object *obj;
 *      AQTransform x(obj);
 *      x.writeObject();
 */
class AQTransform
{
        public:
                // constructor and destructor
                AQTransform();
                AQTransform(const Object *obj);
                AQTransform(const LampRen *lamp);
                AQTransform(const AQTransform &xform);
                ~AQTransform() {};

                // assignment operator
                const AQTransform& operator=(const AQTransform& xform);
                
                // output methods for various objects
                void writeCamera()         const;
                void writeObject()         const;
                void writeLight()          const;
                void writeLightShadowMap() const;

        private:
                // the stored transformation matrix itself
                float m_matrix[4][4];
                // type of the source object
                short m_srcObType;
};



/**
 * A stack of transformation matrices.
 *
 * This stack should be used for storing multiple transformation matrices for
 * motion blur output.  Note that the transformations are treated atomically
 * and are passed by-value (so all 16 elements + type are copied).
 */
class AQTransformStack
{
        public:
                // Constructor and destructor
                AQTransformStack() {}
                ~AQTransformStack() {}
                AQTransformStack(const AQTransformStack& ostack);

                // assignment operator
                const AQTransformStack& operator=(
                        const AQTransformStack& ostack);
                
                // Push transforms onto the stack
                void push(AQTransform xform)
                {
                        m_stack.push_back(xform);
                }

                // Output options for the stack
                void writeObjectMotion(bool mblur) const;
                void writeCameraMotion(bool mblur) const;

                // Returns the number of transforms on the stack
                int size() const
                {
                        return m_stack.size();
                }
                
        private:
                // the stack of transforms
                vector<AQTransform> m_stack;

        private:
                // start and finish a motion block
                void initMotion() const;
                void endMotion() const;
};



#endif /* __cplusplus */

#endif /* __AQSIS_TRANSFORM_H */

