/**
 * colorband.cpp
 *
 * Colorband functions.
 */

#include <assert.h>
#include <stdlib.h>
#include <iostream>

#include "colorband.h"
#include "util.h"
#include "error.h"



/**
 * Selects an input type for a color ramp.
 *
 * \param type          Must be one of: 
 *                              MA_RAMP_IN_SHADER
 *                              MA_RAMP_IN_ENERGY
 *                              MA_RAMP_IN_NOR
 * \param shader        Shader value.
 * \param energy        "Energy" received from a light source.
 * \param dpnorm        Dot product of normal with view vector.
 */
float DY_select_ramp_input(int type, float shader, float energy[3], 
        float dpnorm)
{
        switch (type)
        {
                case MA_RAMP_IN_SHADER:
                        return shader;
                case MA_RAMP_IN_ENERGY:
                        return COLOR_TO_FLOAT(energy);
                case MA_RAMP_IN_NOR:
                        return dpnorm;
                default:
                        DY_writeError("Unknown ramp input type encountered.");
                        return shader;
        }
}



/**
 * Merges a color ramp with an original base color.
 *
 * \param res   Result color.
 * \param orig  Original color.
 * \param cb    Color-band color.
 * \param fac   Blend factor.
 * \param type  Type of blend.
 */
void DY_merge_ramp_color(float res[3], float orig[3], float cb[3], float fac, 
        int type)
{
        float nfac;
        switch (type)
        {
                case MA_RAMP_BLEND:
                        nfac = 1.0-fac;
                        res[0] = nfac*orig[0] + fac*cb[0];
                        res[1] = nfac*orig[1] + fac*cb[1];
                        res[2] = nfac*orig[2] + fac*cb[2];
                        break;
                case MA_RAMP_ADD:
                        res[0] = orig[0] + fac*cb[0];
                        res[1] = orig[1] + fac*cb[1];
                        res[2] = orig[2] + fac*cb[2];
                        break;
                case MA_RAMP_MULT:
                        res[0] = orig[0] * cb[0] * fac;
                        res[1] = orig[1] * cb[1] * fac;
                        res[2] = orig[2] * cb[2] * fac;
                        break;
                case MA_RAMP_SUB:
                        res[0] = orig[0] - fac*cb[0];
                        res[1] = orig[1] - fac*cb[1];
                        res[2] = orig[2] - fac*cb[2];
                        break;
                default:
                        DY_writeError("Unknown color ramp output type.");
                        exit(-1);
        }
}



/**
 * set_four_ipo() is taken directly from key.c
 */
void DY_set_four_ipo(float d, float *data, int type)
{
        float d2, d3, fc;
	
        if (type==KEY_LINEAR) {
                data[0]= 0.0f;
                data[1]= 1.0f-d;
                data[2]= d;
                data[3]= 0.0f;
        }
        else 
        {
                d2= d*d;
                d3= d2*d;
		
                if (type==KEY_CARDINAL) 
                {
                        fc= 0.71f;
			
                        data[0]= -fc*d3	+2.0f*fc*d2 -fc*d;
                        data[1]= (2.0f-fc)*d3 +(fc-3.0f)*d2 +1.0f;
                        data[2]= (fc-2.0f)*d3 +(3.0f-2.0f*fc)*d2 +fc*d;
                        data[3]= fc*d3 -fc*d2;
                }
                else if (type==KEY_BSPLINE) 
                {
                        data[0]= -0.1666f*d3 +0.5f*d2 -0.5f*d +0.16666f;
                        data[1]= 0.5f*d3-d2+0.6666f;
                        data[2]= -0.5f*d3+0.5f*d2+0.5f*d+0.1666f;
                        data[3]= 0.1666f*d3;
                }
        }
}



/**
 * do_colorband() is taken directly from texture.c
 */
#ifndef CLAMP
#define CLAMP(a, b, c) if((a)<(b)) (a)=(b); else if((a)>(c)) (a)=(c)
#endif
int DY_do_colorband(ColorBand *coba, float in, float out[4])
{
        /* These vars form the texture channel, in render/intern/texture.c */
        CBData *cbd1, *cbd2, *cbd0, *cbd3;
        float fac, mfac, t[4];
        int a;
	
        if(coba->tot==0) return 0;
	
        cbd1= coba->data;
        if(coba->tot==1) 
        {
                out[0]= cbd1->r;
                out[1]= cbd1->g;
                out[2]= cbd1->b;
                out[3]= cbd1->a;
        }
        else 
        {
                if(in < cbd1->pos) in= cbd1->pos;

                /* we're looking for first pos > in */
                for(a=0; a<coba->tot; a++, cbd1++) if(cbd1->pos > in) break;
			
                if(a==coba->tot) cbd1--;
                if(in > cbd1->pos) in= cbd1->pos;
		
                cbd2= cbd1-1;
                fac= (in-cbd1->pos)/(cbd2->pos-cbd1->pos);
		
                if(coba->ipotype==2) 
                {
                        /* ipo from right to left: 3 2 1 0 */
			
                        if(a>=coba->tot-1) cbd0= cbd1;
                        else cbd0= cbd1+1;
                        if(a<2) cbd3= cbd2;
                        else cbd3= cbd2-1;
			
                        DY_set_four_ipo(fac, t, KEY_BSPLINE);

                        out[0]= t[3]*cbd3->r +t[2]*cbd2->r +
                                t[1]*cbd1->r +t[0]*cbd0->r;
                        out[1]= t[3]*cbd3->g +t[2]*cbd2->g +
                                t[1]*cbd1->g +t[0]*cbd0->g;
                        out[2]= t[3]*cbd3->b +t[2]*cbd2->b +
                                t[1]*cbd1->b +t[0]*cbd0->b;
                        out[3]= t[3]*cbd3->a +t[2]*cbd2->a +
                                t[1]*cbd1->a +t[0]*cbd0->a;
                        CLAMP(out[0], 0.0, 1.0);
                        CLAMP(out[1], 0.0, 1.0);
                        CLAMP(out[2], 0.0, 1.0);
                        CLAMP(out[3], 0.0, 1.0);
                }
                else 
                {
		        if(coba->ipotype==1) /* EASE */
                        {
                                mfac= fac*fac;
                                fac= 3.0f*mfac-2.0f*mfac*fac;
                        }
                        mfac= 1.0f-fac;
			
                        out[0]= mfac*cbd1->r + fac*cbd2->r;
                        out[1]= mfac*cbd1->g + fac*cbd2->g;
                        out[2]= mfac*cbd1->b + fac*cbd2->b;
                        out[3]= mfac*cbd1->a + fac*cbd2->a;
                }
        }
        return 1;	/* OK */
}
#undef CLAMP


