/**
 * context.cpp
 */

#include <sstream>

#include "context.h"
#include "colorband.h"
#include "error.h"
#include "util.h"


/**
 * Loads or returns a rendering context.
 *
 * Each Dobby context is indexed according to the filename of the file that
 * contains all of the rendering parameters.  This method checks to see if
 * the file containing the parameters has already been parsed.  If it has then
 * a handle to the context of those parameters is returned.  Otherwise, the
 * parameters file is parsed and a new handle is created.
 *
 * \param filename      Name of parameters file.
 *
 * \returns Rendering context for the specified parameters file, or NULL if
 *          an error occurs.
 */
DY_Context *DY_ContextManager::getContext(const string& filename)
{
        // check if we need to create a new context
        if (m_contexts.find(filename) == m_contexts.end())
        {
                DY_Context *c = DY_Context::createContext(filename);
                if (!c) 
                {
                        stringstream msg;
                        msg << "Could not load context for settings file: "
                            << "\"" << filename << "\"";
                        DY_writeError(msg.str());
                }
                m_contexts[filename] = c;
                
        }

        return m_contexts[filename];
}



/**
 * Constructs a new Dobby rendering context.
 *
 * \param filename      Name of parameters file for this context
 *
 * \returns Pointer to the new Dobby context, or NULL if an error occurs.
 */
DY_Context* DY_Context::createContext(const string& filename)
{
        // create the new context and attempt to set up its parameters
        //  manager by reading from the parameters file
        DY_Context *c = new DY_Context();
        if (!c->m_pman.addFile(filename))
        {
                delete c;
                return NULL;
        }

        // all went OK - return the new context
        return c;
}



/**
 * Calculates color ramp modifications to the diffuse color that occur within
 * the illuminance loop of diffuse shaders.
 *
 * This is a very "special-case" SHADEOP.  For correct usage, refer to the
 * shading functions in the SL include file surface_functions.h.
 *
 * \param result        Output color from the calculations.
 * \param shader        Value of the shader itself.
 * \param energy        Received "energy" from the light.
 * \param dpnorm        Dot product of the normal with the view vector.
 * \param diffuse_base  Base diffuse color before color ramps are applied.
 */
void DY_Context::colorramp_diffuse_illum(float result[3], float shader, 
        float energy[3], float dpnorm, float diffuse_base[3])
{
        // bail quickly if we're not using a color ramp for diffuse color
        //  lookup.  in this case, the result color is just the base diffuse
        //  color, unchanged.
        // if the color ramp calculations involve the result of shading then
        //  we also bail at this point.  result ramp calculations are handled
        //  by the colorramp_diffuse_result() method, and don't belong inside
        //  the innermost illuminance loop, which is where this method is
        //  being executed.
        if (!ramp_col__use() || (ramp_col__input() == MA_RAMP_IN_RESULT))
        {
                CP_TRIPLE(result, diffuse_base);
                return;
        }

        // select the ramp input type
        float fac = DY_select_ramp_input(ramp_col__input(), shader, energy,
                dpnorm);

        // construct the ColorBand object to be the same as Blender's
        //  representation
        ColorBand cb;
        constructColorband("ramp_col", &cb);

        // do the colorband interpolation look-up
        float cbres[4];
        DY_do_colorband(&cb, fac, cbres);
        fac = cbres[3] * ramp_col__factor();

        // merge the color appropriately with the base color
        DY_merge_ramp_color(result, diffuse_base, cbres, fac, 
                ramp_col__blend());
}



/**
 * Calculates color ramp modifications to the specular color that occur
 * within the illuminance loop of specular shaders.
 *
 * This is a very "special-case" SHADEOP.  For correct usage, refer to the
 * specular shading functions in the SL include file surface_functions.h
 *
 * \param result        Output color from the calculations.
 * \param shader        Value of the shader.
 * \param energy        Received "energy" from the light.
 * \param dpnorm        Dot product of the normal with the view vector.
 * \param specular_base Base specular color before color ramps are applied.
 */
void DY_Context::colorramp_specular_illum(float result[3], float shader,
        float energy[3], float dpnorm, float specular_base[3])
{
        // bail quickly if we're not using a color ramp, or if the color ramp
        //  calculations occur outside of the innermost illuminance loop
        if (!ramp_spec__use() || (ramp_spec__input() == MA_RAMP_IN_RESULT))
        {
                CP_TRIPLE(result, specular_base);
                return;
        }

        // select the ramp input type
        float fac = DY_select_ramp_input(ramp_spec__input(), shader, energy,
                dpnorm);

        // construct the ColorBand object to mimick Blender's internal
        //  representation
        ColorBand cb;
        constructColorband("ramp_spec", &cb);

        // do the colorband interpolation lookup
        float cbres[4];
        DY_do_colorband(&cb, fac, cbres);
        fac = cbres[3] * ramp_spec__factor();

        // merge the color appropriately with the base color
        DY_merge_ramp_color(result, specular_base, cbres, fac,
                ramp_spec__blend());
}



/**
 * Returns a single floating point parameter.
 *
 * \param group         Group.
 * \param param         Parameter.
 *
 * \returns     Floating-point parameter.
 */
float DY_Context::getFloat(const string& group, const string& param) const
{
        DY_ParamSet g;
        m_pman.getGroup(group, g);

        // getFloat can return floats or integers - first check for a float
        DY_Param *p = g.getParam(param, PARAMFLOAT);
        if (p)
        {
                assert(p->getType() == PARAMFLOAT);
                DY_FloatParam *pf = static_cast<DY_FloatParam*>(p);

                return pf->getFloat(0);
        }
        else
        {
                p = g.getParam(param, PARAMINT);
                if (p)
                {
                        assert(p->getType() == PARAMINT);
                        DY_IntParam *pf = static_cast<DY_IntParam*>(p);

                        int i = pf->getInt(0);
                        return static_cast<float>(i);
                }
        }
        
        stringstream msg;
        msg << "Could not find group: \"" << &group[0] << "\" "
            << ", float parameter: \"" << &param[0] << "\"";
        DY_writeError(msg.str());

        return 0;
}



/**
 * Returns a single integer parameter.
 *
 * \param group         Group.
 * \param name          Parameter.
 *
 * \returns     Integer parameter.
 */
int DY_Context::getInt(const string& group, const string& param) const
{
        DY_ParamSet g;
        m_pman.getGroup(group, g);
        DY_Param *p = g.getParam(param, PARAMINT);
        if (p)
        {
                assert(p->getType() == PARAMINT);
                DY_IntParam *pi = static_cast<DY_IntParam*>(p);

                return pi->getInt(0);
        }
        else
        {
                stringstream msg;
                msg << "Could not find group: \"" << &group[0] << "\" "
                    << ", int parameter: \"" << &param[0] << "\"";
                DY_writeError(msg.str());

                return 0;
        }
}



/**
 * Returns a single triple parameter.
 *
 * \param group         Group.
 * \param name          Parameter.
 *
 * \returns     Tuple parameter.
 */
void DY_Context::getTriple(float result[3], const string& group, 
        const string& param) const
{
        DY_ParamSet g;
        m_pman.getGroup(group, g);
        DY_Param *p = g.getParam(param, PARAMTRIPLE);
        if (p)
        {
                assert(p->getType() == PARAMTRIPLE);
                DY_TripleParam *pt = static_cast<DY_TripleParam*>(p);

                pt->getTriple(0, result);
                return;
        }
        else
        {
                stringstream msg;
                msg << "Could not find group: \"" << &group[0] << "\" "
                    << ", triple parameter: \"" << &param[0] << "\"";
                DY_writeError(msg.str());
        }
}



/**
 * Constructs a Blender ColorBand object from parameters found within a
 * particular group.
 *
 * The used parameters from within the group are the following:
 *      tot
 *      ipotype
 *
 * \param group         Name of the group in which to find the colorband
 *                       parameters.
 * \param cb            Colorband to populate.
 */
void DY_Context::constructColorband(const string& group, ColorBand *cb) const
{
        // fetch the appropriate group of parameters
        DY_ParamSet g;
        m_pman.getGroup(group, g);

        // populate the colorband
        cb->tot     = static_cast<short>(g.getIntParam("tot")->getInt(0));
        cb->ipotype = static_cast<short>(g.getIntParam("ipotype")->getInt(0));
        DY_TripleParam *colors = g.getTripleParam("colors");
        DY_FloatParam  *alpha  = g.getFloatParam("alpha");
        DY_FloatParam  *pos    = g.getFloatParam("pos");
        for (int i=0; i < 16; i++)
        {
                CBData *cbd = &cb->data[i];
                
                float color[3];
                colors->getTriple(i, color);
                cbd->r   = color[0];
                cbd->g   = color[1];
                cbd->b   = color[2];
                cbd->a   = alpha->getFloat(i);
                cbd->pos = pos->getFloat(i);
        }
}

