/**
 * light_functions.h
 *
 * Functions useful to the BL_light light source shader.
 */
#ifndef LIGHT_FUNCTIONS_H
#define LIGHT_FUNCTIONS_H



/**
 * Calculates the normal Blender linear falloff function (without Quad or
 * Sphere selected).
 *
 * r    - distance from the light source
 * dist - Blender dist parameter
 */
float linear_falloff (
	float r;
	float dist;
) {
	return ( dist / (dist + r) );
}



/**
 * Calculates the falloff function when only the Quad option is selected.
 *
 * r     - distance from the light source
 * dist  - Blender dist parameter
 * quad1 - Blender Quad1 parameter
 * quad2 - Blender Quad2 parameter
 */
float quad_only_falloff (
	float r;
	float dist;
	float quad1;
	float quad2;
) {
	float dist2 = dist * dist;
	return ( (dist / (dist + quad1*r)) * (dist2 / (dist2 + quad2*r*r)) );
}



/**
 * Calculates the falloff function when only the Sphere option is selected.
 *
 * r    - distance from the light source
 * dist - Blender dist parameter
 */
float sphere_only_falloff (
	float r;
	float dist;
) {
	float i = 0;
	if (r <= dist) {
		i = (dist - r) / (dist + r);
	}
	return i;
}



/**
 * Calculates the falloff function when both Quad and Sphere options are
 * selected.
 *
 * r     - distance from the light source
 * dist  - Blender dist parameter
 * quad1 - Blender Quad1 parameter
 * quad2 - Blender Quad2 parameter
 */
float quad_sphere_falloff (
	float r;
	float dist;
	float quad1;
	float quad2;
) {
	float i = 0;
	if (r <= dist) {
		float dist2 = dist * dist;
		float f1 = (dist - r) / (dist + quad1*r);
		float f2 = dist2 / (dist2 + quad2*r*r);
		i = f1*f2;
	}
	return i;
}



/**
 * Lamp light source.
 *
 * from   - location of the light source
 * colour - color of the light source (pre-multiplied by its intensity)
 * dist   - light source distance parameter
 * quad   - 1 if the Blender "Quad" option has been selected
 * sphere - 1 if the Blender "Sphere" option has been selected
 * quad1  - the Blender Quad1 setting
 * quad2  - the Blender Quad2 setting
 */
void lamp (
	point from;
	color colour;
	float dist;
	float quad;   /* boolean */
	float sphere; /* boolean */
	float quad1;
	float quad2;
) {
	extern vector L;
	extern color Cl;

	illuminate (from) {

		float r = length(L);
		float i;

		/* switch between falloff options */
		if ((sphere > 0.5) && (quad > 0.5)) {
			i = quad_sphere_falloff(r, dist, quad1, quad2);
		} else if ((sphere > 0.5) && (quad < 0.5)) {
			i = sphere_only_falloff(r, dist);
		} else if ((quad > 0.5) && (sphere < 0.5)) {
			i = quad_only_falloff(r, dist, quad1, quad2);
		} else {
			i = linear_falloff(r, dist);
		}

		Cl = colour * i;

	}

}



/**
 * Sun light source.
 *
 * colour - color of the light source (pre-multiplied by its intensity)
 */
void sun (
	color colour;
) {
	extern color Cl;

	point from = point "shader" (0,0,0);
	point to   = point "shader" (0,0,1);

	solar (to-from, 0) {
		Cl = colour;
	}
}



/**
 * Hemi light source.
 *
 * colour - color of the light source (pre-multiplied by its intensity)
 */
void hemi (
	color colour;
) {
	extern color Cl;

	point from = point "shader" (0,0,0);
	point to   = point "shader" (0,0,1);

	solar (to-from, PI/2) {
		Cl = colour;
	}
}



/**
 * Spot light source.
 *
 * from       - location of the light source
 * colour     - color of the light source (pre-multiplied by its intensity)
 * energy     - energy of the light source (used for shadow-only lights)
 * dist       - light source distance parameter
 * quad       - 1 if the Blender "Quad" option has been selected
 * sphere     - 1 if the Blender "Sphere" option has been selected
 * quad1      - the Blender Quad1 setting
 * quad2      - the Blender Quad2 setting
 * spotSi     - spotlight size (expressed as a dot product to the central axis)
 * spotBl     - spot blend (expressed as a dot product to the central axis)
 * square     - 1 if the Blender "Square" option has been selected
 * shadowmap  - name of a shadow map
 * samples    - number of shadow map samples
 * ssoft      - softness of the shadow map
 * bias       - shadow map bias
 * onlyshadow - marks a spotlight as only casting a shadow
 */
void spot (
	point from;
	color colour;
	float energy;
	float dist;
	float quad;	/* boolean */
	float sphere;	/* boolean */
	float quad1;
	float quad2;
	float spotSi;
	float spotBl;
	float square;	/* boolean */
	string shadowmap;
	float samples;
	float ssoft;
	float bias;
	float onlyshadow;
) {

	extern point Ps;
	extern color Cl;
	extern vector L;

	vector axis = vector "shader" (0,0,1);

	/* calculate the angular size of the spotlight */
	float angsize = acos(spotSi);
	if (square > 0.5) {
		/* square falloff */
		angsize = atan(sqrt(2)*tan(angsize));
	}

	illuminate (from, axis, angsize) {

		float r = length(L);

		/* calculate edge falloff softness */
		vector Ln = normalize(L);
		vector Lshad = normalize(vtransform("shader", L));
		float inpr = Ln.axis;
		if (square > 0.5) {
			/* square falloff */
			if (inpr > 0) {
				float x = xcomp(Lshad);
				float y = ycomp(Lshad);
				float z = zcomp(Lshad);
				float v = max(abs(x/z), abs(y/z));
				inpr = 1 / sqrt(1 + v*v);
			} else {
				inpr = 0;
			}
		}
		float soft = 0;
		if (inpr >= spotSi) {
			float t = inpr - spotSi;
			if ((t < spotBl) && (spotBl != 0)) {
				float i = t / spotBl;
				soft = 3*i*i - 2*i*i*i;
				soft *= inpr;
			} else {
				soft = inpr;
			}
		} 

		/* calculate intensity based on distance falloff */
		float intensity;
		if ((sphere > 0.5) && (quad > 0.5)) {
			intensity = quad_sphere_falloff(r, dist, quad1, quad2);
		} else if ((sphere > 0.5) && (quad < 0.5)) {
			intensity = sphere_only_falloff(r, dist);
		} else if ((quad > 0.5) && (sphere < 0.5)) {
			intensity = quad_only_falloff(r, dist, quad1, quad2);
		} else {
			intensity = linear_falloff(r, dist);
		}

		/* sample any shadow map that is present */
		float shad = 0;
		float recvshadow;
		surface("recvshadow", recvshadow);
		if (recvshadow > 0.0) {
			float blur = ssoft / 1000.0;
			float realbias = bias * 0.1;
			float realsamples = samples * 4;
			if (shadowmap != "") {
				shad = shadow( shadowmap, Ps, 
					"samples", realsamples,
					"blur", blur,
					"bias", realbias );
			}
		}
		shad *= recvshadow;

		if (onlyshadow > 0.5) {
			Cl = -energy * intensity * shad;
		} else {
			Cl = colour * intensity * soft * (1-shad);
		}

	}

}


#endif /* LIGHT_FUNCTIONS_H */
