/**
 * surface_functions.h
 *
 * Functions useful to the BL_surface shader.  Initially, these are direct
 * transcriptions of the Blender shading functions to the RenderMan shading
 * language.
 */

#ifndef SURFACE_FUNCTIONS_H
#define SURFACE_FUNCTIONS_H

#include "util.h"

/** Diffuse shader types. */
#define D_TOON          2
#define D_ORENNAYAR     1
#define D_LAMBERT       0

/** Specular shader types. */
#define S_TOON          3
#define S_BLINN         2
#define S_PHONG         1
#define S_COOKTORRENCE  0



/**
 * The Blender basic specular function.
 *
 * Blender has a much longer and verbose version of this function (which may
 * or may not be faster).  This is a much cut-down version that's a little
 * smarter although it may not be *exactly* the same.
 *
 * \param inp   Input value.
 * \param hard  Surface hardness.
 *
 * \return      Basically calculates: pow(inp, hard)
 */
float spec(
	float inp; 
	float hard;
) {
        return pow(clamp(inp,0,1), hard);
}



/**
 * Lambertian diffuse.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward-facing normal.
 * \param diffusecolor  Diffuse color.
 * \param Kref          Reflectance.
 */
color lambertian_diffuse (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
        color diffusecolor;
        float Kref;
) {
	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        color C = 0;
        illuminance(P, Nf, PI/2) 
        {
        
                /* standard shading calculations */
		float nondiffuse = 0;
		lightsource("__nondiffuse", nondiffuse);

                float fshad = (1-nondiffuse) * Nf.normalize(L);
                color energy = Cl * fshad * Kref;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_diffuse = dobby_colorramp_diffuse_illum(
                        pfile, fshad, energy, dpnorm, diffusecolor
                );

                /* add the contribution from the current lamp */
                C += energy * final_diffuse;
	}

	return C;
}



/**
 * Phong specular.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward facing normal.
 * \param V             View vector.
 * \param specularcolor Specular color.
 */
color phong_specular (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
	vector V;	/* V  = -normalize(I) */
        color  specularcolor;
) {
	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float hard    = dobby_get(pfile, "shaders", "Ks_hard");
        uniform float Ks_spec = dobby_get(pfile, "shaders", "Ks_spec");

	color S = 0;
	illuminance(P, Nf, PI/2) {

                /* standard shading calculations */
		float nonspecular = 0, negative = 0;
		lightsource("__nonspecular", nonspecular);
		lightsource("negative", negative);
	
		vector H = normalize(normalize(L) + V);
		float rslt = H.Nf;
		if (rslt > 0.0) {
			rslt = spec(rslt, hard);
		} else {
			rslt = 0;
		}

                float fshad = (1-nonspecular) * (1-negative) * rslt;
                color energy = Cl * fshad * Ks_spec;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_specular = dobby_colorramp_specular_illum(
                        pfile, fshad, energy, dpnorm, specularcolor
                );

                /* add the contribution from the current lamp */
                S += energy * final_specular;

	}
	
	return S;
}



/**
 * Oren-Nayar diffuse.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward facing normal.
 * \param V             View vector.
 * \param diffusecolor  Diffuse color.
 * \param Kref          Reflectance.
 */
color oren_nayar_diffuse (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
	vector V;	/* V  = -normalize(I) */
        color  diffusecolor;
        float  Kref;
) {

	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float rough = dobby_get(pfile, "shaders", "Kd_rough");

	color D = 0;
	illuminance(P, Nf, PI/2) {

                /* shading calculations */
		float nondiffuse = 0;
		lightsource("__nondiffuse", nondiffuse);
	
		vector Ln = normalize(L);
		float nl = Nf.Ln;
		
		vector h  = normalize(V + Ln);
		float nh  = max(0, Nf.h);
		float nv  = max(0, Nf.V);
		float vh  = max(0, V.h);

		float  Lit_A  = acos(nl);
		float  View_A = acos(nv);
		vector Lit_B  = normalize(Ln - nl*Nf);
		vector View_B = normalize(V - nv*Nf);

		float t = max(0, Lit_B.View_B);
		float a, b;
		if (Lit_A > View_A) {
			a = Lit_A;  b = View_A;
		} else {
			a = View_A; b = Lit_A;
		}

		float rough2 = rough*rough;
		float A = 1 - (0.5 * rough2 / (rough2 + 0.33));
		float B = 0.45 * (rough2 / (rough2 + 0.09));
			
		b *= 0.95;
		
                float fshad = (1-nondiffuse) * nl * (A + (B*t*sin(a)*tan(b)));
                color energy = Cl * fshad * Kref;
                
                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_diffuse = dobby_colorramp_diffuse_illum(
                        pfile, fshad, energy, dpnorm, diffusecolor
                );
                
                /* add the contribution from the current lamp */
		D += energy * final_diffuse;

	}

	return D;

}



/**
 * Toon diffuse.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward facing normal.
 * \param V             View vector.
 * \param diffusecolor  Diffuse color.
 * \param Kref          Reflectance.
 */
color toon_diffuse (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
	vector V;	/* V  = -normalize(I) */
        color  diffusecolor;
        float  Kref;
) {

	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float size   = dobby_get(pfile, "shaders", "Kd_size");
        uniform float smooth = dobby_get(pfile, "shaders", "Kd_smooth");

	color D = 0;
	illuminance(P, Nf, PI/2) 
        {

                /* standard shading calculations */
		float nondiffuse = 0;
		lightsource("__nondiffuse", nondiffuse);

		float rslt = Nf.normalize(L);
		float ang  = acos(rslt);

		if (ang < size) {
			rslt = 1.0;
		} else if ((ang >= (size+smooth)) || (smooth <= 0.0)) {
			rslt = 0.0;
		} else {
			rslt = 1.0 - ((ang - size) / smooth);
		}

                float fshad = (1-nondiffuse) * rslt;
                color energy = Cl * fshad * Kref;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_diffuse = dobby_colorramp_diffuse_illum(
                        pfile, fshad, energy, dpnorm, diffusecolor
                );
                
                /* add the contribution from the current lamp */
		D += energy * final_diffuse;
	}

	return D;
}



/**
 * Cook-Torrence specular.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward facing normal.
 * \param V             View vector.
 * \param specularcolor Specular color.
 */
color cook_torrence_specular (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
	vector V;	/* V  = -normalize(I) */
        color  specularcolor;
) {

	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float hard    = dobby_get(pfile, "shaders", "Ks_hard");
        uniform float Ks_spec = dobby_get(pfile, "shaders", "Ks_spec");

	color S = 0;
	illuminance(P, Nf, PI/2) {

                /* standard shading calculations */
		float nonspecular = 0, negative = 0;
		lightsource("__nonspecular", nonspecular);
		lightsource("negative", negative);

		vector h = normalize(normalize(L) + V);
		float nh = Nf.h;
		float nv = max(0, Nf.V);

		float i = spec(nh, hard) / (0.1 + nv);

                float fshad = (1-nonspecular) * (1-negative) * i;
                color energy = Cl * fshad * Ks_spec;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_specular = dobby_colorramp_specular_illum(
                        pfile, fshad, energy, dpnorm, specularcolor
                );

                /* add the contribution from the current lamp */
                S += energy * final_specular;
	}

	return S;

}



/**
 * Blinn specular.
 *
 * pfile          Dobby parameter file.
 * Nf             Forward facing normal.
 * V              View vector.
 * specularcolor  Specular color.
 */
color blinn_specular (
        uniform string pfile;
	normal Nf;		/* Nf = faceforward (normalize(N), I) */
	vector V;		/* V  = -normalize(I) */
        color  specularcolor;
) {
	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float refrac     = dobby_get(pfile, "shaders", "Ks_refr");
        uniform float spec_power = dobby_get(pfile, "shaders", "Ks_hard");
        uniform float Ks_spec    = dobby_get(pfile, "shaders", "Ks_spec");

	float sp;
	if (spec_power < 100) {
		sp = sqrt(1 / spec_power);
	} else {
		sp = 10 / spec_power;
	}
	float sp2 = sp * sp;

	color S = 0;
	illuminance(P, Nf, PI/2) {

                /* standard shading calculations */
		float nonspecular = 0, negative = 0;
		lightsource("__nonspecular", nonspecular);
		lightsource("negative", negative);

		vector Ln = normalize(L);
		vector h  = normalize(Ln + V);
		float nh = max(0, Nf.h);
		float nv = Nf.V;
		if (nv < 0.0)
			nv = 0.01;
		float nl = max(0, Nf.Ln);
		float vh = V.h;
		if (vh < 0.0)
			vh = 0.01;

		float a = 1;
		float b = 2 * nh * nv / vh;
		float c = 2 * nh * nl / vh;

		float g = 0;
		if ((a < b) && (a < c))
			g = a;
		else if ((b < a) && (b < c))
			g = b;
		else if ((c < a) && (c < b))
			g = c;
		
		float p = sqrt(refrac*refrac + vh*vh - 1);

		float pnvh2  = (p - vh) * (p - vh);
		float ppvh2  = (p + vh) * (p + vh);
		float ppvhn2 = (vh*(p+vh)-1) * (vh*(p+vh)-1);
		float pnvhn2 = (vh*(p-vh)+1) * (vh*(p-vh)+1);
		float f = (pnvh2 / ppvh2) * (1 + ppvhn2 / pnvhn2);

		float ang = acos(nh);
		float ang2 = ang * ang;

		float i = f * g * exp(-ang2 / (2*sp2));

                float fshad = (1-nonspecular) * (1-negative) * i;
                color energy = Cl * fshad * Ks_spec;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_specular = dobby_colorramp_specular_illum(
                        pfile, fshad, energy, dpnorm, specularcolor
                );

                /* add the contribution from the current lamp */
		S += energy * final_specular;

	}

	return S;
	
}



/**
 * Toon specular.
 *
 * \param pfile         Dobby parameter file.
 * \param Nf            Forward facing normal.
 * \param V             View vector.
 * \param specularcolor Specular color.
 */
color toon_specular (
        uniform string pfile;
	normal Nf;	/* Nf = faceforward (normalize(N), I) */
	vector V;	/* V  = -normalize(I) */
        color  specularcolor;
) {

	extern point P;
	extern vector L;
	extern color Cl;
        extern vector I;

        uniform float size    = dobby_get(pfile, "shaders", "Ks_size");
        uniform float smooth  = dobby_get(pfile, "shaders", "Ks_smooth");
        uniform float Ks_spec = dobby_get(pfile, "shaders", "Ks_spec");

	color S = 0;
	illuminance(P, Nf, PI/2) {

                /* standard shading calculations */
		float nonspecular = 0, negative = 0;
		lightsource("__nonspecular", nonspecular);
		lightsource("negative", negative);

		vector h = normalize(normalize(L) + V);
		float rslt = h.Nf;
		float ang = acos(rslt);

		if (ang < size) {
			rslt = 1;
		} else if ((ang > (size+smooth)) || (smooth < EPSILON)) {
			rslt = 0;
		} else {
			rslt = 1 - ((ang-size) / smooth);
		}

                float fshad = (1-nonspecular) * (1-negative) * rslt;
                color energy = Cl * fshad * Ks_spec;

                /* color ramp */
                float dpnorm = abs(normalize(Nf).normalize(I));
                color final_specular = dobby_colorramp_specular_illum(
                        pfile, fshad, energy, dpnorm, specularcolor
                );

                /* add the contribution from the current lamp */		
		S += energy * final_specular;

	}

	return S;
	
}



#endif /* SURFACE_FUNCTIONS_H */

