/**
 * util.h
 *
 * Utility functions.
 */

#ifndef UTIL_H
#define UTIL_H

/**
 * EPSILON - A small value which is larger than the expected floating point
 *           error.
 */
#define EPSILON 1.0e-5

/**
 * IX - Adds a small amount to a floating point value to make it "safe" to
 *      use as an array index.
 *
 * From the RI Spec 3.3:
 *      "An array element index is a float expression, which is rounded down
 *       to generate an integer index."
 *
 * If the float expression is to be rounded down to generate the array index,
 * the reasoning here is that we should bump it up a little first, in case
 * floating point imprecision leaves it just *below* the integer we're 
 * expecting.  I don't know if that will ever happen, but just-in-case.
 */
#define IX(x) ((x) + 10*EPSILON)

/**
 * FEQ - Test for equality of a pair of floating point values up to the
 *       precision EPSILON.
 *
 * \param x     One value to test.
 * \param y     Second value to test.
 *
 * \returns     The result of the test (x == y).
 */
#define FEQ(x, y) ( float dobby_feq(x,y) )

/**
 * NFEQ - Test for inequality between a pair of floating point values up to
 *        the precision EPSILON.
 *
 * \param x     First value to test.
 * \param y     Second value to test.
 *
 * \returns     The result of the test (x != y).
 */
#define NFEQ(x, y) \
        ( \
                ( \
                        (x) + (EPSILON) < (y) \
                ) \
                && \
                ( \
                        (x) - (EPSILON) < (y) \
                ) \
        ) || ( \
                ( \
                        (x) + (EPSILON) > (y) \
                ) \
                && \
                ( \
                        (x) - (EPSILON) > (y) \
                ) \
        )

/**
 * Converts a color to b&w using the Blender formula for energy:
 *      0.3 red + 0.58 green + 0.12 blue
 */
float colorToFloat(
        color c;
) {
        return 0.30*comp(c,0) + 0.58*comp(c,1) + 0.12*comp(c,2);
}

#endif /* UTIL_H */
