/*
 * $Id: KX_Frustum.cpp 721 2004-07-10 11:54:00Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 * Camera in the gameengine. Cameras are also used for views.
 */
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "KX_Frustum.h"


void KX_IFrustum::ExtractClipPlanes()
{
	if (!m_dirty)
		return;

	MT_Matrix4x4 m = GetProjectionMatrix() * GetWorldToCamera();
	// Left clip plane
	m_planes[0] = m[3] + m[0];
	// Right clip plane
	m_planes[1] = m[3] - m[0];
	// Top clip plane
	m_planes[2] = m[3] - m[1];
	// Bottom clip plane
	m_planes[3] = m[3] + m[1];
	// Near clip plane
	m_planes[4] = m[3] + m[2];
	// Far clip plane
	m_planes[5] = m[3] - m[2];
	
	m_dirty = false;
	m_normalised = false;
}


void KX_IFrustum::NormaliseClipPlanes()
{
	if (m_normalised)
		return;
	
	for (unsigned int p = 0; p < 6; p++)
	{
		MT_Scalar factor = sqrt(m_planes[p][0]*m_planes[p][0] + m_planes[p][1]*m_planes[p][1] + m_planes[p][2]*m_planes[p][2]);
		if (!MT_fuzzyZero(factor))
			m_planes[p] /= factor;
	}
	
	m_normalised = true;
}


void KX_IFrustum::ExtractFrustumSphere()
{
	if (m_set_frustum_centre)
		return;

	// The most extreme points on the near and far plane. (normalised device coords)
	MT_Vector4 hnear(1., 1., 0., 1.), hfar(1., 1., 1., 1.);
	MT_Matrix4x4 clip_camcs_matrix = GetProjectionMatrix();
	clip_camcs_matrix.invert();
	
	// Transform to hom camera local space
	hnear = clip_camcs_matrix*hnear;
	hfar = clip_camcs_matrix*hfar;
	
	// Tranform to 3d camera local space.
	MT_Point3 nearpoint(hnear[0]/hnear[3], hnear[1]/hnear[3], hnear[2]/hnear[3]);
	MT_Point3 farpoint(hfar[0]/hfar[3], hfar[1]/hfar[3], hfar[2]/hfar[3]);
	
	// Compute centre
	m_frustum_centre = MT_Point3(0., 0.,
		(nearpoint.dot(nearpoint) - farpoint.dot(farpoint))/(2.0*(GetCameraFar() - GetCameraNear())));
	m_frustum_radius = m_frustum_centre.distance(farpoint);
	
	// Transform to world space.
	m_frustum_centre = GetCameraToWorld()(m_frustum_centre);
	m_frustum_radius /= fabs(GetWorldScaling()[GetWorldScaling().closestAxis()]);
	
	m_set_frustum_centre = true;
}


int KX_IFrustum::SphereInsideFrustum(const MT_Point3& centre, const MT_Scalar &radius)
{
	ExtractFrustumSphere();
	if (centre.distance2(m_frustum_centre) > (radius + m_frustum_radius)*(radius + m_frustum_radius))
		return OUTSIDE;

	unsigned int p;
	ExtractClipPlanes();
	NormaliseClipPlanes();
		
	MT_Scalar distance;
	int intersect = INSIDE;
	// distance:  <-------- OUTSIDE -----|----- INTERSECT -----0----- INTERSECT -----|----- INSIDE -------->
	//                                -radius                                      radius
	for (p = 0; p < 6; p++)
	{
		distance = m_planes[p][0]*centre[0] + m_planes[p][1]*centre[1] + m_planes[p][2]*centre[2] + m_planes[p][3];
		if (fabs(distance) <= radius)
			intersect = INTERSECT;
		else if (distance < -radius)
			return OUTSIDE;
	}
	
	return intersect;
}


int KX_IFrustum::BoxInsideFrustum(const MT_Point3 *box)
{
	ExtractClipPlanes();
	
	unsigned int insideCount = 0;
	// 6 view frustum planes
	for( unsigned int p = 0; p < 6 ; p++ )
	{
		unsigned int behindCount = 0;
		// 8 box verticies.
		for (unsigned int v = 0; v < 8 ; v++)
		{
			if (m_planes[p][0]*box[v][0] + m_planes[p][1]*box[v][1] + m_planes[p][2]*box[v][2] + m_planes[p][3] < 0.)
				behindCount++;
		}
		
		// 8 points behind this plane
		if (behindCount == 8)
			return OUTSIDE;

		// Every box vertex is on the front side of this plane
		if (!behindCount)
			insideCount++;
	}
	
	// All box verticies are on the front side of all frustum planes.
	if (insideCount == 6)
		return INSIDE;
	
	return INTERSECT;
}


bool KX_IFrustum::PointInsideFrustum(const MT_Point3& x)
{
	ExtractClipPlanes();
	
	for( unsigned int i = 0; i < 6 ; i++ )
	{
		if (m_planes[i][0]*x[0] + m_planes[i][1]*x[1] + m_planes[i][2]*x[2] + m_planes[i][3] < 0.)
			return false;
	}
	return true;
}


void KX_IFrustum::SetDirty()
{
	m_dirty = true;
	m_set_frustum_centre = false;
}


bool KX_IFrustum::GetFrustumCulling() const
{
	return m_frustum_culling;
}


void KX_IFrustum::SetFrustumCulling(bool cull)
{
	m_frustum_culling = cull;
}

