/**
 * $Id: KX_PolygonMaterial.cpp 843 2004-09-18 06:01:59Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "KX_PolygonMaterial.h"

#include "BKE_mesh.h"
#include "BKE_global.h"
#include "BKE_image.h"

extern "C" {
#include "BDR_drawmesh.h"
}

#include "DNA_material_types.h"
#include "DNA_texture_types.h"
#include "DNA_image_types.h"
#include "DNA_mesh_types.h"

#include "IMB_imbuf_types.h"

#include "MEM_guardedalloc.h"

#include "RAS_LightObject.h"
#include "RAS_IRenderTools.h"
#include "RAS_MaterialBucket.h"

#include "KX_PyMath.h"

/**
 *  This struct is for passing a ARB_*_program from Python
 */
struct KX_PolygonMaterial::Program
{
	STR_String m_program;
	STR_String m_name;
	Program(STR_String program, STR_String name)
		: m_program(program),
		  m_name(name)
	{
	}
};

/**
 *  This struct is the shader passed from Python.
 */
struct KX_PolygonMaterial::Shader
{
	RAS_IRasterizer::IShader::Target m_target;
	STR_String m_shader;
	STR_String m_name;
	Shader(RAS_IRasterizer::IShader::Target target, STR_String shader, STR_String name)
		: m_target(target),
		  m_shader(shader),
		  m_name(name)
	{
	}
};

/**
 *  This class is for attribute binding.
 */
class KX_PolygonMaterial::Attribute
{
	STR_String	m_name;
	int		m_location;
public:
	Attribute(const STR_String &name, int loc)
		: m_name(name),
		  m_location(loc)
	{
	}
	
	~Attribute()
	{
	}
	
	void BindAttribute(RAS_IRasterizer::IShader *shader)
	{
		shader->BindAttribute(m_name, m_location);
	}
};	

/**
 *  This is the uniform base class.
 */
class KX_PolygonMaterial::Uniform
{
protected:
	STR_String	m_name;
	
	Uniform(const STR_String &name)
		: m_name(name)
	{}
public:
	virtual ~Uniform()
	{}
	
	/**
	 *  Set the uniform when material is activated.
	 */
	virtual void Activate(RAS_IRasterizer* rasty, RAS_IRasterizer::IShader *shader) = 0;
	/**
	 *  Set the uniform for each meshslot.
	 */
	virtual void Update(const KX_MeshSlot&, RAS_IRasterizer *, RAS_IRasterizer::IShader *)
	{
	}
	
	STR_String&  GetName()
	{
		return m_name;
	}
};

/**
 *  This is for vector and scalar uniforms.
 */
template<class T>
class KX_PolygonMaterial::VecUniform : public KX_PolygonMaterial::Uniform
{
	int		m_index;
	T		m_value;
	
public:
	VecUniform(const STR_String &name, T value)
		: Uniform(name),
		  m_index(-1),
		  m_value(value)
	{
	}
	
	virtual ~VecUniform()
	{
	}
	
	virtual void Activate(RAS_IRasterizer*, RAS_IRasterizer::IShader *shader)
	{
		if (m_index < 0)
			m_index = shader->GetUniformLocation(m_name);
		if (m_index >= 0)
			shader->SetUniform(m_index, m_value);
	}
};

/**
 *  This is for Matrix uniforms - they have an optional transpose parameter.
 */
template<class T>
class KX_PolygonMaterial::MatUniform : public KX_PolygonMaterial::Uniform
{
	int		m_index;
	T		m_value;
	bool		m_transpose;
	
public:
	MatUniform(const STR_String &name, T value, bool transpose = false)
		: Uniform(name),
		  m_index(-1),
		  m_value(value),
		  m_transpose(transpose)
	{
	}
	
	virtual ~MatUniform()
	{
	}
	
	virtual void Activate(RAS_IRasterizer*, RAS_IRasterizer::IShader *shader)
	{
		if (m_index < 0)
			m_index = shader->GetUniformLocation(m_name);
		if (m_index >= 0)
			shader->SetUniform(m_index, m_value, m_transpose);
	}
};

/**
 *  This is for uniforms that are generated.
 */
class KX_PolygonMaterial::GenUniform : public KX_PolygonMaterial::Uniform
{
public:
	enum GenType {
		// Model view matrix (Obsoleted by GLSL v1.1
		// MV = V*M
		MODELVIEWMATRIX,
		MODELVIEWMATRIX_TRANSPOSE,
		MODELVIEWMATRIX_INVERSE,
		MODELVIEWMATRIX_INVERSETRANSPOSE,
	
		// Model matrix
		MODELMATRIX,
		MODELMATRIX_TRANSPOSE,
		MODELMATRIX_INVERSE,
		MODELMATRIX_INVERSETRANSPOSE,
	
		// View Matrix
		VIEWMATRIX,
		VIEWMATRIX_TRANSPOSE,
		VIEWMATRIX_INVERSE,
		VIEWMATRIX_INVERSETRANSPOSE,
		
		NUMLIGHTS,
		
		// Current Eye (Stereo)
		EYE
	};
	
	GenUniform(const STR_String &name, GenType value)
		: Uniform(name),
		  m_index(-1),
		  m_value(value)
	{
	}
	
	virtual ~GenUniform()
	{
	}
	
	virtual void Activate(RAS_IRasterizer *rasty, RAS_IRasterizer::IShader *shader)
	{
		if (m_index < 0)
			m_index = shader->GetUniformLocation(m_name);
		if (m_index >= 0)
		{
			MT_Matrix4x4 view;
			rasty->GetViewMatrix(view);
			
			switch (m_value)
			{
				case VIEWMATRIX:
					shader->SetUniform(m_index, view);
					break;
				case VIEWMATRIX_TRANSPOSE:
					shader->SetUniform(m_index, view, true);
					break;
				case VIEWMATRIX_INVERSE:
					view.invert();
					shader->SetUniform(m_index, view);
					break;
				case VIEWMATRIX_INVERSETRANSPOSE:
					view.invert();
					shader->SetUniform(m_index, view, true);
					break;
				/*
				case NUMLIGHTS:
					shader->SetUniform(m_index, (unsigned int) material->m->m_numlights);
					break;
				*/
				case EYE:
					shader->SetUniform(m_index, (unsigned int) (rasty->GetEye() == RAS_IRasterizer::RAS_STEREO_LEFTEYE?0:1));
					break;
				default:
					break;
			}
		}
	}
	
	virtual void Update(const KX_MeshSlot &ms, RAS_IRasterizer *rasty, RAS_IRasterizer::IShader *shader)
	{
		if (m_index >= 0)
		{
			MT_Matrix4x4 model;
			model.setValue(ms.m_OpenGLMatrix);
			MT_Matrix4x4 view;
			rasty->GetViewMatrix(view);
			switch (m_value)
			{
				case MODELMATRIX:
					shader->SetUniform(m_index, model);
					break;
				case MODELMATRIX_TRANSPOSE:
					shader->SetUniform(m_index, model, true);
					break;
				case MODELMATRIX_INVERSE:
					model.invert();
					shader->SetUniform(m_index, model);
					break;
				case MODELMATRIX_INVERSETRANSPOSE:
					model.invert();
					shader->SetUniform(m_index, model, true);
					break;
				case MODELVIEWMATRIX:
					shader->SetUniform(m_index, view*model);
					break;
				case MODELVIEWMATRIX_TRANSPOSE:
				{
					MT_Matrix4x4 mat(view*model);
					shader->SetUniform(m_index, mat, true);
					break;
				}
				case MODELVIEWMATRIX_INVERSE:
				{
					MT_Matrix4x4 mat(view*model);
					mat.invert();
					shader->SetUniform(m_index, mat);
					break;
				}
				case MODELVIEWMATRIX_INVERSETRANSPOSE:
				{
					MT_Matrix4x4 mat(view*model);
					mat.invert();
					shader->SetUniform(m_index, mat, true);
					break;
				}
				default:
					break;
			}
		}
	}
	
private:
	int		m_index;
	GenType		m_value;
};

/**
 * Private to KX_PolygonMaterial 
 * 
 * (IRIX can't use foward declared types in std::vector)
 */
class KX_PolygonMaterial::Private
{
public:
	enum LightState
	{
		AMBIENT,
		DIFFUSE,
		TEXTURE,
		SPECULAR,
		FINISHED
	};

	std::vector<RAS_IRasterizer::IProgram*>		m_programs;
	std::vector<KX_PolygonMaterial::Program>	m_programsource;
	RAS_IRasterizer::IShader*			m_shader;
	
	std::vector<RAS_IRasterizer::ITexture*>		m_textures;
	std::vector<KX_PolygonMaterial::Shader>		m_shadersource;
	std::vector<KX_PolygonMaterial::Uniform*>	m_uniforms;
	std::vector<KX_PolygonMaterial::Attribute*>	m_attributes;
	
	/** Lighting pass variables */
	bool			m_doshadows;
	unsigned int		m_numlights;
	unsigned int		m_lightpass;
	
	LightState			m_lightstate;
	
	Private()
		: m_shader(0),
		m_lightpass(0),
		m_lightstate(AMBIENT)
	{
	}
	
	~Private()
	{
		delete m_shader;
		
		std::vector<RAS_IRasterizer::ITexture*>::iterator tit = m_textures.begin();
		for (; tit != m_textures.end(); ++tit)
			delete (*tit);
		m_textures.clear();
		
		std::vector<Uniform*>::iterator uit = m_uniforms.begin();
		for (; uit != m_uniforms.end(); ++uit)
			delete (*uit);
		m_uniforms.clear();
		
		for (std::vector<Attribute*>::iterator ait = m_attributes.begin(); ait != m_attributes.end(); ++ait)
			delete (*ait);
		m_attributes.clear();
	}

};


KX_PolygonMaterial::KX_PolygonMaterial(const STR_String &texname,
											   bool ba,
											   Material *material,
											   int tile,
											   int tilexrep,
											   int tileyrep,
											   int mode,
											   bool smooth,
											   bool transparant,
											   bool zsort,
											   int subsurf,
											   bool linearsubsurf,
											   int lightlayer,
											   bool bIsTriangle,
											   void* clientobject,
											   struct TFace* tface,
											   RAS_IRenderTools* rendertools,
											   PyTypeObject *T)
		: PyObjectPlus(T),
		  RAS_IPolyMaterial(texname,
							false,
							material?STR_String(material->id.name):"",
							tile,
							tilexrep,
							tileyrep,
							mode,
							smooth,
							transparant,
							zsort,
							subsurf,
							linearsubsurf,
							lightlayer,
							bIsTriangle,
							clientobject),
		m(new Private()),
		m_tface(tface),
		m_material(material),
		m_rendertools(rendertools),
		m_pymaterial(0)
{
}

KX_PolygonMaterial::~KX_PolygonMaterial()
{
	delete m;
	
	if (m_pymaterial)
		Py_DECREF(m_pymaterial);
}

bool KX_PolygonMaterial::ActivatePrograms(RAS_IRasterizer *rasty) const
{
	bool activated = false;
	std::vector<RAS_IRasterizer::IProgram*>::iterator pit;
	if (!m->m_programsource.size()) // No new shaders
	{
		for (pit = m->m_programs.begin(); pit != m->m_programs.end(); ++pit)
		{
			if ((*pit)->isValid())
			{
				(*pit)->Enable();
				activated = true;
			}
		}
		return activated;
	}
	
	for (pit = m->m_programs.begin(); pit != m->m_programs.end(); ++pit)
		delete (*pit);
	m->m_programs.clear();
	
	std::vector<Program>::iterator psit;
	for (psit = m->m_programsource.begin(); psit != m->m_programsource.end(); ++psit)
	{
		RAS_IRasterizer::IProgram *program = NULL;
		if ((*psit).m_program.Length() > 24) // Program must be source
			program = rasty->NewProgram((*psit).m_program, (*psit).m_name);
		else
		{
			// Program might be name of text datablock...
			STR_String source(m_rendertools->GetText((*psit).m_program));
			program = rasty->NewProgram(source.Length()?source:(*psit).m_program, (*psit).m_name);
		}
		
		if (program && program->isValid())
		{
			m->m_programs.push_back(program);
			program->Enable();
			activated = true;
		}
	}
	m->m_programsource.clear();
	return activated;
}

bool KX_PolygonMaterial::ActivateShader(RAS_IRasterizer *rasty) const
{
	if (!m->m_shadersource.size()) // No new shaders
	{
		if (m->m_shader)
		{
			if (m->m_shader->isValid()) // activate previous shader
			{
				m->m_shader->Enable();
				return true;
			}
		}
		return false;
	}
	
	// New shaders to load!
	delete m->m_shader;
	
	m->m_shader = rasty->NewShader();
	if (!m->m_shader)
	{
		m->m_shadersource.clear();
		return false; // GLSL not supported.
	}
	
	std::vector<STR_String> vshaders;
	std::vector<STR_String> fshaders;
	
	std::vector<Shader>::iterator shit = m->m_shadersource.begin();
	for (; shit != m->m_shadersource.end(); ++shit)
	{
		if ((*shit).m_shader.Length() > 24) // 24 == max length of Blender datablock; m_shader must be source
		{
			if ((*shit).m_target == RAS_IRasterizer::IShader::VERTEX)
				vshaders.push_back((*shit).m_shader);
			else
				fshaders.push_back((*shit).m_shader);
		}
		else
		{
			STR_String source = m_rendertools->GetText((*shit).m_shader);
			if (!source.Length()) // Not a datablock, must be source
				source = (*shit).m_shader;
			if ((*shit).m_target == RAS_IRasterizer::IShader::VERTEX)
				vshaders.push_back(source);
			else
				fshaders.push_back(source);
		}
			
	}
	
	if (vshaders.size())
		m->m_shader->LoadShader(RAS_IRasterizer::IShader::VERTEX, vshaders, m->m_shadersource[0].m_name);
	if (fshaders.size())
		m->m_shader->LoadShader(RAS_IRasterizer::IShader::FRAGMENT, fshaders, m->m_shadersource[0].m_name);
	
	m->m_shadersource.clear();
	for (std::vector<Attribute*>::iterator ait = m->m_attributes.begin(); ait != m->m_attributes.end(); ++ait)
		(*ait)->BindAttribute(m->m_shader);
	
	if (m->m_shader->LinkShader())
	{
		m->m_shader->Enable();
		return true;
	}
	
	m->m_shader->Disable();
	return false;
}

void KX_PolygonMaterial::ConvertMaterial(RAS_IRasterizer* rasty) const
{
	unsigned int t;
	unsigned int i;
	if (m->m_textures.size())
		return;
	// TODO
	for ( i = 0, t = 0; i < 8; i++)
	{
		MTex* mtex = m_material->mtex[i];
		if (!mtex)
			continue;
			
		if ((m_material->septex & (1<<i))) // Tex channel disabled
			continue;
		
		RAS_IRasterizer::ITexture *tex = m_rendertools->CreateTexture(rasty, mtex);
		if (tex)
			m->m_textures.push_back(tex);
	}
}

bool KX_PolygonMaterial::Activate(RAS_IRasterizer* rasty, TCachingInfo& cachingInfo) const 
{
	unsigned int i;
	unsigned int unit = 0;
	bool multipass = false;
	
	if (GetCachingInfo() != cachingInfo)
	{
		if (!cachingInfo)
			m_rendertools->SetTPage(NULL);
		
		cachingInfo = GetCachingInfo();

		if (m_material)
			ConvertMaterial(rasty);
			
		if (m_pymaterial)
		{
			PyObject *ret = PyObject_CallMethod(m_pymaterial, "Activate", NULL);
			if (ret)
			{
				bool value = PyInt_AsLong(ret);
				Py_DECREF(ret);
				return value;
			}
		}
		
		if (IsTextured() && ((rasty->GetDrawingMode() == RAS_IRasterizer::KX_TEXTURED)
		                 ||  (IsTransparant() && rasty->GetDrawingMode() == RAS_IRasterizer::KX_SHADOWMAP)))
		{
			ActivateUVTexture(rasty, unit);
			ActivateShaderTexture(rasty, unit);
		}
		
		m->m_doshadows = false;
		if (IsLit() && rasty->GetDrawingMode() >= RAS_IRasterizer::KX_SHADED && rasty->GetDrawingMode() <= RAS_IRasterizer::KX_TEXTURED)
		{
			for (i = 0; i < rasty->Lighting()->GetNumLights() && rasty->Lighting()->IsEnabled(i) && unit < rasty->GetNumTextureUnits(); i++)
			{
				if (ActivateShadowMap(rasty, i, unit))
				{
					m->m_doshadows = !IsTransparant();
					unit++;
				}
			}
			m->m_lightpass = 0;
			for (m->m_numlights = 0; m->m_numlights < rasty->Lighting()->GetNumLights() && rasty->Lighting()->IsEnabled(m->m_numlights); m->m_numlights++);
		}
		
		DisableTextures(rasty, unit);
		
		if (rasty->GetDrawingMode() <= RAS_IRasterizer::KX_TEXTURED)
		{
			if (ActivateShader(rasty))
			{
				for (std::vector<Uniform*>::iterator uit = m->m_uniforms.begin(); uit != m->m_uniforms.end(); ++uit)
					(*uit)->Activate(rasty, m->m_shader);
			} else 
				ActivatePrograms(rasty);
		}
		
		if(m_drawingmode & RAS_IRasterizer::KX_TWOSIDE)
		{
			rasty->SetCullFace(false);
		}
		else
		{
			rasty->SetCullFace(true);
		}

		if (m_drawingmode & RAS_IRasterizer::KX_LINES) {
			rasty->SetLines(true);
		}
		else {
			rasty->SetLines(false);
		}
		rasty->SetSubsurf(m_subsurf, m_linearsubsurf);
	}
	
	ActivateMaterialProperties(rasty);

	return multipass;
}

void KX_PolygonMaterial::DisableTextures(RAS_IRasterizer *rasty, unsigned int unit) const
{
	for (; unit < rasty->GetNumTextureUnits(); unit++)
	{
		rasty->SetTexUnit(unit);
		rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_DISABLE);
	}
	rasty->SetTexUnit(0);
}

void KX_PolygonMaterial::ActivateMaterialProperties(RAS_IRasterizer *rasty) const
{
	rasty->SetSpecularity(m_specular[0],m_specular[1],m_specular[2],m_specularity);
	rasty->SetShinyness(m_shininess);
	if (m_material)
	{
		rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], m_material->alpha);
		rasty->SetAmbient(MT_Vector4(m_diffuse[0]*m_material->amb, m_diffuse[1]*m_material->amb,m_diffuse[2]*m_material->amb, 1.0));
		//rasty->SetEmit(MT_Vector4(m_diffuse[0], m_diffuse[1],m_diffuse[2], m_material->emit));
	} else {
		rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], 1.0);
	}
}

void KX_PolygonMaterial::ActivateUVTexture(RAS_IRasterizer *rasty, unsigned int &unit) const
{
	rasty->SetTexUnit(unit);
	update_realtime_texture(m_tface, rasty->GetTime());
	m_rendertools->SetTPage(0); // Clear blender's cache - we've overwritten it.
	rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
	m_rendertools->SetTPage(m_tface);
	unit++;
}

bool KX_PolygonMaterial::ActivateShadowMap(RAS_IRasterizer *rasty, unsigned int li, unsigned int unit) const
{
	RAS_LightObject* light = rasty->Lighting()->Light(li);
	if (light->m_shadow && light->m_shadowmap.size())
	{
		rasty->SetTexUnit(unit);
		rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_GEN);
		//rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
		light->m_shadowmap[0]->BindShadow(light);
		rasty->Lighting()->Enable(li);
		return true;
	}
	return false;
}

void KX_PolygonMaterial::ActivateShaderTexture(RAS_IRasterizer *rasty, unsigned int &unit) const
{
	if (m->m_shader && m->m_shader->isValid() || m->m_programs.size())
	{
		for (unsigned int i = 0; i < m->m_textures.size() && unit < rasty->GetNumTextureUnits(); i++)
		{
			rasty->SetTexUnit(unit);
			rasty->SetTexCoords(RAS_IRasterizer::RAS_TEXCO_UV1);
			m->m_textures[i]->Bind();
			unit++;
		}
	}
}

bool KX_PolygonMaterial::ActivateLightAmbientPass(RAS_IRasterizer *rasty) const
{
	const float AMBIENTMIX = 0.5;
	const float DIFFUSEMIX = 0.5;
	
	DisableTextures(rasty, 0);
	
	if (!m_material)
		return false;
		
	rasty->SetDiffuse(m_diffuse[0]*m_material->amb*DIFFUSEMIX, m_diffuse[1]*m_material->amb*DIFFUSEMIX, m_diffuse[2]*m_material->amb*DIFFUSEMIX, 1.0);
	rasty->SetSpecularity(0.0, 0.0, 0.0, 0.0);
	rasty->SetAmbient(MT_Vector4(m_diffuse[0]*m_material->amb*AMBIENTMIX, m_diffuse[1]*m_material->amb*AMBIENTMIX, m_diffuse[2]*m_material->amb*AMBIENTMIX, 1.0));
	return true;
}

bool KX_PolygonMaterial::ActivateLightDiffusePass(RAS_IRasterizer *rasty) const
{
	unsigned int i;
	bool dopass = false;
	// Do non shadow lamps
	if (m->m_lightpass == m->m_numlights)
	{
		for (i = 0; i < m->m_numlights; i++)
		{
			RAS_LightObject* light = rasty->Lighting()->Light(i);
			if (!(light->m_shadow && light->m_shadowmap.size() && !light->m_nodiffuse))
			{
				rasty->Lighting()->Enable(i);
				dopass = true;
			} else
				rasty->Lighting()->Disable(i);
		}
		m->m_lightpass++;
		DisableTextures(rasty, 0);
		return dopass;
	}
	
	// Finished.
	if (m->m_lightpass > m->m_numlights)
		return false;
		
	rasty->SetDiffuse(m_diffuse[0], m_diffuse[1],m_diffuse[2], 1.0);
	for (i = 0; i < m->m_numlights; i++)
	{
		if (m->m_lightpass == i)
		{
			if (!rasty->Lighting()->Light(i)->m_nodiffuse && ActivateShadowMap(rasty, i, 0))
			{
				dopass = true;
			}
			else
			{
				// Not a shadow light: ignore
				m->m_lightpass++;
				rasty->Lighting()->Disable(i);
			}
		} else {
			// Not the active light for this pass: ignore
			rasty->Lighting()->Disable(i);
		}
	}
	m->m_lightpass++;
	
	if (!dopass)
		return ActivateLightDiffusePass(rasty);

	return dopass;
}

bool KX_PolygonMaterial::ActivateLightSpecularPass(RAS_IRasterizer *rasty) const
{
	unsigned int i;
	bool dopass = false;
	
	if (m_specular.fuzzyZero())
		return false;
	
	if (!m->m_lightpass)
	{
		rasty->SetSpecularity(m_specular[0],m_specular[1],m_specular[2],m_specularity);
		rasty->SetShinyness(m_shininess);
		rasty->SetDiffuse(0.0, 0.0, 0.0, 0.0);
		rasty->SetAmbient(MT_Vector4(0.0, 0.0, 0.0, 0.0));
		rasty->Lighting()->EnableLighting();
	}
	
	// Do non shadow lamps
	if (m->m_lightpass == m->m_numlights)
	{
		for (i = 0; i < m->m_numlights; i++)
		{
			RAS_LightObject* light = rasty->Lighting()->Light(i);
			if (!(light->m_shadow && light->m_shadowmap.size() && !light->m_nospecular))
			{
				rasty->Lighting()->Enable(i);
				dopass = true;
			} else
				rasty->Lighting()->Disable(i);
		}
		m->m_lightpass++;
		DisableTextures(rasty, 0);
		return dopass;
	}
	
	// Finished.
	if (m->m_lightpass > m->m_numlights)
		return false;
		
	for (i = 0; i < m->m_numlights; i++)
	{
		if (m->m_lightpass == i)
		{
			if (!rasty->Lighting()->Light(i)->m_nospecular && ActivateShadowMap(rasty, i, 0))
			{
				dopass = true;
			}
			else
			{
				// Not a shadow light: ignore
				m->m_lightpass++;
				rasty->Lighting()->Disable(i);
			}
		} else {
			// Not the active light for this pass: ignore
			rasty->Lighting()->Disable(i);
		}
	}
	m->m_lightpass++;

	return dopass;
}


bool KX_PolygonMaterial::ActivateLightTexturePass(RAS_IRasterizer *rasty) const
{
	if (IsTextured() && (rasty->GetDrawingMode() == RAS_IRasterizer::KX_TEXTURED))
	{
		unsigned int unit = 0;
		rasty->Lighting()->DisableLighting();
		//ActivateMaterialProperties(rasty);
		ActivateUVTexture(rasty, unit);
		DisableTextures(rasty, 1);
		rasty->SetTexUnit(0);
		return true;
	}
	
	return false;
}

bool KX_PolygonMaterial::ActivateShaderPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	if (!pass)
		return true;

	m->m_shader->Disable();
	return false;
}

bool KX_PolygonMaterial::ActivateProgramPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	if (!pass)
		return true;
	
	std::vector<RAS_IRasterizer::IProgram*>::iterator pit;
	for (pit = m->m_programs.begin(); pit != m->m_programs.end(); ++pit)
		(*pit)->Disable();

	return false;
}

bool KX_PolygonMaterial::ActivateShadowPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	switch (m->m_lightstate)
	{
		case Private::AMBIENT:
			m->m_lightstate = Private::DIFFUSE;
			rasty->SetMeshCache(false);
			if (ActivateLightAmbientPass(rasty))
				return true;
		case Private::DIFFUSE:
			if (pass)
				rasty->Blender()->SetBlendMode(RAS_IRasterizer::IBlend::ADD);
			if (ActivateLightDiffusePass(rasty))
				return true;
			m->m_lightstate = Private::TEXTURE;
			m->m_lightpass = 0;
		case Private::TEXTURE:
			m->m_lightstate = Private::SPECULAR;
			if (ActivateLightTexturePass(rasty))
			{
				rasty->Blender()->SetBlendMode(RAS_IRasterizer::IBlend::MULTIPLY);
				return true;
			}
		case Private::SPECULAR:
			if (ActivateLightSpecularPass(rasty))
			{
				rasty->Blender()->SetBlendMode(RAS_IRasterizer::IBlend::ADD);
				return true;
			}
			m->m_lightstate = Private::FINISHED;
			m->m_lightpass = 0;
		case Private::FINISHED:
			rasty->Blender()->SetBlendMode(RAS_IRasterizer::IBlend::DISABLE);
			m->m_lightstate = Private::AMBIENT;
			rasty->SetMeshCache(true);
			return false;
	}
	// FAILSAFE!
	std::cout << "ActivateShadowPass: m_lightstate is wrong!" << std::endl;
	m->m_lightstate = Private::AMBIENT;
	return false;
}

bool KX_PolygonMaterial::ActivateNextPass(RAS_IRasterizer *rasty, unsigned int pass) const
{
	if (m_pymaterial)
	{
		PyObject *ret = PyObject_CallMethod(m_pymaterial, "ActivateNextPass", "i", pass);
		if (ret)
		{
			bool value = PyInt_AsLong(ret);
			Py_DECREF(ret);
			
			if (!value)
			{
				ret = PyObject_CallMethod(m_pymaterial, "Deactivate", NULL);
				if (ret)
					Py_DECREF(ret);
				return false;
			}
			return value;
		}
	}

	if (m->m_shader)
		return ActivateShaderPass(rasty, pass);
		
	if (m->m_programs.size())
		return ActivateProgramPass(rasty, pass);
	
	if (m->m_doshadows && IsLit() && rasty->GetDrawingMode() >= RAS_IRasterizer::KX_SHADED && rasty->GetDrawingMode() <= RAS_IRasterizer::KX_TEXTURED)
		return ActivateShadowPass(rasty, pass);
		
	return !pass;
}

void KX_PolygonMaterial::ActivateMeshSlot(RAS_IRasterizer* rasty, const KX_MeshSlot &ms) const 
{ 
	if (m_pymaterial)
	{
		PyObject *ret = PyObject_CallMethod(m_pymaterial, "ActivateMeshSlot", NULL);
		if (ret)
		{
			Py_DECREF(ret);
			return;
		}
	}

	if (m->m_shader && m->m_shader->isValid())
	{
		for (std::vector<Uniform*>::iterator uit = m->m_uniforms.begin(); uit != m->m_uniforms.end(); ++uit)
			(*uit)->Update(ms, rasty, m->m_shader);
	}
}

//----------------------------------------------------------------------------
//Python


PyMethodDef KX_PolygonMaterial::Methods[] = {
	KX_PYMETHODTABLE(KX_PolygonMaterial, loadProgram),
	KX_PYMETHODTABLE(KX_PolygonMaterial, loadShader),
	KX_PYMETHODTABLE(KX_PolygonMaterial, genUniform),
	KX_PYMETHODTABLE(KX_PolygonMaterial, setUniform),
	KX_PYMETHODTABLE(KX_PolygonMaterial, bindAttribute),
	KX_PYMETHODTABLE(KX_PolygonMaterial, setCustomMaterial),
//	KX_PYMETHODTABLE(KX_PolygonMaterial, setPerPixelLights),
	
	{NULL,NULL} //Sentinel
};


PyTypeObject KX_PolygonMaterial::Type = {
	PyObject_HEAD_INIT(&PyType_Type)
		0,
		"KX_PolygonMaterial",
		sizeof(KX_PolygonMaterial),
		0,
		PyDestructor,
		0,
		__getattr,
		__setattr,
		0, //&MyPyCompare,
		__repr,
		0 //&cvalue_as_number,
};

PyParentObject KX_PolygonMaterial::Parents[] = {
	&KX_PolygonMaterial::Type,
	NULL
};

PyObject* KX_PolygonMaterial::_getattr(const STR_String& attr)
{
	if (attr == "VERTEX_SHADER")
			return PyLong_FromLong(RAS_IRasterizer::IShader::VERTEX);
	if (attr == "FRAGMENT_SHADER")
			return PyLong_FromLong(RAS_IRasterizer::IShader::FRAGMENT);
	
	if (attr == "texture")
		return PyString_FromString(m_texturename.ReadPtr());
	if (attr == "material")
		return PyString_FromString(m_materialname.ReadPtr());
	
	if (attr == "tile")
		return PyInt_FromLong(m_tile);
	if (attr == "tilexrep")
		return PyInt_FromLong(m_tilexrep);
	if (attr == "tileyrep")
		return PyInt_FromLong(m_tileyrep);
	
	if (attr == "drawingmode")
		return PyInt_FromLong(m_drawingmode);
	if (attr == "smooth")
		return PyInt_FromLong(m_smooth);
	if (attr == "transparent")
		return PyInt_FromLong(m_transparant);
	if (attr == "zsort")
		return PyInt_FromLong(m_zsort);
	if (attr == "lightlayer")
		return PyInt_FromLong(m_lightlayer);
	if (attr == "triangle")
		return PyInt_FromLong(m_bIsTriangle);
	if (attr == "subsurf")
		return PyInt_FromLong(m_subsurf);
	if (attr == "linearsubsurf")
		return PyInt_FromLong(m_linearsubsurf);
		
	if (attr == "diffuse")
		return PyObjectFrom(m_diffuse);
	if (attr == "shininess")
		return PyFloat_FromDouble(m_shininess);
	if (attr == "specular")
		return PyObjectFrom(m_specular);
	if (attr == "specularity")
		return PyFloat_FromDouble(m_specularity);
	
	_getattr_up(PyObjectPlus);
}

int KX_PolygonMaterial::_setattr(const STR_String &attr, PyObject *pyvalue)
{
	return PyObjectPlus::_setattr(attr, pyvalue);
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, setCustomMaterial, "setCustomMaterial(material)")
{
	PyObject *material;
	if (PyArg_ParseTuple(args, "O", &material))
	{
		if (m_pymaterial)
			Py_DECREF(m_pymaterial);

		m_pymaterial = material;
		Py_INCREF(m_pymaterial);
		Py_Return;
	}
	
	return NULL;
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, loadProgram, "loadProgram(source, name)")
{
	char *source;
	char *name = NULL;
	if (PyArg_ParseTuple(args, "s|s", &source, &name))
	{
		m->m_programsource.push_back(Program(source, name?name:"untitled program"));
		Py_Return;
	}
	
	return NULL;
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, loadShader, "loadShader(type, source, name)")
{
	char *source;
	char *name = NULL;
	int target;
	if (PyArg_ParseTuple(args, "is|s", &target, &source, &name))
	{
		switch (target)
		{
			case RAS_IRasterizer::IShader::VERTEX:
				m->m_shadersource.push_back(Shader(RAS_IRasterizer::IShader::VERTEX, source, name?name:"untitled vertex shader"));
				Py_Return;
			
			case RAS_IRasterizer::IShader::FRAGMENT:
				m->m_shadersource.push_back(Shader(RAS_IRasterizer::IShader::FRAGMENT, source, name?name:"untitled fragment shader"));
				Py_Return;
		}
	}
	return NULL;
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, bindAttribute, "bindAttribute(name, location)")
{
	char *name;
	int  location;
	if (PyArg_ParseTuple(args, "si", &name, &location))
	{
		m->m_attributes.push_back(new Attribute(name, location));
		Py_Return;
	}
	return NULL;
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, genUniform, "genUniform(name, value)")
{
	char *name;
	int value;
	if (PyArg_ParseTuple(args, "si", &name, &value))
	{
		m->m_uniforms.push_back(new GenUniform(name, (GenUniform::GenType) value));
		Py_Return;
	}
	
	return NULL;
}

KX_PYMETHODDEF_DOC(KX_PolygonMaterial, setUniform, "setUniform(name, value, transpose = false)")
{
	char *name;
	PyObject* value;
	int transpose = 0;
	if (PyArg_ParseTuple(args, "sO|i", &name, &value, &transpose))
	{
		if (PySequence_Check(value))
		{
			switch (PySequence_Size(value))
			{
				case 2:
				{
					MT_Vector2 vec;
					if (PyVecTo(value, vec))
					{
						m->m_uniforms.push_back(new VecUniform<MT_Vector2>(name, vec));
						Py_Return;
					}
					PyErr_SetString(PyExc_TypeError, "Unable to convert vec2");
					break;
				}
				case 3:
					if (PyObject_IsMT_Matrix(value, 3))
					{
						MT_Matrix3x3 mat;
						if (PyMatTo(value, mat))
						{
							m->m_uniforms.push_back(new MatUniform<MT_Matrix3x3>(name, mat, transpose));
							Py_Return;
						}
						PyErr_SetString(PyExc_TypeError, "Unable to convert mat3");
					}
					else
					{
						MT_Vector3 vec;
						if (PyVecTo(value, vec))
						{
							m->m_uniforms.push_back(new VecUniform<MT_Vector3>(name, vec));
							Py_Return;
						}
						PyErr_SetString(PyExc_TypeError, "Unable to convert vec3");
					}
					break;
				case 4:
					if (PyObject_IsMT_Matrix(value, 4))
					{
						MT_Matrix4x4 mat;
						if (PyMatTo(value, mat))
						{
							m->m_uniforms.push_back(new MatUniform<MT_Matrix4x4>(name, mat, transpose));
							Py_Return;
						}
						PyErr_SetString(PyExc_TypeError, "Unable to convert mat4");
					}
					else
					{
						MT_Vector4 vec;
						if (PyVecTo(value, vec))
						{
							m->m_uniforms.push_back(new VecUniform<MT_Vector4>(name, vec));
							Py_Return;
						}
						PyErr_SetString(PyExc_TypeError, "Unable to convert vec4");
					}
					break;
				default:
					PyErr_SetString(PyExc_TypeError, "Bad sequence length");
					break;
			}
		} else {
			if (PyInt_Check(value))
			{
				m->m_uniforms.push_back(new VecUniform<unsigned int>(name, PyInt_AsLong(value)));
				Py_Return;
			}
			else if (PyFloat_Check(value))
			{
				m->m_uniforms.push_back(new VecUniform<float>(name, PyFloat_AsDouble(value)));
				Py_Return;
			}
			PyErr_SetString(PyExc_TypeError, "Unable to convert type");
		}
	}
	return NULL;
}


