/**
 * $Id: Shading.fs 780 2004-08-05 01:58:38Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 *
 * Shading Functions.
 *
 * Diffuse Shaders:
 *   Lambertian
 *   Oren-Nayar
 *   Toon
 *
 * Specular Shaders:
 *   Phong
 *   Blinn
 *   Toon
 *
 * TARGET: Fragment Processor (Any)
 */

/**
 *  Lambertian Diffuse shader.
 *  ref = Reflection factor.
 *  FIXME: check this is sane
 */
vec3 BL_LambertianDiffuse(const vec3 normal, const vec3 lightvec, const float ref, const vec3 diffuse_col)
{
	return diffuse_col * ref * max(dot(normal, lightvec), 0.0);
}

/**
 *  Phong Specular shader
 *
 *  hard = hardness (specular exponent)
 *  spec = specular value
 */
vec3 BL_PhongSpecular(const vec3 normal, const vec3 lightvec, const vec3 viewvec, const float hard, const float spec, const vec3 speccol)
{
	vec3 refl = reflect(normal, lightvec);
	return spec * speccol * pow(dot(refl, viewvec), hard);
}

/**
 *  Oren-Nayar Diffuse shader.
 *
 *  ref = reflection coefficient.
 *  rough = roughness.
 */
vec3 BL_OrenNayarDiffuse(const vec3 normal, const vec3 lightvec, const vec3 viewvec, const float ref, const float rough, const vec3 diffuse_col)
{
	//vec3 h = normalize(lightvec + viewvec);
	//float nh = max(dot(normal, h), 0.0);
	float nv = max(dot(normal, viewvec), 0.0);
	float nl = max(dot(normal, lightvec), 0.0);
	//float vh = max(dot(viewvec, h), 0.0);
	float Lit_A = acos(nl);
	float View_A = acos(nv);
	
	vec3 Lit_B = normalize(lightvec - nl*normal);
	vec3 View_B = normalize(viewvec - nv*normal);
	
	float t = max(dot(Lit_B, View_B), 0.0);
	float a = max(Lit_A, View_A);
	float b = min(Lit_A, View_A)*0.95;
	
	float A = 1.0 - (0.5 * ((rough * rough) / ((rough * rough) + 0.33)));
	float B = 0.45 * ((rough * rough) / ((rough * rough) + 0.09));
	
	return diffuse_col * ref * nl * ( A + ( B * t * sin(a) * tan(b) ) );
}

/**
 *  Blinn Specular shader
 *
 *  refrac = coefficient of refraction
 *  hard = specular exponent
 *  spec = specular coefficient
 */
vec3 BL_BlinnSpecular(const vec3 normal, const vec3 lightvec, const vec3 viewvec, const float refrac, const float hard, const float spec, const vec3 speccol)
{
	vec3 h = normalize(lightvec + viewvec);
	float nh = max(dot(normal, h), 0.0);
	float nv = max(dot(normal, viewvec), 0.0);
	float nl = max(dot(normal, lightvec), 0.0);
	float vh = max(dot(viewvec, h), 0.01);
	
	float a = 1.0;
	float b = (2.0*nh*nv)/vh;
	float c = (2.0*nh*nl)/vh;

	float g = min(a, min(b, c));
	
	float p = sqrt( refrac*refrac + vh*vh - 1.0);
	float f = (((p-vh)*(p-vh))/((p+vh)*(p+vh)))*(1+((((vh*(p+vh))-1.0)*((vh*(p+vh))-1.0))/(((vh*(p-vh))+1.0)*((vh*(p-vh))+1.0))));
	float ang = acos(nh);

	return speccol * spec * f * g * exp((-(ang*ang) / (2.0*hard*hard)));
}

/**
 *  Toon Diffuse shader.
 *
 *  size = size of diffuse shading region (0.0 < size < 1.0)
 *  smooth = size of smooth diffuse shading region (0.0 < smooth < 1.0 - size)
 */
vec3 BL_ToonDiffuse(const vec3 normal, const vec3 lightvec, const vec3 viewvec, const float size, const float smooth, const vec3 diffuse_col)
{
	float ang = acos(dot(normal, lightvec));

	return diffuse_col*(1.0 - smoothstep(size, size + smooth, ang));
}

/**
 *  Toon Specular Shader.
 *
 *  size = size of specular region (0.0 < size < 1.0)
 *  smooth = size of smooth specular shading region (0.0 < smooth < 1.0 - size)
 */
vec3 BL_ToonSpecular(const vec3 normal, const vec3 lightvec, const vec3 viewvec, const float size, const float smooth, const vec3 speccol)
{
	vec3 h = normalize(l + v);
	float ang = acos(dot(h, normal));

	return speccol*(1.0 - smoothstep(size, size + smooth, ang));
}

