# $Id: KX_PolygonMaterial.py 837 2004-09-11 01:25:13Z kester $

class KX_PolygonMaterial:
	"""
	This is the interface to materials in Blender.
	
	Materials define the render state to be applied to mesh objects.
	
	B{Example}:
	
	This example does perpixel normal mapping.  It needs a textured mesh,
	and a normal map in the first active texture slot.
	
	B{vertex.vs}::
		uniform mat4 MVI; // Inverse ModelView Matrix
		
		attribute vec4 vertex_tangent;
		
		varying vec3 lightvec;
		varying vec3 viewvec;
		
		vec3 BL_TangentTransform(const vec3 vector, const vec3 tan1, const vec3 tan2, const vec3 origin)
		{
			return vec3(dot(tan1, vector),
				dot(tan2, vector),
				dot(gl_Normal, vector)) - origin;
		}
		
		void main ()
		{
			// Transform vertex to clip space
			gl_Position = gl_ModelViewProjectionMatrix*gl_Vertex;	
			
			// Create Tangent Space Matrix
			vec3 tan2 =  cross(vertex_tangent.xyz, gl_Normal.xyz)*vertex_tangent.w;
			
			vec4 lightpos = MVI*gl_LightSource[0].position;
			lightvec = BL_TangentTransform(lightpos, vertex_tangent.xyz, tan2, gl_Vertex);
			viewvec = BL_TangentTransform(MVI[3], vertex_tangent.xyz, tan2, gl_Vertex);
			
			// UV Coordinates
			gl_TexCoord[0] = gl_MultiTexCoord0;
			// Vertex colours
			gl_FrontColor = gl_Color;
		}

	B{fragment.fs}::
		uniform sampler2D colourmap;
		uniform sampler2D normap;
		
		varying vec3 lightvec;
		varying vec3 viewvec;
		
		void main()
		{
			// Tangent space light vector
			vec3 lv = normalize(lightvec);
		
			// Tangent space view vector
			vec3 vv = normalize(viewvec);
		
			// Lookup normal
			// Scale & bias normal
			// Normalise normal
			vec3 normal = normalize(2.0*texture2D(normap, gl_TexCoord[0].xy).xyz - 1.0);
		
			// Compute diffuse lighting
			float diffuse = dot(normal, lv);
			vec3 refl = normalize(2.0*diffuse*normal - lv);
		
			// Compute specular lighting
			float spec = pow(max(dot(refl, vv), 0.0), gl_FrontMaterial.shininess);
		
			// Lookup colour map
			vec4 colour = texture2D(colourmap, gl_TexCoord[0].xy) * gl_LightSource[0].diffuse * diffuse;
		
			// light colour map
			gl_FragColor = spec * gl_FrontMaterial.specular * gl_LightSource[0].specular + colour;
			// Pass alpha
			gl_FragColor.a = colour.a;
		}
	
	B{LoadShaders.py}::
		import GameLogic
		
		mats = GameLogic.getCurrentController().getOwner().getMesh().materials
		shaders = [(mats[0].VERTEX_SHADER, "vertex.vs"),
			   (mats[0].FRAGMENT_SHADER, "fragment.fs")]
		
		for mat in mats:
			for shader in shaders:
				mat.loadShader(shader[0], shader[1], shader[1])
			
			# Set Textures...
			mat.setUniform("colourmap", 0)
			mat.setUniform("normap", 1)
			# Set MV inverse matrix
			mat.genUniform("MVI", 2)
			# Bind tangent space
			mat.bindAttribute("vertex_tangent", 1)

	@bug: All attributes are read only.
	
	@group Constants: VERTEX_SHADER, FRAGMENT_SHADER
	@ivar VERTEX_SHADER: Vertex shader (see loadShader)
	@ivar FRAGMENT_SHADER: Fragment shader (see loadShader)
	
	@ivar texture: Texture name
	@type texture: string
	
	@ivar material: Material name
	@type material: string
	
	@ivar tile: Texture is tiling
	@type tile: boolean
	@ivar tilexrep: Number of tile repetitions in x direction.
	@type tilexrep: integer
	@ivar tileyrep: Number of tile repetitions in y direction.
	@type tileyrep: integer
	
	@ivar drawingmode: Drawing mode for the material.
		- 2  (drawingmode & 4)     Textured
		- 4  (drawingmode & 16)    Light
		- 14 (drawingmode & 16384) 3d Polygon Text
	@type drawingmode: bitfield
	
	@ivar smooth: See "Set Smooth/Set Solid"
		If smooth is false (Solid) and the mesh has armatures,
		the face normals will be recalculated.  This is slower.
	@type smooth: boolean
	
	@ivar transparent: This material is transparent.  All meshes with this
		material will be rendered after non transparent meshes from back
		to front.
	@type transparent: boolean
	
	@ivar zsort: Transparent polygons in meshes with this material will be sorted back to
		front before rendering.
		Non-Transparent polygons will be sorted front to back before rendering.
	@type zsort: boolean
	
	@ivar lightlayer: Light layers this material affects.
	@type lightlayer: bitfield.
	
	@ivar triangle: Mesh data with this material is triangles.
	@type triangle: boolean
	
	@ivar subsurf: The amount of mesh smoothing to apply.  This requires hardware support.
	@type subsurf: integer
	@ivar linearsubsurf: True if the subdivision method does not alter the shape of the mesh.
		linearsubsurf is used to increase lighting quality.
		
	@ivar diffuse: The diffuse colour of the material.  black = [0.0, 0.0, 0.0, 1.0] white = [1.0, 1.0, 1.0, 1.0]
	@type diffuse: list [r, g, b, a]
	@ivar specular: The specular colour of the material. black = [0.0, 0.0, 0.0, 1.0] white = [1.0, 1.0, 1.0, 1.0]
	@type specular: list [r, g, b, a] 
	@ivar shininess: The shininess (specular exponent) of the material. 0.0 <= shininess <= 128.0
	@type shininess: float
	@ivar specularity: The amount of specular of the material. 0.0 <= specularity <= 1.0
	@type specularity: float
	"""
	def loadProgram(source, name):
		"""
		Loads an assembly program into the material.
		
		The type of program will be deduced from the source header:
			1. !!ARBvp1.0	Vertex Program (GL_ARB_vertex_program)
			2. !!ARBfp1.0	Fragment Program (GL_ARB_fragment_program)
		
		@param source: The program to load.  Can be either the source itself, or the name
			of a text space.
		@type source: string
		@param name: (Optional) The name of the program.  The name is printed to the console for debugging.
		@type name: string
		"""
	def loadShader(shadertype, source, name):
		"""
		Loads a GLSL shader into the material.
		
		@param shadertype: VERTEX_SHADER or FRAGMENT_SHADER
		@param source: The shader to load.  Can be either the source itself, or the name
			of a text space.
		@type source: string
		@param name: (Optional) The name of the shader.  The name is printed to the console for debugging.
		@type name: string
		"""
	def setUniform(name, value, transpose = False):
		"""
		Sets a shader uniform parameter.
		
		You can pass parameters to your shader with this method.
		@param name: The name of the uniform to load.
		@type name: string
		@param value: The value of the uniform to load.
		@type value: integer, float, vec2, vec3, vec4, mat3, mat4
		@param transpose: (Optional) For matrix types, whether to transpose the matrix on load.
		@type transpose: boolean
		"""
	def genUniform(name, value):
		"""
		Generates a shader uniform parameter.
		
			0.	Model-View Matrix
			1.	Model-View Matrix Transposed
			2.	Inverse Model-View Matrix
			3.	Inverse Model-View Transposed
			4.	Model Matrix
			5.	Model Matrix Transposed
			6.	Inverse Model Matrix
			7.	Inverse Model Matrix Transposed
			8.	View Matrix
			9.	View Matrix Transposed
			10.	Inverse View Matrix
			11.	Inverse View Matrix Transposed
			12.	(unsigned int) Number of Lights (Unimplemented)
			13.	(unsigned int) Current eye in stereo mode. (0 = Left, 1 = Right)
			
		@bug: These constants should be given names.
		
		@param name: The name of the uniform to load.
		@type name: string
		@param value: The value to load into the uniform. (See list above)
		@type value: integer
		"""
	def bindAttribute(name, location):
		"""
		Binds an attribute location to a name.
		
		Available attributes:
			1. Vertex Tangent.  Bitangent = (normal X tangent)*tangent.w
		
		@param name: attribute to bind.
		@type name: string
		@param location: attribute location.
		@type location: integer
		"""
	def setCustomMaterial(material):
		"""
		Sets the material state setup object.
		
		Example::
			class PyMaterial:
				def Activate(self):
					# Activate is called first.
					# It returns True if this material is multipass
					return False
				def ActivateNextPass(self, p):
					# Called before each pass.
					# Returns True if this pass should be rendered.
					return p == 0 # Single pass
				def ActivateMeshSlot(self):
					# Called for each mesh to be rendered.
					pass
				def Deactivate(self):
					# Called to finish rendering.
					pass
			
			# Create a new Python Material and pass it to the renderer.
			mat.setCustomMaterial(PyMaterial())
		
		@param material: The material object.
		@type material: instance
		"""
		
