# $Id: Rasterizer.py 838 2004-09-11 01:32:13Z kester $
"""
Documentation for the Rasterizer module.

Example Uses an L{SCA_MouseSensor}, and two L{KX_ObjectActuator}s to implement MouseLook::
	# To use a mouse movement sensor "Mouse" and a 
	# motion actuator to mouse look:
	import Rasterizer
	import GameLogic

	# SCALE sets the speed of motion
	SCALE=[1, 0.5]
	
	co = GameLogic.getCurrentController()
	obj = co.getOwner()
	mouse = co.getSensor("Mouse")
	lmotion = co.getActuator("LMove")
	wmotion = co.getActuator("WMove")
	
	# Transform the mouse coordinates to see how far the mouse has moved.
	def mousePos():
		x = (Rasterizer.getWindowWidth()/2 - mouse.getXPosition())*SCALE[0]
		y = (Rasterizer.getWindowHeight()/2 - mouse.getYPosition())*SCALE[1]
		return (x, y)
	
	pos = mousePos()
	
	# Set the amount of motion: X is applied in world coordinates...
	lmotion.setTorque(0.0, 0.0, pos[0], False)
	# ...Y is applied in local coordinates
	wmotion.setTorque(-pos[1], 0.0, 0.0, True)
	
	# Activate both actuators
	GameLogic.addActiveActuator(lmotion, True)
	GameLogic.addActiveActuator(wmotion, True)
	
	# Centre the mouse
	Rasterizer.setMousePosition(Rasterizer.getWindowWidth()/2, Rasterizer.getWindowHeight()/2)

	
"""

def getWindowWidth():
	"""
	Gets the width of the window (in pixels)
	
	@rtype: integer
	"""
def getWindowHeight():
	"""
	Gets the height of the window (in pixels)
	
	@rtype: integer
	"""
def makeScreenshot(filename):
	"""
	Writes a screenshot to the given filename.
	
	If filename starts with // the image will be saved relative to the current directory.
	If the filename contains # it will be replaced with the frame number.
	
	The standalone player saves the file based on the extension. .jpg, .png and .tga are supported.
	It will also not do colour space conversion or gamma correction.
	
	When run from Blender, makeScreenshot supports Iris, IrisZ, TGA, Raw TGA, PNG, HamX, and Jpeg.
	Gamma, Colourspace conversion and Jpeg compression are taken from the Render settings panels.
	
	@type filename: string
	"""

def enableVisibility(visible):
	"""
	Doesn't really do anything...
	"""

def showMouse(visible):
	"""
	Enables or disables the operating system mouse cursor.
	
	@type visible: boolean
	"""

def setMousePosition(x, y):
	"""
	Sets the mouse cursor position.
	
	@type x, y: integer
	"""

def setBackgroundColor(rgba):
	"""
	Sets the window background colour.
	
	@type rgba: list [r, g, b, a]
	"""

def setMistColor(rgb):
	"""
	Sets the mist colour.
	
	@type rgb: list [r, g, b]
	"""

def setMistStart(start):
	"""
	Sets the mist start value.  Objects further away than start will have mist applied to them.
	
	@type start: float
	"""

def setMistEnd(end):
	"""
	Sets the mist end value.  Objects further away from this will be coloured solid with
	the colour set by setMistColor().
	
	@type end: float
	"""
	
def setDisplayResolution(width, height, bpp, rate):
	"""
	Sets the display resolution.
	
	The user's desktop resolution will be restored on Blender exit.
	
	Note: X11 platforms can't set bpp.
	
	Example::
		import Rasterizer
		
		Rasterizer.setDisplayResolution(800, 600, 32, 60)
	
	@param width: New display width, in pixels
	@type width: integer
	@param height: New display height, in pixels
	@type height: integer
	@param bpp: The colour depth to set the display to: valid values are 8, 16, 24 and 32.
	@type bpp: integer
	@param rate: The refresh rate to set the display
	@type rate: integer
	"""

def getCurrentResolution():
	"""
	Gets the current display resolution.
	
	Example::
		import Rasterizer
		
		currentres = Rasterizer.getCurrentResolution()
		print "Width: ", currentres[0]
		print "Height: ", currentres[1]
		print "BPP: ", currentres[2]
		print "Refresh Rate: ", currentres[3]
	
	@return: The current display resolution, in the form (width, height, bpp, rate)
	@rtype: tuple
	"""

def getValidResolutions():
	"""
	Gets a list of valid display resolutions.
	
	Example::
		import Rasterizer
		
		validres = Rasterizer.getValidResolutions()
		print "Width\\tHeight\\tBPP\\tRefresh Rate"
		for res in validres:
			print res[0], "\\t", res[1], "\\t", res[2], "\\t", res[3]
	
	@return: A list of valid display resolutions, in the form [(width, height, bpp, rate)]
	@rtype: list
	"""

def setEyeSeparation(eyesep):
	"""
	Sets the eye separation for stereo mode.
	
	@param eyesep: The distance between the left and right eye.
		If eyesep < 0.0, eye separation will be automatically determined from the projection
		matrix.
	@type eyesep: float
	"""

def getEyeSeparation():
	"""
	Gets the current eye separation for stereo mode.
	
	If the returned eye separation is < 0.0, the eye separation will be
	automatically set on the next frame.
	
	@rtype: float
	"""
	
def setFocalLength(focallength):
	"""
	Sets the focal length for stereo mode.
	
	@param focallength: The focal length.  If focallength < 0.0, the focal length will
		be automatically determined.
	@type focallength: float
	"""

def getFocalLength():
	"""
	Gets the current focal length for stereo mode.
	
	If the returned focal length is < 0.0, the focal length will be
	automatically set on the next frame.
	
	@rtype: float
	"""
