/**
 * $Id: RAS_IRasterizer.h 838 2004-09-11 01:32:13Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifndef __RAS_IRASTERIZER
#define __RAS_IRASTERIZER

#ifdef WIN32
#pragma warning (disable:4786)
#endif

#include "MT_CmMatrix4x4.h"
#include "MT_Matrix4x4.h"

#include "RAS_MaterialBucket.h"
#include "RAS_Rect.h"

class STR_String;

class RAS_ICanvas;
class RAS_IPolyMaterial;
struct RAS_LightObject;

/**
 *  Draw handle class.
 *
 *  This class is used by index primitives to speed up drawing.
 */
class RAS_IDrawHandle
{
protected:
	RAS_IDrawHandle() {}
public:
	virtual ~RAS_IDrawHandle() {}
	/**
	 *  Returns true if the object's vertex data has been modified, or if it has never been initialised.
	 */
	virtual bool IsVertexModified() = 0;
	/**
	 *  Sets that the buffered vertex data has changed, and should be reloaded.
	 */
	virtual void SetVertexModified(bool modified) = 0;
	/**
	 *  Returns true if the buffered index data has been modified, or if it has never been initialised.
	 */
	virtual bool IsIndexModified() = 0;
	/**
	 *  Sets that the buffered index data has changed, and should be reloaded.
	 */
	virtual void SetIndexModified(bool modified) = 0;
};	


/**
 * 3D rendering device context interface. 
 */
class RAS_IRasterizer
{

public:
	/**
	 *  Interface for lighting.
	 */
	class ILighting
	{
	protected:
		ILighting() {};
		
	public:
		virtual ~ILighting() {};
		/**
		 *  Returns the number of lights supported by the rasterizer.
		 */
		virtual unsigned int GetNumLights() = 0;
		
		/**
		 *  Enable the given light.
		 */
		virtual void Enable (unsigned int light, RAS_LightObject *data) = 0;
		virtual void Enable (unsigned int light) = 0;
		/**
		 *  Disable the given light.
		 */
		virtual void Disable (unsigned int light) = 0;
		
		/**
		 *  Gets light data for a given light.
		 */
		virtual RAS_LightObject* Light(unsigned int light) = 0;
		
		/**
		 *  Returns true if the given light is enabled.
		 */
		virtual bool IsEnabled(unsigned int light) = 0;
		
		/**
		 *  Begin lighting.
		 *
		 *  This sets up the world transform.
		 */
		virtual void BeginLighting() = 0;
		
		/**
		 *  End lighting.
		 *
		 *  Returns the world transform to its previous state.
		 */
		virtual void EndLighting() = 0;
		
		/**
		 *  Enable Lighting.
		 *
		 *  This sets lighting state to enable lighting.
		 */
		virtual void EnableLighting() = 0;
		
		/**
		 *  Disable Lighting.
		 *
		 *  This disables lighting state.
		 */
		virtual void DisableLighting() = 0;
		
	};
	
	/**
	 *  Interface for vertex & fragment programs.
	 */
	class IProgram
	{
	protected:
		IProgram() {};
	
	public:
		virtual ~IProgram() {}
		/**
		 *  Loads the given program.
		 *
		 *  Name is an optional name used in debug messages.
		 */
		virtual bool LoadProgram(const STR_String &program, const STR_String &name) = 0;
		
		/**
		 *  Enable enables the program for future rendered primitives.
		 */
		virtual void Enable() = 0;
		/**
		 *  Disable disables the program for future rendered primitives.
		 */
		virtual void Disable() = 0;
		/**
		 *  Returns true if the program loaded successfully.
		 */
		virtual bool isValid() = 0;
	};
	
	/**
	 *  Interface for vertex & fragment shaders.
	 */
	class IShader
	{
	protected:
		IShader() {};
	public:
		enum Target {
			VERTEX,
			FRAGMENT
		};
		
		virtual ~IShader() {}
		
		/**
		 *  Loads the shader into the given target.  The optional name is used for debug messages.
		 *
		 *  Returns true on success.
		 */
		//virtual bool LoadShader(Target target, const STR_String &shader, const STR_String &name = "untitled") = 0;
		virtual bool LoadShader(Target target, const std::vector<STR_String> &shaders, const STR_String &name = "untitled") = 0;
		/**
		 *  Links all shaders loaded with LoadShader.
		 *
		 *  Returns true on success.
		 */
		virtual bool LinkShader() = 0;
		/**
		 *  Enables this shader.
		 */
		virtual void Enable() = 0;
		/**
		 *  Disables this shader.
		 */
		virtual void Disable() = 0;
		/**
		 *  Returns true if this shader is valid.
		 *
		 *  A shader is valid if it links.
		 */
		virtual bool isValid() = 0;
		
		virtual int GetUniformLocation(const STR_String& name) = 0;
		virtual void SetUniform(int uniform, const MT_Tuple2& vec) = 0;
		virtual void SetUniform(int uniform, const MT_Tuple3& vec) = 0;
		virtual void SetUniform(int uniform, const MT_Tuple4& vec) = 0;
		virtual void SetUniform(int uniform, const unsigned int& val) = 0;
		virtual void SetUniform(int uniform, const float& val) = 0;
		virtual void SetUniform(int uniform, const MT_Matrix3x3& vec, bool transpose = false) = 0;
		virtual void SetUniform(int uniform, const MT_Matrix4x4& vec, bool transpose = false) = 0;

		virtual int GetAttribLocation(const STR_String& name) = 0;
		virtual void BindAttribute(const STR_String& attr, int loc) = 0;

/*		
		template<typename T>
		void SetUniform(const STR_String& uniform, const T& vec)
		{
			int loc = GetUniformLocation(uniform);
			if (loc >= 0)
				SetUniform(loc, vec);
		} */
	};

	/**
	 *  Interface for renderer textures.
	 */
	class ITexture
	{
	protected:
		ITexture() {}
		
	public:
		virtual ~ITexture() {}
		enum Filter
		{
			DISABLE,
			NEAREST,
			LINEAR
		};
		
		enum CubeFace
		{
			X_POS,
			X_NEG,
			Y_POS,
			Y_NEG,
			Z_POS,
			Z_NEG
		};
	
		/**
		 *  Sets the internal texture format.
		 */
		virtual void SetTextureFormat(bool luminance, bool alpha) = 0;
		virtual void SetShadowFormat(int size) = 0;
		virtual void SetCubeMap(CubeFace face) = 0;
		virtual ITexture *NewCubeMap() = 0;

		/**
		 *  Shares an image handle.  The renderer will use the given handle,
		 *  and not delete it when the texture object is deleted.
		 */
		virtual void ShareImage(unsigned int bind) = 0;
		/**
		 *  Sets this texture to be the active texture.
		 */
		virtual void Bind() = 0;
		/**
		 *  Binds the texture and sets up rendering for a shadow map.
		 */
		virtual void BindShadow(RAS_LightObject* light, int axis = 0) = 0;
		/**
		 *  Sets Tile mode.
		 */
		virtual void Tile(bool tile) = 0;
		/**
		 *  Enables texture compression.
		 *
		 *  Returns true if compression is enabled.
		 */
		virtual bool SetCompression(bool compress) = 0;

		/**
		 *  Sets the texture mip map filter.
		 *
		 *  Valid values for mipmap and linear: DISABLE, NEAREST, LINEAR
		 */		
		virtual void SetMipMap(Filter mipmap, Filter filter) = 0;

		/**
		 *  Copies the depth buffer rectangle specified by rect into the texture.
		 */
		virtual void ReadDepthBuffer(const RAS_Rect &rect) = 0;

		/**
		 *  Copies image data into the texture.
		 */
		virtual void BindImage(unsigned int *rect, int rectw, int recth) = 0;
		
	};
	
	/**
	 *  Blending class.
	 */
	class IBlend
	{
	protected:
		IBlend() {}
		
	public:
		virtual ~IBlend() {}
		enum BlendMode {
			DISABLE,
			ALPHA,
			ADD,
			SUBTRACT,
			MULTIPLY,
			REPLACE
		};
	
		/**
		 *  Sets the rasterizer blend mode.
		 *
		 *  See the above enum for valid values.
		 */
		virtual bool SetBlendMode(BlendMode blendmode) = 0;
	};
	
	RAS_IRasterizer(RAS_ICanvas* canv){};
	virtual ~RAS_IRasterizer(){};
	/**
	 */
	enum	{
			RAS_RENDER_3DPOLYGON_TEXT = 16384
	};
	/**
	 * Drawing types
	 */
	enum DrawType {
			KX_BOUNDINGBOX = 1,
			KX_WIREFRAME,
			KX_SOLID,
			KX_SHADED,
			KX_TEXTURED,
			KX_SHADOWMAP
	};

	/**
	 * Valid SetDepthMask parameters
	 */
	enum DepthMask {
			KX_DEPTHMASK_ENABLED =1,
			KX_DEPTHMASK_DISABLED
	};

	/**
	 */
	enum    { 	 
			KX_TWOSIDE = 512, 	 
			KX_LINES = 32768 	 
	};

	/**
	 * Stereo mode types
	 */
	enum StereoMode {
			RAS_STEREO_NOSTEREO = 1,
			RAS_STEREO_QUADBUFFERED,
			RAS_STEREO_ABOVEBELOW,
			RAS_STEREO_INTERLACED,
			RAS_STEREO_ANAGLYPH,
			RAS_STEREO_SIDEBYSIDE,
			RAS_STEREO_VINTERLACE,
			
			RAS_STEREO_MAXSTEREO
	};
	/**
	 * Render pass identifiers for stereo.
	 */
	enum StereoEye {
			RAS_STEREO_LEFTEYE = 1,
			RAS_STEREO_RIGHTEYE
	};
	
	/**
	 * Texture gen modes.
	 */
	enum TexCoGen {
		RAS_TEXCO_GEN,		//< GPU will generate texture coordinates
		RAS_TEXCO_ORCO,		//< Vertex coordinates (object space)
		RAS_TEXCO_GLOB,		//< Vertex coordinates (world space)
		RAS_TEXCO_UV1,		//< UV coordinates
		RAS_TEXCO_OBJECT,	//< Use another object's position as coordinates
		RAS_TEXCO_LAVECTOR,	//< Light vector as coordinates
		RAS_TEXCO_VIEW,		//< View vector as coordinates
		RAS_TEXCO_STICKY,	//< Sticky coordinates
		RAS_TEXCO_WINDOW,	//< Window coordinates
		RAS_TEXCO_DISABLE	//< Disable this texture unit (cached)
	};

	/**
	 * SetDepthMask enables or disables writing a fragment's depth value
	 * to the Z buffer.
	 */
	virtual void	SetDepthMask(DepthMask depthmask)=0;
	/**
	 * SetMaterial sets the material settings for subsequent primitives
	 * to be rendered with.
	 * The material will be cached.
	 */
	virtual bool	SetMaterial(const RAS_IPolyMaterial& mat)=0;
	/**
	 * Set the material for the next render pass.
	 */
	virtual bool	ActivateNextPass(const RAS_IPolyMaterial& mat)=0;
	
	/**
	 * Init initialises the renderer.
	 */
	virtual bool	Init()=0;
	/**
	 * Exit cleans up the renderer.
	 */
	virtual void	Exit()=0;
	/**
	 * BeginFrame is called at the start of each frame.
	 */
	virtual bool	BeginFrame(int drawingmode, double time)=0;
	virtual bool	BeginShadowMap(float bias = 1.0) = 0;
	/**
	 * ClearDepthBuffer clears the depth buffer.
	 */
	virtual void	ClearDepthBuffer()=0;
	/**
	 * ClearCachingInfo clears the currently cached material.
	 */
	virtual void	ClearCachingInfo(void)=0;
	/**
	 * EndFrame is called at the end of each frame.
	 */
	virtual void	EndFrame()=0;
	virtual void	EndShadowMap() = 0;
	/**
	 * SetRenderArea sets the render area from the 2d canvas
	 */
	virtual void	SetRenderArea()=0;

	// Stereo Functions
	/**
	 * SetStereoMode will set the stereo mode
	 */
	virtual void	SetStereoMode(const StereoMode stereomode)=0;
	/**
	 * Stereo can be used to query if the rasterizer is in stereo mode.
	 * @return true if stereo mode is enabled.
	 */
	virtual bool	Stereo()=0;
	/**
	 * Sets which eye buffer subsequent primitives will be rendered to.
	 */
	virtual void	SetEye(const StereoEye eye)=0;
	virtual StereoEye	GetEye()=0;
	/**
	 */
	virtual void	SetEyeSeparation(const float eyeseparation)=0;
	virtual float	GetEyeSeparation() = 0;
	/**
	 */
	virtual void	SetFocalLength(const float focallength)=0;
	virtual float	GetFocalLength() = 0;
	/**
	 * SwapBuffers swaps the back buffer with the front buffer.
	 */
	virtual void	SwapBuffers()=0;
	
	virtual void	SetMeshCache(bool cache) = 0;
	
	// Drawing Functions
	/**
	 * IndexPrimitives: Renders primitives.
	 * @param vertexarrays is an array of vertex arrays
	 * @param indexarrays is an array of index arrays
	 * @param mode determines the type of primitive stored in the vertex/index arrays:
	 *              0 triangles
	 *              1 lines (default)
	 *              2 quads
	 * @param polymat (reserved)
	 * @param useObjectColor will render the object using @param rgbacolor instead of 
	 *  vertex colours.
	 */
	virtual RAS_IDrawHandle*	IndexPrimitives( RAS_IDrawHandle* drawhandle,
							const vecVertexArray& vertexarrays,
							const vecIndexArrays & indexarrays,
							int mode,
							class RAS_IPolyMaterial* polymat,
							class RAS_IRenderTools* rendertools,
							bool useObjectColor,
							const MT_Vector4& rgbacolor)=0;
	/**
	 * @copydoc IndexPrimitives
	 * IndexPrimitivesEx will renormalize faces if @param vertexarrays[i].getFlag() & TV_CALCFACENORMAL
	 */
	virtual void	IndexPrimitives_Ex( const vecVertexArray& vertexarrays,
							const vecIndexArrays & indexarrays,
							int mode,
							class RAS_IPolyMaterial* polymat,
							class RAS_IRenderTools* rendertools,
							bool useObjectColor,
							const MT_Vector4& rgbacolor)=0;
	/**
	 * IndexPrimitives_3DText will render text into the polygons.
	 * The text to be rendered is from @param rendertools client object's text property.
	 */
	virtual void	IndexPrimitives_3DText( const vecVertexArray& vertexarrays,
							const vecIndexArrays & indexarrays,
							int mode,
							class RAS_IPolyMaterial* polymat,
							class RAS_IRenderTools* rendertools,
							bool useObjectColor,
							const MT_Vector4& rgbacolor)=0;

	virtual void	SetProjectionMatrix(MT_CmMatrix4x4 & mat)=0;
	/* This one should become our final version, methinks. */
	/**
	 * Set the projection matrix for the rasterizer. This projects
	 * from camera coordinates to window coordinates.
	 * @param mat The projection matrix.
	 */
	virtual void	SetProjectionMatrix(const MT_Matrix4x4 & mat)=0;
	/**
	 * Sets the modelview matrix.
	 */
	virtual void	SetViewMatrix(const MT_Matrix4x4 & mat,
						const MT_Vector3& campos,
						const MT_Point3 &camLoc,
						const MT_Quaternion &camOrientQuat)=0;
	virtual void	GetViewMatrix(MT_Matrix4x4 &mat) const = 0;
	/**
	 */
	virtual const	MT_Point3& GetCameraPosition()=0;
	/**
	 */
	virtual void	LoadViewMatrix()=0;
	/**
	 */
	virtual void	SetFog(float start,
						   float dist,
						   float r,
						   float g,
						   float b)=0;
	
	virtual void	SetFogColor(float r,
								float g,
								float b)=0;

	virtual void	SetFogStart(float start)=0;
	/**
	 */
	virtual void	SetFogEnd(float end)=0;
	/**
	 */
	virtual void	DisplayFog()=0;
	/**
	 */
	virtual void	DisableFog()=0;

	virtual void	SetBackColor(float red,
								 float green,
								 float blue,
								 float alpha)=0;
	
	/**
	 * @param drawingmode = KX_BOUNDINGBOX, KX_WIREFRAME, KX_SOLID, KX_SHADED or KX_TEXTURED.
	 */
	virtual void	SetDrawingMode(int drawingmode)=0;
	/**
	 * @return the current drawing mode: KX_BOUNDINGBOX, KX_WIREFRAME, KX_SOLID, KX_SHADED or KX_TEXTURED.
	 */
	virtual int	GetDrawingMode()=0;
	/**
	 */
	virtual void	EnableTextures(bool enable)=0;
	/**
	 * Sets face culling
	 */	
	virtual void	SetCullFace(bool enable)=0;
	/**
	 * Sets wireframe mode.
	 */
	virtual void    SetLines(bool enable)=0;
	/**
	 */
	virtual double	GetTime()=0;
	/**
	 * Generates a projection matrix from the specified frustum.
	 * @param left the left clipping plane
	 * @param right the right clipping plane
	 * @param bottom the bottom clipping plane
	 * @param top the top clipping plane
	 * @param frustnear the near clipping plane
	 * @param frustfar the far clipping plane
	 * @return a 4x4 matrix representing the projection transform.
	 */
	virtual MT_Matrix4x4 GetFrustumMatrix(
		float left,
		float right,
		float bottom,
		float top,
		float frustnear,
		float frustfar,
		bool perspective = true
	)=0;
	
	virtual MT_Matrix4x4 GetFrustumMatrix(
		float fov,
		float frustnear,
		float frustfar
	) = 0;
	/**
	 * Sets the specular colour component of the lighting equation.
	 */
	virtual void	SetSpecularity(float specX,
								   float specY,
								   float specZ,
								   float specval)=0;
	
	/**
	 * Sets the specular exponent component of the lighting equation.
	 */
	virtual void	SetShinyness(float shiny)=0;
	/**
	 * Sets the diffuse colour component of the lighting equation.
	 */
	virtual void	SetDiffuse(float difX,
							   float difY,
							   float difZ,
							   float diffuse)=0;
	/**
	 * Sets the ambient colour component of the lighting equation.
	 */
	virtual void	SetAmbient(const MT_Vector4 &amb)=0;
	/**
	 * Sets the emit colour component of the lighting equation.
	 */
	virtual void	SetEmit(const MT_Vector4 &emit)=0;

	/**
	 * Enables subsurf, if supported by hardware.
	 * @param level the number of evaluation points per edge (0 disables.)
	 * @param linear specifies linear subsurf
	 */
	virtual void	SetSubsurf(int, bool) {};
	
	virtual unsigned int	GetNumTextureUnits() = 0;
	virtual void	SetTexUnit(unsigned int unit) = 0;
	virtual void	SetTexCoords(TexCoGen coords) = 0;
	
	/**
	 *  Gets a handle to the lighting manager.
	 */
	virtual ILighting *Lighting() = 0;
	/**
	 *  Gets a handle to the blending state.
	 */
	virtual IBlend *Blender() = 0;
	/**
	 *  Creates a new vertex/fragment program.
	 */
	virtual IProgram* NewProgram(const STR_String &program, const STR_String &name = "unnamed") = 0;
	/**
	 *  Sets the number of per pixel lights to use.
	 */
	virtual void SetPerPixelLights(unsigned int lights) = 0;
	/**
	 *  Gets the number of per pixel lights to use.
	 */
	virtual unsigned int GetPerPixelLights() = 0;
	/**
	 *  Create a new texture object.
	 *
	 *  bind is an optional handle to share - the texture will use an already 
	 *  existing texture.
	 */
	virtual ITexture* NewTexture(int bind = 0) = 0;

	virtual bool QueryShadowMaps() = 0;
	/**
	 *  Create a new high level shader object.
	 *
	 *  Will return NULL if shaders are not available.
	 */
	virtual IShader *NewShader() = 0;
};

#endif //__RAS_IRASTERIZER

