/**
 * $Id: RAS_OpenGLRasterizer.cpp 838 2004-09-11 01:32:13Z kester $
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "GL/glew.h"

#include "RAS_OpenGLRasterizer.h"

#include "RAS_Rect.h"
#include "RAS_TexVert.h"
#include "MT_CmMatrix4x4.h"
#include "RAS_IRenderTools.h" // rendering text
#include "RAS_LightObject.h"

// Use the accumulation buffer for motion blur
//#define RAS_MOTIONBLUR
// Enable antialiasing
//#define RAS_MULTISAMPLE
// Enable texture anisotropic filter
//#define RAS_ANISO

/******************************************************************************
                       Blending Class
******************************************************************************/
class RAS_OpenGLRasterizer::GLBlend : public RAS_IRasterizer::IBlend
{
protected:
	BlendMode m_blendmode;
public:
	GLBlend() :
		m_blendmode((BlendMode) -1)
	{}
	
	virtual ~GLBlend()
	{}

	virtual bool SetBlendMode(BlendMode blendmode)
	{
		bool result = true;
		if (m_blendmode != blendmode)
		{
			m_blendmode = blendmode;
			
			if (m_blendmode)
			{
				glEnable(GL_BLEND);
				
				switch (m_blendmode)
				{
					case ALPHA:
						glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
						break;
					case ADD:
						glBlendFunc(GL_ONE, GL_ONE);
						break;
					case SUBTRACT:
						// Can't do this without extensions
						glBlendFunc(GL_ZERO, GL_ONE);
						result = false;
						break;
					case MULTIPLY:
						glBlendFunc(GL_ZERO, GL_SRC_COLOR);
						break;
				}
			}
			else
			{
				glDisable(GL_BLEND);
			}
		}
		return result;
	}
	
	virtual BlendMode GetBlendMode()
	{
		return m_blendmode;
	}
};

/**
 *  Extended blend class
 */
class RAS_OpenGLRasterizer::GLBlendEXT : public RAS_OpenGLRasterizer::GLBlend
{
public:
	GLBlendEXT()
		: GLBlend()
	{}
	
	virtual ~GLBlendEXT()
	{}
	
	virtual bool SetBlendMode(BlendMode blendmode)
	{
		if (m_blendmode != blendmode)
		{
			m_blendmode = blendmode;
			
			if (m_blendmode)
			{
				glEnable(GL_BLEND);
				
				switch (m_blendmode)
				{
					case ALPHA:
						glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
						glBlendEquationEXT(GL_FUNC_ADD);
						break;
					case ADD:
						glBlendFunc(GL_ONE, GL_ONE);
						glBlendEquationEXT(GL_FUNC_ADD);
						break;
					case SUBTRACT:
						glBlendFunc(GL_ONE, GL_ONE);
						glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT);
						break;
					case MULTIPLY:
						glBlendFunc(GL_ZERO, GL_SRC_COLOR);
						glBlendEquationEXT(GL_FUNC_ADD);
						break;
					case REPLACE:
						glBlendFunc(GL_ONE, GL_ZERO);
						glBlendEquationEXT(GL_FUNC_ADD);
						break;
					default:
						std::cout << "GLBlendImaging::SetBlendMode: Crazy Blend Mode selected. " << m_blendmode << std::endl;
						break;
				}
			}
			else
			{
				glDisable(GL_BLEND);
			}
		}
		return true;
	}

};


/******************************************************************************
                       Texture Class
******************************************************************************/
class RAS_OpenGLRasterizer::GLTexture : public RAS_IRasterizer::ITexture
{
	GLuint	m_bind;
	bool	m_shared; // m_bind is shared with Blender (don't delete)
	
	GLint	m_format; // GL_RGBA, GL_LUMINANCE etc
	GLenum	m_target; // GL_TEXTURE_2D
	GLenum	m_subtarget; // GL_TEXTURE_2D (cube mapping)
	
	bool	m_doMipMap; // Mipmapping enabled
	GLint	m_mipMap;   // Mipmapping level: GL_LINEAR_MIPMAP_NEAREST etc
	
	bool	m_compress;
	
	static bool	m_gcompress;   // Texture compression enabled globally.
public:
	
	void SetTextureFormat(bool luminance, bool alpha)
	{
		if (luminance)
		{
			if (alpha)
				m_format = m_compress?GL_COMPRESSED_LUMINANCE_ALPHA_ARB:GL_LUMINANCE_ALPHA;
			else
				m_format = m_compress?GL_COMPRESSED_LUMINANCE_ARB:GL_LUMINANCE;
		} else {
			if (!alpha)
				m_format = m_compress?GL_COMPRESSED_RGB_ARB:GL_RGB;
			else
				m_format = m_compress?GL_COMPRESSED_RGBA_ARB:GL_RGBA;
		}
	}
	
	void SetShadowFormat(int size)
	{
		m_format = GL_DEPTH_COMPONENT;
		Bind();
		glTexImage2D(m_subtarget, 0, m_format, size, size, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, 0);
		glTexParameteri(m_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(m_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(m_target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
		glTexParameteri(m_target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
		float border[4] = {1.0, 0.0, 0.0, 0.0};
		glTexParameterfv(m_target, GL_TEXTURE_BORDER_COLOR, border);
		glTexParameteri(m_target, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB);
		glTexParameteri(m_target, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
		//glTexParameteri(m_target, GL_DEPTH_TEXTURE_MODE_ARB, GL_INTENSITY);
	}
	
	void SetCubeMap(CubeFace face)
	{
		m_target = GL_TEXTURE_CUBE_MAP;
		m_subtarget = GL_TEXTURE_CUBE_MAP_POSITIVE_X + face;
	}
	
	ITexture* NewCubeMap()
	{
		return new GLTexture(m_bind);
	}
	
	void ShareImage(unsigned int bind)
	{
		m_bind = (GLuint) bind;
		m_shared = true;
	}
	
	void Bind()
	{
		glBindTexture( m_target, m_bind);
	}
	
	void BindShadow(RAS_LightObject* light, int axis)
	{
		Bind();
		
		MT_Matrix4x4 eyeplane(0.5f, 0.0f, 0.0f, 0.5f,
					0.0f, 0.5f, 0.0f, 0.5f,
					0.0f, 0.0f, 0.5f, 0.5f,
					0.0f, 0.0f, 0.0f, 1.0f);
		eyeplane = eyeplane*light->m_projection_matrix * light->m_modelview_matrix; 
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGendv(GL_S, GL_EYE_PLANE, eyeplane[0].getValue());
		glTexGendv(GL_T, GL_EYE_PLANE, eyeplane[1].getValue());
		glTexGendv(GL_R, GL_EYE_PLANE, eyeplane[2].getValue());
		glTexGendv(GL_Q, GL_EYE_PLANE, eyeplane[3].getValue());
	}
	
	void DeleteTexture()
	{
		if (m_bind)
		{
			if (!m_shared)
				glDeleteTextures(1, &m_bind);
			m_bind = 0;
		}
	}
	
	void Tile(bool tile)
	{
		if (tile)
		{
			glTexParameteri(m_target, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(m_target, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(m_target, GL_TEXTURE_WRAP_R, GL_REPEAT);
		} else {
			float border[4] = {0.0, 0.0, 0.0, 0.0};
			glTexParameteri(m_target, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(m_target, GL_TEXTURE_WRAP_T, GL_CLAMP);
			glTexParameteri(m_target, GL_TEXTURE_WRAP_R, GL_CLAMP);
			glTexParameterfv(m_target, GL_TEXTURE_BORDER_COLOR, border);
		}
	}
	
	void ReadDepthBuffer(const RAS_Rect &rect)
	{
		Bind();
		glCopyTexImage2D(m_subtarget, 0, m_format, rect.GetLeft(), rect.GetBottom(), rect.GetWidth(), rect.GetHeight(), 0);
	}
	
	
	void BindImage(unsigned int *rect, int rectw, int recth)
	{
		switch (m_target)
		{
			case GL_TEXTURE_CUBE_MAP:
			case GL_TEXTURE_2D:
				if (!m_doMipMap)
					glTexImage2D(m_subtarget, 0,  m_format,  rectw, recth, 0, GL_RGBA, GL_UNSIGNED_BYTE, rect);
				else
					gluBuild2DMipmaps(m_subtarget, m_format, rectw, recth, GL_RGBA, GL_UNSIGNED_BYTE, rect);
				break;
			case GL_TEXTURE_1D:
				if (!m_doMipMap)
					glTexImage1D(m_subtarget, 0,  m_format,  rectw, 0, GL_RGBA, GL_UNSIGNED_BYTE, rect);
				else
					gluBuild1DMipmaps(m_subtarget, m_format, rectw, GL_RGBA, GL_UNSIGNED_BYTE, rect);
				break;
		}
	
		if (!m_doMipMap)
		{
			glTexParameteri(m_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(m_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		} else
		{
			glTexParameteri(m_target, GL_TEXTURE_MIN_FILTER, m_mipMap);
			glTexParameteri(m_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		}
	}
	
	static unsigned int GetNumTextureUnits()
	{
		GLint units = 1;
		if (GLEW_ARB_fragment_shader || GLEW_ARB_fragment_program)
			glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS_ARB, &units);
		else if (GLEW_ARB_multitexture || GLEW_VERSION_1_3)
			glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &units);
		return units;
	}
	
	static unsigned int GetNumTextureCoords()
	{
		GLint coords = 1;
		if (GLEW_ARB_fragment_shader || GLEW_ARB_fragment_program)
			glGetIntegerv(GL_MAX_TEXTURE_COORDS_ARB, &coords);
		else if (GLEW_ARB_multitexture || GLEW_VERSION_1_3)
			glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &coords);
		return coords;
	}
	
	static bool SetTextureCompression(bool compress)
	{
		m_gcompress = compress && GLEW_ARB_texture_compression;
		return m_gcompress;
	}
	
	bool SetCompression(bool compress)
	{
		m_compress = compress && m_gcompress;
		return m_compress;
	}
	
	void SetMipMap(Filter mipmap, Filter filter)
	{
		m_doMipMap = true;
		switch (mipmap)
		{
			case DISABLE:
				m_doMipMap = false;
				break;
			case LINEAR:
				switch (filter)
				{
					case NEAREST:
						m_mipMap = GL_LINEAR_MIPMAP_NEAREST;
						break;
					default:
					case LINEAR:
						m_mipMap = GL_LINEAR_MIPMAP_LINEAR;
						break;
				}
				break;
			case NEAREST:
				switch (filter)
				{
					case NEAREST:
						m_mipMap = GL_NEAREST_MIPMAP_NEAREST;
						break;
					default:
					case LINEAR:
						m_mipMap = GL_NEAREST_MIPMAP_LINEAR;
						break;
				}
				break;
		}
	}
	
	GLTexture() :
		m_bind(0),
		m_shared(false),
		m_format(GL_RGBA),
		m_target(GL_TEXTURE_2D),
		m_subtarget(GL_TEXTURE_2D),
		m_compress(m_gcompress)
	{
		glGenTextures(1, &m_bind);
		Bind();
	}
	
	GLTexture(int share) :
		m_bind(share),
		m_shared(true),
		m_format(GL_RGBA),
		m_target(GL_TEXTURE_2D),
		m_subtarget(GL_TEXTURE_2D),
		m_compress(m_gcompress)
	{
		Bind();
	}
		
	
	virtual ~GLTexture()
	{
		DeleteTexture();
	}
	
};

bool	RAS_OpenGLRasterizer::GLTexture::m_gcompress = true;
/******************************************************************************
                       Vertex/Fragment Program Class
******************************************************************************/
class RAS_OpenGLRasterizer::GLProgram : public RAS_IRasterizer::IProgram
{
	GLuint	m_bind;
	GLenum	m_target;
	bool	m_valid;
public:
	GLProgram()
		: IProgram(),
		m_bind(0),
		m_valid(false)
	{
	}
	
	GLProgram(const STR_String &program, const STR_String &name)
		: IProgram(),
		m_bind(0),
		m_valid(false)
	{
		LoadProgram(program, name);
	}
	
	virtual ~GLProgram()
	{
	}

	void PrintProgramError(const STR_String &program, const STR_String &name, int errorPos, const STR_String &ptype)
	{
		std::cout << "Error in " << ptype << " program \"" << name << "\" at " << errorPos << ":" << std::endl;
		std::cout << (const char *) glGetString(GL_PROGRAM_ERROR_STRING_ARB) << std::endl;
		std::vector<STR_String> lines = program.Explode('\n');
		int start = 0;
		for (std::vector<STR_String>::iterator line = lines.begin(); line != lines.end(); ++line)
		{
			int finish = start + (*line).Length() + 1;
			if (errorPos >= start && errorPos < finish)
			{
				std::cout << (*line) << std::endl;
				std::cout << STR_String(' ', errorPos - start) << "^" << std::endl;
				break;
			}
			start = finish;
		}
	}
	
	void PrintResourceUsage(const STR_String &name, const STR_String &ptype)
	{
		int count, native;
		int maxcount, maxnative;
		std::cout << ptype << " program \"" << name << "\" resource profile:" << std::endl;
		
		#define glGetProgramv(name) \
			glGetProgramivARB(m_target, GL_PROGRAM_##name##_ARB, (GLint*) &count); \
			glGetProgramivARB(m_target, GL_MAX_PROGRAM_##name##_ARB, (GLint*) &maxcount); \
			glGetProgramivARB(m_target, GL_PROGRAM_NATIVE_##name##_ARB, (GLint*) &native); \
			glGetProgramivARB(m_target, GL_MAX_PROGRAM_NATIVE_##name##_ARB, (GLint*) &maxnative) 
		
		#define output \
			count << "/" << maxcount << " (" << native << "/" << maxnative << ")"

		glGetProgramv(INSTRUCTIONS);
		std::cout << "Instructions (native): " << output << std::endl;
		if (m_target == GL_FRAGMENT_PROGRAM_ARB)
		{
			glGetProgramv(ALU_INSTRUCTIONS);
			std::cout << "ALU Instructions (native): " << output << std::endl;
			glGetProgramv(TEX_INSTRUCTIONS);
			std::cout << "TEX Instructions (native): " << output << std::endl;
			glGetProgramv(TEX_INDIRECTIONS);
			std::cout << "TEX Indirections (native): " << output << std::endl;
		}
		
		glGetProgramv(TEMPORARIES);
		std::cout << "TEMPs (native): " << output << std::endl;
		glGetProgramv(PARAMETERS);
		std::cout << "PARAMs (native): " << output << std::endl;
		glGetProgramv(ATTRIBS);
		std::cout << "ATTRIBs (native): " << output << std::endl;
		std::cout << std::endl;
		
		#undef glGetProgramv
		#undef output
	}
	
	static void PrintResources()
	{
		std::cout << "   Programs   : ";
		if (GLEW_ARB_vertex_program && GLEW_ARB_fragment_program)
			std::cout << "Vertex & Fragment programs supported" << std::endl;
		else if (GLEW_ARB_vertex_program)
			std::cout << "Vertex programs supported" << std::endl;
		else if (GLEW_ARB_fragment_program)
			std::cout << "Fragment programs supported" << std::endl;
		else
		{
			std::cout << "Not Supported" << std::endl;
			return;
		}
		
		GLint maxcount, maxnative;
		#define glGetProgramv(target, name) \
			glGetProgramivARB(GL_##target##_PROGRAM_ARB, GL_MAX_PROGRAM_##name##_ARB, (GLint*) &maxcount); \
			glGetProgramivARB(GL_##target##_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_##name##_ARB, (GLint*) &maxnative) 
		
		#define output \
			maxcount << " (" << maxnative << ")"
		
		if (GLEW_ARB_vertex_program)
		{
			glGetProgramv(VERTEX, INSTRUCTIONS);
			std::cout << "      Vertex Instructions (native): " << output << std::endl;
			glGetProgramv(VERTEX, TEMPORARIES);
			std::cout << "             TEMPs (native)       : " << output << std::endl;
			glGetProgramv(VERTEX, PARAMETERS);
			std::cout << "             PARAMs (native)      : " << output << std::endl;
			glGetProgramv(VERTEX, ATTRIBS);
			std::cout << "             ATTRIBs (native)     : " << output << std::endl;
		}
		
		if (GLEW_ARB_fragment_program)
		{
			glGetProgramv(FRAGMENT, INSTRUCTIONS);
			std::cout << "      Fragment Instructions (native): " << output << std::endl;
			glGetProgramv(FRAGMENT, ALU_INSTRUCTIONS);
			std::cout << "           ALU Instructions (native): " << output << std::endl;
			glGetProgramv(FRAGMENT, TEX_INSTRUCTIONS);
			std::cout << "           TEX Instructions (native): " << output << std::endl;
			glGetProgramv(FRAGMENT, TEX_INDIRECTIONS);
			std::cout << "           TEX Indirections (native): " << output << std::endl;
			glGetProgramv(FRAGMENT, TEMPORARIES);
			std::cout << "               TEMPs (native)       : " << output << std::endl;
			glGetProgramv(FRAGMENT, PARAMETERS);
			std::cout << "               PARAMs (native)      : " << output << std::endl;
			glGetProgramv(FRAGMENT, ATTRIBS);
			std::cout << "               ATTRIBs (native)     : " << output << std::endl;
		}
		#undef glGetProgramv
		#undef output
	}
	
	static GLenum QueryProgramType(const STR_String &program)
	{
		if (program.Find("!!ARBvp1.0\n") == 0)
			return GL_VERTEX_PROGRAM_ARB;
		else if (program.Find("!!ARBfp1.0\n") == 0)
			return GL_FRAGMENT_PROGRAM_ARB;
		return GL_INVALID_ENUM;
	}
	
	static bool QueryExtensions(const STR_String &program)
	{
		GLenum type = QueryProgramType(program);
		switch (type)
		{
			case GL_VERTEX_PROGRAM_ARB:
				return GLEW_ARB_vertex_program;
			case GL_FRAGMENT_PROGRAM_ARB:
				return GLEW_ARB_fragment_program;
		}
		
		return false;
	}

	virtual bool LoadProgram(const STR_String &program, const STR_String &name)
	{
		int errorPos, isNative;
		
		STR_String ptype("unknown");
		m_target = QueryProgramType(program);
		switch (m_target)
		{
			case GL_VERTEX_PROGRAM_ARB:
				ptype = "vertex";
				break;
			case GL_FRAGMENT_PROGRAM_ARB:
				ptype = "fragment";
				break;
			default:
				std::cout << "Invalid program header." << std::endl;
				return false;
		}
		
		if (m_target == GL_VERTEX_PROGRAM_ARB && !GLEW_ARB_vertex_program)
			return false;
			
		if (m_target == GL_FRAGMENT_PROGRAM_ARB && !GLEW_ARB_fragment_program)
			return false;
		
		glGenProgramsARB(1, &m_bind);
		glBindProgramARB(m_target, m_bind);
		glProgramStringARB(m_target, GL_PROGRAM_FORMAT_ASCII_ARB, program.Length(), program.ReadPtr());
		glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errorPos);
		glGetProgramivARB(m_target, GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB, &isNative);
		if (errorPos == -1)
		{
			if (isNative != 1)
			{
				std::cout << "Error: \"" << name << "\"" << ptype << " program running in software!\n" << std::endl;
				// Don't use software fragment programs, they're way too slow.
				if (m_target == GL_VERTEX_PROGRAM_ARB)
					m_valid = true;
			} else {
				m_valid = true;
			}
			PrintResourceUsage(name, ptype);
		} else {
			PrintProgramError(program, name, errorPos, ptype);
		}
		return m_valid;
	}
	
	virtual void Enable()
	{
		if (m_valid)
		{
			glDisable(GL_COLOR_MATERIAL);
			glEnable(m_target);
			glBindProgramARB(m_target, m_bind);
		}
	} 
	
	virtual void Disable()
	{
		glDisable(m_target);
		glEnable(GL_COLOR_MATERIAL);
	}
	
	virtual bool isValid()
	{
		return m_valid;
	}
};

class RAS_OpenGLRasterizer::GLSLShader : public RAS_IRasterizer::IShader
{
	GLhandleARB	m_program;
	GLint		m_valid;
	
	void PrintInfoLog(GLhandleARB object, const STR_String &stage, const STR_String &name)
	{
		int maxLength = 0;
		glGetObjectParameterivARB(object, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);
		
		char *infoLog = new char[maxLength];
		glGetInfoLogARB(object, maxLength, &maxLength, infoLog);
		
		if (maxLength > 0)
		{
			std::cout << stage << name << ":" << std::endl;
			std::cout << infoLog << std::endl;
		}
		
		delete [] infoLog;
	}
protected:
public:
	GLSLShader() 
		: IShader(),
		  m_program(glCreateProgramObjectARB()),
		  m_valid(0)
	{
		
	}
	
	virtual ~GLSLShader() 
	{
		glDeleteObjectARB(m_program);
	}
	
	static void PrintResourceUsage()
	{
		GLint uniform, varying, attribs;
		glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB, &uniform);
		glGetIntegerv(GL_MAX_VARYING_FLOATS_ARB, &varying);
		glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &attribs);
		std::cout << "      Vertex uniforms   : " << uniform << std::endl;
		std::cout << "             varyings   : " << varying << std::endl;
		std::cout << "             attribs    : " << attribs << std::endl;
		glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB, &uniform);
		std::cout << "      Fragment uniforms : " << uniform << std::endl;
	}
	
	virtual bool LoadShader(Target target, const std::vector<STR_String> &shaders, const STR_String &name)
	{
		GLenum shaderType;
		switch (target)
		{
			case VERTEX:
				shaderType = GL_VERTEX_SHADER_ARB;
				if (!GLEW_ARB_vertex_shader)
					return false;
				break;
			case FRAGMENT:
				shaderType = GL_FRAGMENT_SHADER_ARB;
				if (!GLEW_ARB_fragment_shader)
					return false;
				break;
		}

		glGetError();
				
		GLhandleARB object = glCreateShaderObjectARB(shaderType);
		if (!object)
			return false;
		
		std::vector<STR_String>::const_iterator shit;
		
		std::vector<const GLcharARB*> shaderSource;
		std::vector<GLint> length;
		
		for (shit = shaders.begin(); shit != shaders.end(); ++shit)
		{
			shaderSource.push_back((*shit).ReadPtr());
			length.push_back((*shit).Length());
		}
		
		// Load code
		glShaderSourceARB(object, shaders.size(), &shaderSource[0], &length[0]);
			
		// Compile shader
		glCompileShaderARB(object);
		
		// Print the compile log
		PrintInfoLog(object, "Compiling Shader: ", name);
		
		// Check compiled
		GLint compiled = 0;
		glGetObjectParameterivARB(object, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
		if (compiled)
		{
			// Attach to program
			glAttachObjectARB(m_program, object);
		}
		
		// Delete vertex shader (no longer required)
		glDeleteObjectARB(object);
		
		GLenum err = glGetError();
		if (err != GL_NO_ERROR)
			std::cout << "OpenGL error: " << gluErrorString(err) << std::endl;
		
		return compiled;
	}
	
	virtual bool LinkShader()
	{
		glGetError();
		
		glLinkProgramARB(m_program);
		
		m_valid = false;
		
		GLint linked = false;
		glGetObjectParameterivARB(m_program, GL_OBJECT_LINK_STATUS_ARB, &linked);
		if (!linked)
		{
			PrintInfoLog(m_program, "Linking Shaders", "");
			std::cout << "Shaders failed to link." << std::endl;
			return false;
		}
		
		glValidateProgramARB(m_program);
		
		glGetObjectParameterivARB(m_program, GL_OBJECT_VALIDATE_STATUS_ARB, &m_valid);
		if (!m_valid)
		{
			PrintInfoLog(m_program, "Validating Shaders", "");
			std::cout << "Shaders failed to validate." << std::endl;
			return false;
		}
		
		PrintInfoLog(m_program, "Linked Shaders", "");
		
		return true;
	}
	
	virtual void Enable()
	{
		if (m_valid)
		{
			glDisable(GL_COLOR_MATERIAL);
			glUseProgramObjectARB(m_program);
		}
	}
	
	virtual void Disable()
	{
		glUseProgramObjectARB(0);
		glEnable(GL_COLOR_MATERIAL);
	}
	
	virtual bool isValid()
	{
		return m_valid;
	}
	
	virtual int GetAttribLocation(const STR_String& name)
	{
		return glGetAttribLocationARB(m_program, name.ReadPtr());
	}
	
	virtual void BindAttribute(const STR_String& attr, int loc)
	{
		glBindAttribLocationARB(m_program, loc, attr.ReadPtr());
	}
	
	virtual int GetUniformLocation(const STR_String& name)
	{
		return glGetUniformLocationARB(m_program, name.ReadPtr());
	}
	
	virtual void SetUniform(int uniform, const MT_Tuple2& vec)
	{
		float value[2];
		vec.getValue(value);
		glUniform2fvARB(uniform, 1, value);
	}
	
	virtual void SetUniform(int uniform, const MT_Tuple3& vec)
	{
		float value[3];
		vec.getValue(value);
		glUniform3fvARB(uniform, 1, value);
	}
	
	virtual void SetUniform(int uniform, const MT_Tuple4& vec)
	{
		float value[4];
		vec.getValue(value);
		glUniform4fvARB(uniform, 1, value);
	}
	
	virtual void SetUniform(int uniform, const unsigned int& val)
	{
		glUniform1iARB(uniform, val);
	}
	
	virtual void SetUniform(int uniform, const float& val)
	{
		glUniform1fARB(uniform, val);
	}

	virtual void SetUniform(int uniform, const MT_Matrix4x4& vec, bool transpose)
	{
		float value[16];
		vec.getValue(value);
		glUniformMatrix4fvARB(uniform, 1, transpose?GL_TRUE:GL_FALSE, value);
	}
	
	virtual void SetUniform(int uniform, const MT_Matrix3x3& vec, bool transpose)
	{
		float value[9];
		value[0] = vec[0][0]; value[1] = vec[1][0]; value[2] = vec[2][0]; 
		value[3] = vec[0][1]; value[4] = vec[1][1]; value[5] = vec[2][1]; 
		value[6] = vec[0][2]; value[7] = vec[1][2]; value[7] = vec[2][2]; 
		glUniformMatrix3fvARB(uniform, 1, transpose?GL_TRUE:GL_FALSE, value);
	}
	
	static bool QueryExtensions()
	{
		return GLEW_ARB_shading_language_100 && GLEW_ARB_shader_objects;
	}
};



/******************************************************************************
                       Lighting Class
******************************************************************************/
class RAS_OpenGLRasterizer::GLLight : public RAS_IRasterizer::ILighting
{
	RAS_OpenGLRasterizer		*m_rasty;
	std::vector<RAS_LightObject*>	 m_lights;
	GLint				 m_numLights;
public:
	GLLight(RAS_OpenGLRasterizer *rasty)
		: ILighting(),
		m_rasty(rasty)
	{
		glGetIntegerv(GL_MAX_LIGHTS, &m_numLights);
		m_lights.resize(m_numLights);
	}
	virtual ~GLLight()
	{
	}
	
	virtual unsigned int GetNumLights()
	{
		return m_numLights;
	}
	
	virtual void Enable (unsigned int light, RAS_LightObject *data)
	{
		if (light > m_numLights)
			return;
			
		float vec[4];
		vec[0] = (*(data->m_worldmatrix))(0,3);
		vec[1] = (*(data->m_worldmatrix))(1,3);
		vec[2] = (*(data->m_worldmatrix))(2,3);

		if(data->m_type==RAS_LightObject::LIGHT_SUN) 
		{
			MT_Vector3 lightdir((*(data->m_worldmatrix))(0,2),
					(*(data->m_worldmatrix))(1,2),
					(*(data->m_worldmatrix))(2,2));
			lightdir.normalize();
			lightdir.getValue(vec);
			vec[3]= 0.0;
			glLightfv((GLenum)(GL_LIGHT0+light), GL_POSITION, vec); 
		}
		else 
		{
			vec[3]= 1.0;
			glLightfv((GLenum)(GL_LIGHT0+light), GL_POSITION, vec); 
			glLightf((GLenum)(GL_LIGHT0+light), GL_CONSTANT_ATTENUATION, 1.0);
			glLightf((GLenum)(GL_LIGHT0+light), GL_LINEAR_ATTENUATION, data->m_att1/data->m_distance);
			// without this next line it looks backward compatible.
			//attennuation still is acceptable 
			glLightf((GLenum)(GL_LIGHT0+light), GL_QUADRATIC_ATTENUATION, data->m_att2/(data->m_distance*data->m_distance)); 
			
			if(data->m_type==RAS_LightObject::LIGHT_SPOT) 
			{
				MT_Vector3 lightdir(-(*(data->m_worldmatrix))(0,2),
						-(*(data->m_worldmatrix))(1,2),
						-(*(data->m_worldmatrix))(2,2));
				lightdir.normalize();
				lightdir.getValue(vec);
				glLightfv((GLenum)(GL_LIGHT0+light), GL_SPOT_DIRECTION, vec);
				glLightf((GLenum)(GL_LIGHT0+light), GL_SPOT_CUTOFF, data->m_spotsize/2.0);
				glLightf((GLenum)(GL_LIGHT0+light), GL_SPOT_EXPONENT, 128.0*data->m_spotblend);
			}
			else glLightf((GLenum)(GL_LIGHT0+light), GL_SPOT_CUTOFF, 180.0);
		}
		
		if (data->m_nodiffuse)
		{
			vec[0] = vec[1] = vec[2] = vec[3] = 0.0;
		} else {
			vec[0]= data->m_energy*data->m_red;
			vec[1]= data->m_energy*data->m_green;
			vec[2]= data->m_energy*data->m_blue;
			vec[3]= 1.0;
		}
		glLightfv((GLenum)(GL_LIGHT0+light), GL_DIFFUSE, vec);
		glLightfv((GLenum)(GL_LIGHT0+light), GL_AMBIENT, vec);
		if (data->m_nospecular)
		{
			vec[0] = vec[1] = vec[2] = vec[3] = 0.0;
		} else if (data->m_nodiffuse) {
			vec[0]= data->m_energy*data->m_red;
			vec[1]= data->m_energy*data->m_green;
			vec[2]= data->m_energy*data->m_blue;
			vec[3]= 1.0;
		}
		glLightfv((GLenum)(GL_LIGHT0+light), GL_SPECULAR, vec);
		glEnable((GLenum)(GL_LIGHT0+light));
		
		m_lights[light] = data;
	
	}
	
	virtual void Enable (unsigned int light)
	{
		glEnable((GLenum)(GL_LIGHT0+light));
	}
	
	virtual void Disable (unsigned int light)
	{
		glDisable(GL_LIGHT0+light);
	}
	
	virtual RAS_LightObject* Light(unsigned int light)
	{
		return m_lights[light];
	}
	
	virtual bool IsEnabled(unsigned int light)
	{
		return (light < m_numLights) && glIsEnabled(GL_LIGHT0 + light);
	}
	
	virtual void BeginLighting()
	{
		glPushMatrix();
		m_rasty->LoadViewMatrix();
	}
	
	virtual void EndLighting()
	{
		glPopMatrix();
	}
	
	virtual void EnableLighting()
	{
		glEnable(GL_LIGHTING);
		
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);
		glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
		if (GLEW_EXT_separate_specular_color || GLEW_VERSION_1_2)
			glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
	}
	
	virtual void DisableLighting()
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
	}
	
};

/******************************************************************************
                       Drawhandle Class
******************************************************************************/
class RAS_OpenGLRasterizer::GLDrawHandle : public RAS_IDrawHandle
{
	GLuint  m_dlist;
	bool    m_modified;
	bool    m_stream;                // Object is changing, don't bother creating lists.      
	bool    m_create;
public:	
	GLDrawHandle()
		: RAS_IDrawHandle(),
		m_dlist(glGenLists(1)),
		m_modified(true),
		m_stream(false)
	{
	}
	
	virtual ~GLDrawHandle()
	{
		if (!m_stream)
			glDeleteLists(m_dlist, 1);
	}
	
	virtual bool IsVertexModified()
	{
		return m_modified;
	}
	
	virtual void SetVertexModified(bool modified)
	{
		if (modified)
		{
			m_stream = true;
			glDeleteLists(m_dlist, 1);
		}
		m_modified = modified;
	}
	
	virtual bool IsIndexModified()
	{
		return m_modified;
	}
	
	virtual void SetIndexModified(bool modified)
	{
		if (modified)
		{
			m_stream = true;
			glDeleteLists(m_dlist, 1);
		}
		m_modified = modified;
	}
	
	bool DrawList(bool create)
	{
		if (m_stream)
			return true;

		if (m_modified)
		{
			if (create)
				glNewList(m_dlist, GL_COMPILE_AND_EXECUTE);
			m_create = create;
			return true;
		}
		
		glCallList(m_dlist);
		return false;
	}
	
	void EndList()
	{
		if (m_modified && !m_stream && m_create)
		{
			glEndList();
			m_modified = false;
		}
	}
};


/******************************************************************************
                       Main Rasterizer Class
******************************************************************************/

RAS_OpenGLRasterizer::RAS_OpenGLRasterizer(RAS_ICanvas* canvas)
	:RAS_IRasterizer(canvas),
	m_2DCanvas(canvas),
	m_fogenabled(false),
	m_time(0.0),
	m_stereomode(RAS_STEREO_NOSTEREO),
	m_curreye(RAS_STEREO_LEFTEYE),
	m_eyeseparation(-1.0f),
	m_focallength(-1.0f),
	m_noOfScanlines(32),
	m_meshcache(true),
	m_depthmask(true),
	m_pass(0),
	m_activetexture(0),
	m_materialCachingInfo(0)
{
	//DEBUG:
	//GLEW_ARB_shadow = 0;
	//GLEW_ARB_multitexture = 0;
	//GLEW_ARB_shader_objects = 0;
	//GLEW_ARB_fragment_shader = 0;
	//GLEW_ARB_fragment_program = 0;
	//GLEW_VERSION_1_3 = 0;

	GLint numtexunits = GetNumTextureUnits();
	m_viewmatrix.Identity();
		
	if (!GLEW_ARB_fragment_program || !GLEW_ARB_vertex_program)
		m_perpixellights = 0;
		
	m_texunits.resize(numtexunits);
	for (unsigned int i = 0; i < numtexunits; i++)
	{
		m_texunits[i].texco = RAS_TEXCO_DISABLE;
		m_texunits[i].enable = false;
	}
	SetTexUnit(0);
	
	SetTextureCompression(true);
	if (GLEW_EXT_blend_subtract)
		m_blending = new GLBlendEXT();
	else
		m_blending = new GLBlend();
	
	m_lighting = new GLLight(this);
	
	PrintGLInfo();
}

RAS_OpenGLRasterizer::~RAS_OpenGLRasterizer()
{
	if (GLSLShader::QueryExtensions())
		glUseProgramObjectARB(0);
	if (GLEW_ARB_vertex_program)
		glDisable(GL_VERTEX_PROGRAM_ARB);
	if (GLEW_ARB_fragment_program)
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
	for (unsigned int i = 0; i < m_texunits.size(); i++)
	{
		SetTexUnit(i);
		EnableTextures(false);
	}
	SetTexUnit(0);
	
	m_texunits.clear();
	
	delete m_lighting;
	delete m_blending;
}


bool RAS_OpenGLRasterizer::PrintGLInfo()
{
	static bool debug = true;
	if (!debug)
		return false;
	
	debug = false;
	std::cout << "RAS_OpenGLRasterizer:" << std::endl;
	std::cout << "   GL Vendor  : " << (const char*)glGetString(GL_VENDOR) << std::endl;
	std::cout << "   GL Renderer: " << (const char*)glGetString(GL_RENDERER) << std::endl;
	std::cout << "   GL Version : " << (const char*)glGetString(GL_VERSION) << std::endl;

	GLint r, g, b, a, bits;
	glGetIntegerv(GL_RED_BITS, &r);
	glGetIntegerv(GL_GREEN_BITS, &g);
	glGetIntegerv(GL_BLUE_BITS, &b);
	glGetIntegerv(GL_ALPHA_BITS, &a);
	bits = r + g + b + a;
	std::cout << "   Colour     : " << bits << " bits (" << r << "/" << g << "/" << b << "/" << a << ")" << std::endl;
	glGetIntegerv(GL_ACCUM_RED_BITS, &r);
	glGetIntegerv(GL_ACCUM_GREEN_BITS, &g);
	glGetIntegerv(GL_ACCUM_BLUE_BITS, &b);
	glGetIntegerv(GL_ACCUM_ALPHA_BITS, &a);
	bits = r + g + b + a;
	if (bits)
		std::cout << "   Accum      : " << bits << " bits (" << r << "/" << g << "/" << b << "/" << a << ")" << std::endl;
	else
		std::cout << "   Accum      : Disabled" << std::endl;
	glGetIntegerv(GL_DEPTH_BITS, &r);
	glGetIntegerv(GL_STENCIL_BITS, &g);
	std::cout << "   Depth      : " << r << " bits" << std::endl;
	std::cout << "   Stencil    : " << g << " bits" << std::endl;
	if (GLEW_ARB_multisample)
	{
		glGetIntegerv(GL_SAMPLES_ARB, &b);
		if (b)
			std::cout << "   Multisample: " << b << " samples" << std::endl;
		else
			std::cout << "   Multisample: Disabled" << std::endl;
	} else
		std::cout << "   Multisample: Not supported" << std::endl;
	GLboolean s;
	glGetBooleanv(GL_STEREO, &s);
	std::cout << "   Stereo     : " << (s?"Enabled":"Disabled") << std::endl;
	std::cout << "   Shadow maps: " << (QueryShadowMaps()?"Supported":"Not supported") << std::endl;
	std::cout << "   GLSL       : " << (GLSLShader::QueryExtensions()?"Supported":"Not supported") << std::endl;
	if (GLSLShader::QueryExtensions())
		 GLSLShader::PrintResourceUsage();
	GLProgram::PrintResources();

	std::cout << "   Texunits   : " << GetNumTextureUnits() << std::endl;
	
	return true;
}



static void Myinit_gl_stuff(void)	
{
	float mat_specular[] = { 0.5, 0.5, 0.5, 1.0 };
	float mat_shininess[] = { 35.0 };
/*  	float one= 1.0; */
	int a, x, y;
	GLubyte pat[32*32];
	const GLubyte *patc= pat;
		
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);


//#if defined(__FreeBSD) || defined(__linux__)
//	glDisable(GL_DITHER);	/* op sgi/sun hardware && 12 bits */
//#endif
	
	/* no local viewer, looks ugly in ortho mode */
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE); 
	
	glDepthFunc(GL_LESS);
	/* scaling matrices */
	glEnable(GL_NORMALIZE);

	glShadeModel(GL_FLAT);

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glDisable(GL_LOGIC_OP);
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_COLOR_MATERIAL);

	glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
	glPixelTransferi(GL_RED_SCALE, 1);
	glPixelTransferi(GL_RED_BIAS, 0);
	glPixelTransferi(GL_GREEN_SCALE, 1);
	glPixelTransferi(GL_GREEN_BIAS, 0);
	glPixelTransferi(GL_BLUE_SCALE, 1);
	glPixelTransferi(GL_BLUE_BIAS, 0);
	glPixelTransferi(GL_ALPHA_SCALE, 1);
	glPixelTransferi(GL_ALPHA_BIAS, 0);

	a = 0;
	for(x=0; x<32; x++)
	{
		for(y=0; y<4; y++)
		{
			if( (x) & 1) pat[a++]= 0x88;
			else pat[a++]= 0x22;
		}
	}
	
	glPolygonStipple(patc);
	glReadBuffer(GL_BACK);
	glDrawBuffer(GL_BACK);
	
#ifdef RAS_MOTIONBLUR
	glClearAccum(0.0, 0.0, 0.0, 1.0);
	glClear(GL_ACCUM_BUFFER_BIT);
#endif

}

bool RAS_OpenGLRasterizer::Init()
{
	Myinit_gl_stuff();
	
#ifdef RAS_MULTISAMPLE
	if (GLEW_ARB_multisample)
	{
		glEnable(GL_MULTISAMPLE_ARB);
	}
	
	glEnable(GL_POLYGON_SMOOTH);
#endif
	
#ifdef RAS_ANISO
	if (GLEW_EXT_texture_filter_anisotropic)
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 2.0);
#endif

	m_redback = 0.4375;
	m_greenback = 0.4375;
	m_blueback = 0.4375;
	m_alphaback = 0.0;
	
	glClearColor(m_redback,m_greenback,m_blueback,m_alphaback);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glShadeModel(GL_SMOOTH);

	return true;
}



void RAS_OpenGLRasterizer::SetBackColor(float red,
										float green,
										float blue,
										float alpha)
{
	m_redback = red;
	m_greenback = green;
	m_blueback = blue;
	m_alphaback = alpha;
}



void RAS_OpenGLRasterizer::SetFogColor(float r,
									   float g,
									   float b)
{
	m_fogr = r;
	m_fogg = g;
	m_fogb = b;
	m_fogenabled = true;
}



void RAS_OpenGLRasterizer::SetFogStart(float start)
{
	m_fogstart = start;
	m_fogenabled = true;
}



void RAS_OpenGLRasterizer::SetFogEnd(float fogend)
{
	m_fogdist = fogend;
	m_fogenabled = true;
}



void RAS_OpenGLRasterizer::SetFog(float start,
								  float dist,
								  float r,
								  float g,
								  float b)
{
	m_fogstart = start;
	m_fogdist = dist;
	m_fogr = r;
	m_fogg = g;
	m_fogb = b;
	m_fogenabled = true;
}



void RAS_OpenGLRasterizer::DisableFog()
{
	m_fogenabled = false;
}



void RAS_OpenGLRasterizer::DisplayFog()
{
	if ((m_drawingmode >= KX_SOLID) && m_fogenabled)
	{
		float params[5];
		glFogi(GL_FOG_MODE, GL_LINEAR);
		glFogf(GL_FOG_DENSITY, 0.1f);
		glFogf(GL_FOG_START, m_fogstart);
		glFogf(GL_FOG_END, m_fogstart + m_fogdist);
		params[0]= m_fogr;
		params[1]= m_fogg;
		params[2]= m_fogb;
		params[3]= 0.0;
		glFogfv(GL_FOG_COLOR, params); 
		glEnable(GL_FOG);
	} 
	else
	{
		glDisable(GL_FOG);
	}
}



bool RAS_OpenGLRasterizer::SetMaterial(const RAS_IPolyMaterial& mat)
{
	if (m_pass)
		glDepthFunc(GL_LESS);
	m_pass = 0;
	return mat.Activate(this, m_materialCachingInfo);
}

bool RAS_OpenGLRasterizer::ActivateNextPass(const RAS_IPolyMaterial& mat)
{
	bool dopass = mat.ActivateNextPass(this, m_pass);
	if (dopass && m_pass == 1)
		glDepthFunc(GL_EQUAL);
	m_pass++;
	return dopass;
}


void RAS_OpenGLRasterizer::Exit()
{

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glClearDepth(1.0); 
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glClearColor(m_redback, m_greenback, m_blueback, m_alphaback);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDepthMask (GL_TRUE);
	glDepthFunc(GL_LEQUAL);
	glBlendFunc(GL_ONE, GL_ZERO);
	
	glDisable(GL_POLYGON_STIPPLE);
	
	glDisable(GL_LIGHTING);
	if (GLEW_ATI_pn_triangles)
		glDisable(GL_PN_TRIANGLES_ATI);
	
	if (GLEW_EXT_separate_specular_color || GLEW_VERSION_1_2)
		glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SINGLE_COLOR);
	
#ifdef RAS_MULTISAMPLE
	glDisable(GL_POLYGON_SMOOTH);

	if (GLEW_ARB_multisample)
		glDisable(GL_MULTISAMPLE_ARB);
#endif
		
	EndFrame();
}

bool RAS_OpenGLRasterizer::InterlacedStereo() const
{
	return m_stereomode == RAS_STEREO_VINTERLACE || m_stereomode == RAS_STEREO_INTERLACED;
}

bool RAS_OpenGLRasterizer::BeginFrame(int drawingmode, double time)
{
	m_time = time;
	m_drawingmode = drawingmode;
	
	if (!InterlacedStereo() || m_curreye == RAS_STEREO_LEFTEYE)
	{
		m_2DCanvas->ClearColor(m_redback,m_greenback,m_blueback,m_alphaback);
		m_2DCanvas->ClearBuffer(RAS_ICanvas::COLOR_BUFFER);
	}

	// Blender camera routine destroys the settings
	if (m_drawingmode < KX_SOLID)
	{
		glDisable (GL_CULL_FACE);
		glDisable (GL_DEPTH_TEST);
	}
	else
	{
		glEnable(GL_DEPTH_TEST);
		glEnable (GL_CULL_FACE);
	}

	glShadeModel(GL_SMOOTH);

	m_2DCanvas->BeginFrame();
	
	return true;
}

bool RAS_OpenGLRasterizer::BeginShadowMap(float bias)
{
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	m_2DCanvas->ClearBuffer(RAS_ICanvas::DEPTH_BUFFER);
	m_2DCanvas->BeginFrame();
	
	//glDisable(GL_LIGHTING);
	
	glPolygonOffset(bias, 10.0f);
	glEnable(GL_POLYGON_OFFSET_FILL);
	
	glCullFace(GL_FRONT);
	glEnable(GL_CULL_FACE);
	
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER, 0.5);
	
	/*for (int i = 0; i < GetNumTextureUnits(); i++)
	{
		SetTexUnit(i);
		SetTexCoords(RAS_TEXCO_DISABLE);
	}*/
	return true;
}

void RAS_OpenGLRasterizer::EndShadowMap()
{
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	m_2DCanvas->EndFrame();
	glDisable(GL_POLYGON_OFFSET_FILL);
	glCullFace(GL_BACK);
	glDisable(GL_ALPHA_TEST);
}

void RAS_OpenGLRasterizer::SetDrawingMode(int drawingmode)
{
	m_drawingmode = drawingmode;

	switch (m_drawingmode)
	{
	case KX_BOUNDINGBOX:
		{
		}
	case KX_WIREFRAME:
		{
			glDisable (GL_CULL_FACE);
			break;
		}
	case KX_TEXTURED:
		{
		}
	case KX_SHADED:
		{
		}
	case KX_SOLID:
		{
		}
	default:
		{
		}
	}
}



int RAS_OpenGLRasterizer::GetDrawingMode()
{
	return m_drawingmode;
}



void RAS_OpenGLRasterizer::SetDepthMask(DepthMask depthmask)
{
	switch (depthmask)
	{
	default:
	case KX_DEPTHMASK_ENABLED:
		{
			glDepthMask(GL_TRUE);
			m_depthmask = true;
			break;
		};
	case KX_DEPTHMASK_DISABLED:
		{
			glDepthMask(GL_FALSE);
			m_depthmask = false;
			break;
		};
	}
	glDepthMask(depthmask == KX_DEPTHMASK_DISABLED ? GL_FALSE : GL_TRUE);
}



void RAS_OpenGLRasterizer::ClearDepthBuffer()
{
	m_2DCanvas->ClearBuffer(RAS_ICanvas::DEPTH_BUFFER);
}


void RAS_OpenGLRasterizer::ClearCachingInfo(void)
{
	m_materialCachingInfo = 0;
}


void RAS_OpenGLRasterizer::EndFrame()
{
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	m_2DCanvas->EndFrame();
}	

void RAS_OpenGLRasterizer::SetRenderArea()
{
	// only above/below stereo method needs viewport adjustment
	switch (m_stereomode)
	{
		case RAS_STEREO_ABOVEBELOW:
			switch(m_curreye)
			{
				case RAS_STEREO_LEFTEYE:
					// upper half of window
					m_2DCanvas->GetDisplayArea().SetLeft(0);
					m_2DCanvas->GetDisplayArea().SetBottom(m_2DCanvas->GetHeight() -
						int(m_2DCanvas->GetHeight() - m_noOfScanlines) / 2);
	
					m_2DCanvas->GetDisplayArea().SetRight(int(m_2DCanvas->GetWidth()));
					m_2DCanvas->GetDisplayArea().SetTop(int(m_2DCanvas->GetHeight()));
					break;
				case RAS_STEREO_RIGHTEYE:
					// lower half of window
					m_2DCanvas->GetDisplayArea().SetLeft(0);
					m_2DCanvas->GetDisplayArea().SetBottom(0);
					m_2DCanvas->GetDisplayArea().SetRight(int(m_2DCanvas->GetWidth()));
					m_2DCanvas->GetDisplayArea().SetTop(int(m_2DCanvas->GetHeight() - m_noOfScanlines) / 2);
					break;
			}
			break;
		case RAS_STEREO_SIDEBYSIDE:
			switch (m_curreye)
			{
				case RAS_STEREO_LEFTEYE:
					// Left half of window
					m_2DCanvas->GetDisplayArea().SetLeft(0);
					m_2DCanvas->GetDisplayArea().SetBottom(0);
					m_2DCanvas->GetDisplayArea().SetRight(m_2DCanvas->GetWidth()/2);
					m_2DCanvas->GetDisplayArea().SetTop(m_2DCanvas->GetHeight());
					break;
				case RAS_STEREO_RIGHTEYE:
					// Right half of window
					m_2DCanvas->GetDisplayArea().SetLeft(m_2DCanvas->GetWidth()/2);
					m_2DCanvas->GetDisplayArea().SetBottom(0);
					m_2DCanvas->GetDisplayArea().SetRight(m_2DCanvas->GetWidth());
					m_2DCanvas->GetDisplayArea().SetTop(m_2DCanvas->GetHeight());
					break;
			}
			break;
		default:
			// every available pixel
			m_2DCanvas->GetDisplayArea().SetLeft(0);
			m_2DCanvas->GetDisplayArea().SetBottom(0);
			m_2DCanvas->GetDisplayArea().SetRight(int(m_2DCanvas->GetWidth()));
			m_2DCanvas->GetDisplayArea().SetTop(int(m_2DCanvas->GetHeight()));
			break;
	}
}

	
void RAS_OpenGLRasterizer::SetStereoMode(const StereoMode stereomode)
{
	m_stereomode = stereomode;
}



bool RAS_OpenGLRasterizer::Stereo()
{
	if(m_stereomode == RAS_STEREO_NOSTEREO)
		return false;
	else
		return true;
}


void RAS_OpenGLRasterizer::SetEye(StereoEye eye)
{
	m_curreye = eye;
	switch (m_stereomode)
	{
		case RAS_STEREO_QUADBUFFERED:
			glDrawBuffer(m_curreye == RAS_STEREO_LEFTEYE ? GL_BACK_LEFT : GL_BACK_RIGHT);
			break;
		case RAS_STEREO_ANAGLYPH:
			if (m_curreye == RAS_STEREO_LEFTEYE)
			{
				glColorMask(GL_FALSE, GL_TRUE, GL_TRUE, GL_FALSE);
			} else {
				//glAccum(GL_LOAD, 1.0);
				glColorMask(GL_TRUE, GL_FALSE, GL_FALSE, GL_FALSE);
				ClearDepthBuffer();
			}
			break;
		case RAS_STEREO_VINTERLACE:
		{
			GLuint pat[32];
			const unsigned char mask = 0x55;  // 01010101
			memset(pat, m_curreye == RAS_STEREO_RIGHTEYE?~mask:mask, sizeof(pat));
			glEnable(GL_POLYGON_STIPPLE);
			glPolygonStipple((const GLubyte*) pat);
			if (m_curreye == RAS_STEREO_RIGHTEYE)
				ClearDepthBuffer();
			break;
		}
		case RAS_STEREO_INTERLACED:
		{
			GLuint pat[32];
			GLuint mask = m_curreye == RAS_STEREO_LEFTEYE?~0:0;
			for (int y = 0; y < 32; y+=2)
			{
				pat[y] = mask;
				pat[y+1] = ~mask;
			}
			glEnable(GL_POLYGON_STIPPLE);
			glPolygonStipple((const GLubyte*) pat);
			if (m_curreye == RAS_STEREO_RIGHTEYE)
				ClearDepthBuffer();
			break;
		}
		default:
			break;
	}
}

RAS_IRasterizer::StereoEye RAS_OpenGLRasterizer::GetEye()
{
	return m_curreye;
}


void RAS_OpenGLRasterizer::SetEyeSeparation(float eyeseparation)
{
	m_eyeseparation = eyeseparation;
}

float RAS_OpenGLRasterizer::GetEyeSeparation()
{
	return m_eyeseparation;
}

void RAS_OpenGLRasterizer::SetFocalLength(float focallength)
{
	m_focallength = focallength;
}

float RAS_OpenGLRasterizer::GetFocalLength()
{
	return m_focallength;
}


void RAS_OpenGLRasterizer::SwapBuffers()
{
#ifdef RAS_MOTIONBLUR
	static double lasttime = 0.0;
	const float MOTIONFAC = 10.0*(m_time - lasttime);
	lasttime = m_time;
	glAccum(GL_MULT, 1.0 - MOTIONFAC);
	glAccum(GL_ACCUM, MOTIONFAC);
	glAccum(GL_RETURN, 1.0);
#endif
	m_2DCanvas->SwapBuffers();
}

void RAS_OpenGLRasterizer::SetMeshCache(bool cache)
{
	m_meshcache = cache;
}

RAS_IDrawHandle* RAS_OpenGLRasterizer::IndexPrimitives(RAS_IDrawHandle *drawhandle,
						const vecVertexArray & vertexarrays,
						const vecIndexArrays & indexarrays,
						int mode,
						class RAS_IPolyMaterial*,
						class RAS_IRenderTools*,
						bool useObjectColor,
						const MT_Vector4& rgbacolor
						)
{ 
	GLenum drawmode;
	switch (mode)
	{
	case 0:
		drawmode = GL_TRIANGLES;
		break;
	default:
	case 1:
		drawmode = GL_LINES;
		break;
	case 2:
		drawmode = GL_QUADS;
		break;
	}
	
	const RAS_TexVert* vertexarray ;
	unsigned int numindices,vt;
	
	GLDrawHandle *h = static_cast<GLDrawHandle*>(drawhandle);
	if (!h)
		h = new GLDrawHandle();
	
	if (h->DrawList(m_meshcache))
	{
		for (vt=0;vt<vertexarrays.size();vt++)
		{
			vertexarray = &((*vertexarrays[vt]) [0]);
			const KX_IndexArray & indexarray = (*indexarrays[vt]);
			numindices = indexarray.size();
			
			if (!numindices)
				break;
			
			int vindex=0;
			switch (mode)
			{
			case 1:
				{
					glBegin(GL_LINES);
					vindex=0;
					for (unsigned int i=0;i<numindices;i+=2)
					{
						glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
						glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
					}
					glEnd();
				}
				break;
			case 2:
				{
					glBegin(GL_TRIANGLES);
					vindex=0;
					if (useObjectColor)
					{
						for (unsigned int i=0;i<numindices;i+=4)
						{
	
							glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);
	
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 2
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 3
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex-=2;
							// 1
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex+=2;
							// 3
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 4
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
						}
					}
					else
					{
						for (unsigned int i=0;i<numindices;i+=4)
						{
							// This looks curiously endian unsafe to me.
							// However it depends on the way the colors are packed into 
							// the m_rgba field of RAS_TexVert
							// 1
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 2
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 3
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex-=2;
							// 1
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex+=2;
							// 3
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							// 4
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
						}
					}
					glEnd();	
					break;
				}
			case 0:
				{
					glBegin(GL_TRIANGLES);
					vindex=0;
					if (useObjectColor)
					{
						for (unsigned int i=0;i<numindices;i+=3)
						{
	
							glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);
	
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
						}
					}
					else 
					{
						for (unsigned int i=0;i<numindices;i+=3)
						{
	
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
							
							glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
							TexCoord(vertexarray[(indexarray[vindex])]);
							glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
							vindex++;
						}
					}
					glEnd();	
					break;
				}
			default:
				{
				}
				
			} // switch
		} // for each vertexarray
	} // Drawlist
	
	h->EndList();
	
	return h;
}

void RAS_OpenGLRasterizer::Tangent(const RAS_TexVert& v1,
					const RAS_TexVert& v2,
					const RAS_TexVert& v3,
					const MT_Vector3 &no)
{
	if (!GLEW_ARB_vertex_program)
		return;
	MT_Vector3 x1(v1.getLocalXYZ()), x2(v2.getLocalXYZ()), x3(v3.getLocalXYZ());
	MT_Vector2 uv1(v1.getUV1()), uv2(v2.getUV1()), uv3(v3.getUV1());
	MT_Vector3 dx1(x2 - x1), dx2(x3 - x1);
	MT_Vector2 duv1(uv2 - uv1), duv2(uv3 - uv1);
	
	MT_Scalar r = 1.0 / (duv1.x() * duv2.y() - duv2.x() * duv1.y());
	duv1 *= r;
	duv2 *= r;
	MT_Vector3 sdir(duv2.y() * dx1 - duv1.y() * dx2);
	MT_Vector3 tdir(duv1.x() * dx2 - duv2.x() * dx1);
	
	// Gram-Schmidt orthogonalize
	MT_Vector3 t(sdir - no.cross(no.cross(sdir)));
	if (!MT_fuzzyZero(t))
		t /= t.length();

	float tangent[4];
	t.getValue(tangent);
	// Calculate handedness
	tangent[3] = no.dot(sdir.cross(tdir)) < 0.0 ? -1.0 : 1.0;
	glVertexAttrib4fvARB(1, tangent);
}


void RAS_OpenGLRasterizer::IndexPrimitives_Ex(	const vecVertexArray & vertexarrays,
						const vecIndexArrays & indexarrays,
						int mode,
						class RAS_IPolyMaterial*,
						class RAS_IRenderTools*,
						bool useObjectColor,
						const MT_Vector4& rgbacolor
						)
{ 
	bool	recalc;
	GLenum drawmode;
	switch (mode)
	{
	case 0:
		drawmode = GL_TRIANGLES;
		break;
	case 1:
		drawmode = GL_LINES;
		break;
	case 2:
		drawmode = GL_QUADS;
		break;
	default:
		drawmode = GL_LINES;
		break;
	}
	
	const RAS_TexVert* vertexarray ;
	unsigned int numindices,vt;

	for (vt=0;vt<vertexarrays.size();vt++)
	{
		vertexarray = &((*vertexarrays[vt]) [0]);
		const KX_IndexArray & indexarray = (*indexarrays[vt]);
		numindices = indexarray.size();
		
		if (!numindices)
			continue;
		
		int vindex=0;
		switch (mode)
		{
		case 1:
			{
				glBegin(GL_LINES);
				vindex=0;
				for (unsigned int i=0;i<numindices;i+=2)
				{
					glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
					glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
				}
				glEnd();
			}
			break;
		case 2:
			{
				glBegin(GL_TRIANGLES);
				vindex=0;
				if (useObjectColor)
				{
					for (unsigned int i=0;i<numindices;i+=4)
					{
						MT_Point3 mv1, mv2, mv3, mv4;
						MT_Vector3 fnor;
						/* Calc a new face normal */

						if (vertexarray[(indexarray[vindex])].getFlag() & TV_CALCFACENORMAL)
							recalc= true;
						else
							recalc=false;

						if (recalc){
							mv1 = MT_Point3(vertexarray[(indexarray[vindex])].getLocalXYZ());
							mv2 = MT_Point3(vertexarray[(indexarray[vindex+1])].getLocalXYZ());
							mv3 = MT_Point3(vertexarray[(indexarray[vindex+2])].getLocalXYZ());
							mv4 = MT_Point3(vertexarray[(indexarray[vindex+3])].getLocalXYZ());
							
							fnor = (((mv2-mv1).cross(mv3-mv2))+((mv4-mv3).cross(mv1-mv4))).safe_normalized();

							glNormal3f(fnor[0], fnor[1], fnor[2]);
							Tangent(vertexarray[(indexarray[vindex])], vertexarray[(indexarray[vindex+1])], vertexarray[(indexarray[vindex+2])], fnor);
						}

						glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);
						 // 1
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						// 2
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						//3
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex-=2;
						//1
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex+=2;
						//3
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						//4
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
					}
				}
				else
				{
					for (unsigned int i=0;i<numindices;i+=4)
					{
						// This looks curiously endian unsafe to me.
						// However it depends on the way the colors are packed into 
						// the m_rgba field of RAS_TexVert
						MT_Point3 mv1, mv2, mv3, mv4, fnor;
						/* Calc a new face normal */

						if (vertexarray[(indexarray[vindex])].getFlag() & TV_CALCFACENORMAL)
							recalc= true;
						else
							recalc=false;


						if (recalc){
							mv1 = MT_Point3(vertexarray[(indexarray[vindex])].getLocalXYZ());
							mv2 = MT_Point3(vertexarray[(indexarray[vindex+1])].getLocalXYZ());
							mv3 = MT_Point3(vertexarray[(indexarray[vindex+2])].getLocalXYZ());
							mv4 = MT_Point3(vertexarray[(indexarray[vindex+3])].getLocalXYZ());
							
							fnor = (((mv2-mv1).cross(mv3-mv2))+((mv4-mv3).cross(mv1-mv4))).safe_normalized();

							glNormal3f(fnor[0], fnor[1], fnor[2]);
						}
						//1
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						//2
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						//3
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex-=2;
						//1
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex+=2;
						//3
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						//4
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
					}
				}
				glEnd();	
				break;
			}
		case 0:
			{
				glBegin(GL_TRIANGLES);
				vindex=0;
				if (useObjectColor)
				{
					for (unsigned int i=0;i<numindices;i+=3)
					{
						MT_Point3 mv1, mv2, mv3, fnor;
						/* Calc a new face normal */

						if (vertexarray[(indexarray[vindex])].getFlag() & TV_CALCFACENORMAL)
							recalc= true;
						else
							recalc=false;

						if (recalc){
							mv1 = MT_Point3(vertexarray[(indexarray[vindex])].getLocalXYZ());
							mv2 = MT_Point3(vertexarray[(indexarray[vindex+1])].getLocalXYZ());
							mv3 = MT_Point3(vertexarray[(indexarray[vindex+2])].getLocalXYZ());
							
							fnor = ((mv2-mv1).cross(mv3-mv2)).safe_normalized();
							glNormal3f(fnor[0], fnor[1], fnor[2]);
						}

						glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);

						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
					}
				}
				else 
				{
					for (unsigned int i=0;i<numindices;i+=3)
					{
						MT_Point3 mv1, mv2, mv3, fnor;
						/* Calc a new face normal */

						if (vertexarray[(indexarray[vindex])].getFlag() & TV_CALCFACENORMAL)
							recalc= true;
						else
							recalc=false;


						if (recalc){
							mv1 = MT_Point3(vertexarray[(indexarray[vindex])].getLocalXYZ());
							mv2 = MT_Point3(vertexarray[(indexarray[vindex+1])].getLocalXYZ());
							mv3 = MT_Point3(vertexarray[(indexarray[vindex+2])].getLocalXYZ());
							
							fnor = ((mv2-mv1).cross(mv3-mv2)).safe_normalized();
							glNormal3f(fnor[0], fnor[1], fnor[2]);
						}

						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
						
						glColor4ubv((const GLubyte *)(vertexarray[(indexarray[vindex])].getRGBA()));
						if (!recalc)
						{
							glNormal3fv(vertexarray[(indexarray[vindex])].getNormal());
							if (GLEW_ARB_vertex_program) glVertexAttrib4fvARB(1, vertexarray[(indexarray[vindex])].getTangent());
						}
						TexCoord(vertexarray[(indexarray[vindex])]);
						glVertex3fv(vertexarray[(indexarray[vindex])].getLocalXYZ());
						vindex++;
					}
				}
				glEnd();	
				break;
			}
		default:
			{
			}
			
		} // switch
	} // for each vertexarray
}



void RAS_OpenGLRasterizer::IndexPrimitives_3DText(const vecVertexArray & vertexarrays,
									const vecIndexArrays & indexarrays,
									int mode,
									class RAS_IPolyMaterial* polymat,
									class RAS_IRenderTools* rendertools,
									bool useObjectColor,
									const MT_Vector4& rgbacolor
									)
{ 
	GLenum drawmode;
	switch (mode)
	{
	case 0:
		drawmode = GL_TRIANGLES;
		break;
	case 1:
		drawmode = GL_LINES;
		break;
	case 2:
		drawmode = GL_QUADS;
		break;
	default:
		drawmode = GL_LINES;
		break;
	}
	
	const RAS_TexVert* vertexarray ;
	
	unsigned int numindices, vt;
	
	if (useObjectColor)
	{
		glDisableClientState(GL_COLOR_ARRAY);
		glColor4d(rgbacolor[0], rgbacolor[1], rgbacolor[2], rgbacolor[3]);
	}
	else
	{
		glEnableClientState(GL_COLOR_ARRAY);
	}
	
	for (vt=0;vt<vertexarrays.size();vt++)
	{
		vertexarray = &((*vertexarrays[vt]) [0]);
		const KX_IndexArray & indexarray = (*indexarrays[vt]);
		numindices = indexarray.size();
		
		if (!numindices)
			break;
		
		int vindex=0;
		switch (mode)
		{
		case 1:
			{
				glBegin(GL_LINES);
				vindex=0;
				for (unsigned int i=0;i<numindices;i+=2)
				{
					glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
					glVertex3fv(vertexarray[(indexarray[vindex++])].getLocalXYZ());
				}
				glEnd();
			}
			break;
		case 2:
			{
				vindex=0;
				for (unsigned int i=0;i<numindices;i+=4)
				{
					float v1[3],v2[3],v3[3],v4[3];
					
					v1[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v1[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v1[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					v2[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v2[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v2[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					v3[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v3[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v3[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					v4[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v4[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v4[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					
					vindex++;
					
					rendertools->RenderText(polymat->GetDrawingMode(),polymat,v1,v2,v3,v4);
					ClearCachingInfo();
				}
				break;
			}
		case 0:
			{
				glBegin(GL_TRIANGLES);
				vindex=0;
				for (unsigned int i=0;i<numindices;i+=3)
				{
					float v1[3],v2[3],v3[3];
					
					v1[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v1[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v1[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					v2[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v2[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v2[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					v3[0] = vertexarray[(indexarray[vindex])].getLocalXYZ()[0];
					v3[1] = vertexarray[(indexarray[vindex])].getLocalXYZ()[1];
					v3[2] = vertexarray[(indexarray[vindex])].getLocalXYZ()[2];
					vindex++;
					
					rendertools->RenderText(polymat->GetDrawingMode(),polymat,v1,v2,v3,NULL);		
					ClearCachingInfo();
				}
				glEnd();	
				break;
			}
		default:
			{
			}
		}	//switch
	}	//for each vertexarray
}



void RAS_OpenGLRasterizer::SetProjectionMatrix(MT_CmMatrix4x4 &mat)
{
	glMatrixMode(GL_PROJECTION);
	double* matrix = &mat(0,0);
	glLoadMatrixd(matrix);
}


void RAS_OpenGLRasterizer::SetProjectionMatrix(const MT_Matrix4x4 & mat)
{
	glMatrixMode(GL_PROJECTION);
	double matrix[16];
	/* Get into argument. Looks a bit dodgy, but it's ok. */
	mat.getValue(matrix);
	/* Internally, MT_Matrix4x4 uses doubles (MT_Scalar). */
	glLoadMatrixd(matrix);	
}

MT_Matrix4x4 RAS_OpenGLRasterizer::GetFrustumMatrix(
	float fov,
	float frustnear,
	float frustfar
){
	MT_Matrix4x4 result;
	double mat[16];
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fov, 1.0, frustnear, frustfar);
	glGetDoublev(GL_PROJECTION_MATRIX, mat);
	result.setValue(mat);
	return result;
}

MT_Matrix4x4 RAS_OpenGLRasterizer::GetFrustumMatrix(
	float left,
	float right,
	float bottom,
	float top,
	float frustnear,
	float frustfar,
	bool
){
	MT_Matrix4x4 result;
	double mat[16];

	// correction for stereo
	if(m_stereomode != RAS_STEREO_NOSTEREO)
	{
			float near_div_focallength;
			// next 2 params should be specified on command line and in Blender publisher
			if (m_focallength < 0.0f)
				m_focallength = 1.5 * right;  // derived from example
			if (m_eyeseparation < 0.0f)
				m_eyeseparation = 0.18 * right;  // just a guess...

			near_div_focallength = frustnear / m_focallength;
			switch(m_curreye)
			{
				case RAS_STEREO_LEFTEYE:
						left += 0.5 * m_eyeseparation * near_div_focallength;
						right += 0.5 * m_eyeseparation * near_div_focallength;
						break;
				case RAS_STEREO_RIGHTEYE:
						left -= 0.5 * m_eyeseparation * near_div_focallength;
						right -= 0.5 * m_eyeseparation * near_div_focallength;
						break;
			}
			// leave bottom, top, bottom and top untouched
	}
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(left, right, bottom, top, frustnear, frustfar);
		
	glGetDoublev(GL_PROJECTION_MATRIX, mat);
	result.setValue(mat);

	return result;
}


// next arguments probably contain redundant info, for later...
void RAS_OpenGLRasterizer::SetViewMatrix(const MT_Matrix4x4 &mat, const MT_Vector3& campos,
		const MT_Point3 &, const MT_Quaternion &camOrientQuat)
{
	MT_Matrix4x4 viewMat = mat;

	// correction for stereo
	if(m_stereomode != RAS_STEREO_NOSTEREO)
	{
		MT_Matrix3x3 camOrientMat3x3(camOrientQuat);
		MT_Vector3 unitViewDir(0.0, -1.0, 0.0);  // minus y direction, Blender convention
		MT_Vector3 unitViewupVec(0.0, 0.0, 1.0);
		MT_Vector3 viewDir, viewupVec;
		MT_Vector3 eyeline;

		// actual viewDir
		viewDir = camOrientMat3x3 * unitViewDir;  // this is the moto convention, vector on right hand side
		// actual viewup vec
		viewupVec = camOrientMat3x3 * unitViewupVec;

		// vector between eyes
		eyeline = viewDir.cross(viewupVec);

		switch(m_curreye)
		{
			case RAS_STEREO_LEFTEYE:
				{
				// translate to left by half the eye distance
				MT_Transform transform;
				transform.setIdentity();
				transform.translate(-(eyeline * m_eyeseparation / 2.0));
				viewMat *= transform;
				}
				break;
			case RAS_STEREO_RIGHTEYE:
				{
				// translate to right by half the eye distance
				MT_Transform transform;
				transform.setIdentity();
				transform.translate(eyeline * m_eyeseparation / 2.0);
				viewMat *= transform;
				}
				break;
		}
	}

	// convert row major matrix 'viewMat' to column major for OpenGL
	MT_Scalar cammat[16];
	viewMat.getValue(cammat);
	MT_CmMatrix4x4 viewCmmat = cammat;

	glMatrixMode(GL_MODELVIEW);
	m_viewmatrix = viewCmmat;
	glLoadMatrixd(&m_viewmatrix(0,0));
	m_campos = campos;
}

void RAS_OpenGLRasterizer::GetViewMatrix(MT_Matrix4x4 &mat) const
{
	float viewmat[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, viewmat);
	mat.setValue(viewmat);
}

const MT_Point3& RAS_OpenGLRasterizer::GetCameraPosition()
{
	return m_campos;
}



void RAS_OpenGLRasterizer::LoadViewMatrix()
{
	glLoadMatrixd(&m_viewmatrix(0,0));
}



void RAS_OpenGLRasterizer::EnableTextures(bool enable)
{
	if (enable)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);
}



void RAS_OpenGLRasterizer::SetCullFace(bool enable)
{
	if (enable)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);
}

void RAS_OpenGLRasterizer::SetLines(bool enable)
{
	if (enable)
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void RAS_OpenGLRasterizer::SetSpecularity(float specX,
										  float specY,
										  float specZ,
										  float specval)
{
	GLfloat mat_specular[] = {specX, specY, specZ, specval};
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
}



void RAS_OpenGLRasterizer::SetShinyness(float shiny)
{
	GLfloat mat_shininess[] = {	shiny };
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
}



void RAS_OpenGLRasterizer::SetDiffuse(float difX,float difY,float difZ,float diffuse)
{
	GLfloat mat_diffuse [] = {difX, difY,difZ, diffuse};
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);
}

void	RAS_OpenGLRasterizer::SetAmbient(const MT_Vector4 &amb)
{
	GLfloat mat_ambient[4];
	amb.getValue(mat_ambient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient);
}

void	RAS_OpenGLRasterizer::SetEmit(const MT_Vector4 &emit)
{
	GLfloat mat_emit[4];
	emit.getValue(mat_emit);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, mat_emit);
}

double RAS_OpenGLRasterizer::GetTime()
{
	return m_time;
}


void	RAS_OpenGLRasterizer::SetSubsurf(int level, bool linear)
{
	if (!GLEW_ATI_pn_triangles)
		return;
		
	if (level == 0)
	{
		glDisable(GL_PN_TRIANGLES_ATI);
	} else {
		int maxlevel = 0;
		glGetIntegerv(GL_MAX_PN_TRIANGLES_TESSELATION_LEVEL_ATI, &maxlevel);
		if (level > maxlevel)
			level = maxlevel;
		glEnable(GL_PN_TRIANGLES_ATI);
		glPNTrianglesiATI(GL_PN_TRIANGLES_POINT_MODE_ATI, linear ? GL_PN_TRIANGLES_POINT_MODE_LINEAR_ATI : 
									GL_PN_TRIANGLES_POINT_MODE_CUBIC_ATI);
		glPNTrianglesiATI(GL_PN_TRIANGLES_NORMAL_MODE_ATI, GL_PN_TRIANGLES_NORMAL_MODE_LINEAR_ATI );
		glPNTrianglesiATI(GL_PN_TRIANGLES_TESSELATION_LEVEL_ATI, level);
	}
}

void	RAS_OpenGLRasterizer::SetTexUnit(unsigned int unit)
{
	if (GLEW_ARB_multitexture && unit < m_texunits.size())
	{
		glActiveTextureARB(GL_TEXTURE0 + unit);
		//glClientActiveTexture(GL_TEXTURE0 + unit);
		m_activetexture = unit;
	}
}

void	RAS_OpenGLRasterizer::SetTexCoords(TexCoGen coords)
{
	if (m_activetexture < m_texunits.size())
	{
		if (m_texunits[m_activetexture].texco != coords)
		{
			if (coords == RAS_TEXCO_GEN)
			{
				glEnable(GL_TEXTURE_GEN_S);
				glEnable(GL_TEXTURE_GEN_T);
				glEnable(GL_TEXTURE_GEN_R);
				glEnable(GL_TEXTURE_GEN_Q); 
			}
			else if (m_texunits[m_activetexture].texco == RAS_TEXCO_GEN)
			{
				glDisable(GL_TEXTURE_GEN_S);
				glDisable(GL_TEXTURE_GEN_T);
				glDisable(GL_TEXTURE_GEN_R);
				glDisable(GL_TEXTURE_GEN_Q);
			}
			
			if (coords == RAS_TEXCO_DISABLE)
				glDisable(GL_TEXTURE_2D);
			else
				glEnable(GL_TEXTURE_2D);

			m_texunits[m_activetexture].texco = coords;
			m_texunits[m_activetexture].enable = true;
		}
	}
}

void RAS_OpenGLRasterizer::TexCoord(const RAS_TexVert &tv)
{
	if (!GLEW_ARB_multitexture)
	{
		if (!m_texunits[0].enable)
			return;
		switch (m_texunits[0].texco)
		{
			case RAS_TEXCO_GEN:
				// GL is generating tex cos.
				break;
			case RAS_TEXCO_GLOB: // GL texture matrix transforms to world...
			case RAS_TEXCO_ORCO:
				glTexCoord3fv(tv.getLocalXYZ());
				break;
/*			RAS_TEXCO_OBJECT,
			RAS_TEXCO_LAVECTOR,
			RAS_TEXCO_VIEW,
			RAS_TEXCO_STICKY,
			RAS_TEXCO_WINDOW,
			*/
				// not implemented...
			case RAS_TEXCO_UV1:
				glTexCoord2fv(tv.getUV1());
				break;

			case RAS_TEXCO_DISABLE:
				m_texunits[0].enable = false;
				break;
		}
		return;
	}
	
	for (unsigned int unit = 0; unit < m_texunits.size(); unit++)
	{
		if (!m_texunits[unit].enable)
			break;
		switch (m_texunits[unit].texco)
		{
			case RAS_TEXCO_GEN:
				// GL is generating tex cos.
				break;
			case RAS_TEXCO_GLOB: // GL texture matrix transforms to world...
			case RAS_TEXCO_ORCO:
				glMultiTexCoord3fvARB(GL_TEXTURE0_ARB + unit, tv.getLocalXYZ());
				break;
/*			RAS_TEXCO_OBJECT,
			RAS_TEXCO_LAVECTOR,
			RAS_TEXCO_VIEW,
			RAS_TEXCO_STICKY,
			RAS_TEXCO_WINDOW,
			*/
				// not implemented...
			case RAS_TEXCO_UV1:
				glMultiTexCoord2fvARB(GL_TEXTURE0_ARB + unit, tv.getUV1());
				break;

			case RAS_TEXCO_DISABLE:
				m_texunits[unit].enable = false;
				break;
		}
	}
}

RAS_IRasterizer::ILighting *RAS_OpenGLRasterizer::Lighting()
{
	return m_lighting;
}

RAS_IRasterizer::IBlend *RAS_OpenGLRasterizer::Blender()
{
	return m_blending;
}

RAS_IRasterizer::IProgram *RAS_OpenGLRasterizer::NewProgram(const STR_String &program, const STR_String &name)
{
	if (GLProgram::QueryExtensions(program))
		return new GLProgram(program, name);
	
	return NULL;
}

RAS_IRasterizer::IShader *RAS_OpenGLRasterizer::NewShader()
{
	if (GLSLShader::QueryExtensions())
		return new GLSLShader();
	
	return NULL;
}

unsigned int RAS_OpenGLRasterizer::GetNumTextureUnits()
{
	return GLTexture::GetNumTextureUnits();
}
	
unsigned int RAS_OpenGLRasterizer::GetNumTextureCoords()
{
	return GLTexture::GetNumTextureCoords();
}
	
bool RAS_OpenGLRasterizer::SetTextureCompression(bool compress)
{
	return GLTexture::SetTextureCompression(compress);
}

RAS_IRasterizer::ITexture* RAS_OpenGLRasterizer::NewTexture(int bind)
{
	if (bind)
		return new GLTexture(bind);
	else
		return new GLTexture();
}

bool RAS_OpenGLRasterizer::QueryShadowMaps()
{
	return GLEW_ARB_depth_texture && GLEW_ARB_shadow && (GLEW_ARB_texture_border_clamp || GLEW_VERSION_1_3);
}

void RAS_OpenGLRasterizer::SetPerPixelLights(unsigned int lights)
{
	if (!GLEW_ARB_fragment_program || !GLEW_ARB_vertex_program)
	{
		m_perpixellights = 0;
		return;
	}
	
	GLint maxTexcoUnits;
	glGetIntegerv(GL_MAX_TEXTURE_COORDS_ARB, &maxTexcoUnits);
	if (lights > (maxTexcoUnits - 3))
		lights = maxTexcoUnits - 3;
	m_perpixellights = lights;
}

unsigned int RAS_OpenGLRasterizer::m_perpixellights = 2;
