/**
 * $Id: SG_Tree.cpp 730 2004-07-11 05:10:29Z kester $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): none yet.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 * Bounding Box
 *
 * TODO:  Update() method - update tree when node changes.
 *        Add()    method - add node to tree when object is inserted into scene.
 *        Remove() method - remove node from tree when object is remove from scene.
 */

#include <math.h>

#include <algorithm>
 
#include "SG_BBox.h"
#include "SG_Tree.h"
#include "SG_Node.h"

SG_Tree::SG_Tree()
{
}

SG_Tree::SG_Tree(SG_Tree* left, SG_Tree* right) :
		m_left(left),
		m_right(right),
		m_parent(NULL),
		m_client_object(NULL)
{
	if (m_left)
	{
		m_bbox = m_left->m_bbox;
		m_left->m_parent = this;
	}
	if (m_right)
	{
		m_bbox += m_right->m_bbox;
		m_right->m_parent = this;
	}
	m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
	m_radius = (m_bbox.m_max - m_bbox.m_min).length();
	m_axis = m_bbox.split();
}
	
SG_Tree::SG_Tree(SG_Node* client) :
		m_left(NULL),
		m_right(NULL),
		m_parent(NULL),
		m_client_object(client)
{
	m_bbox = SG_BBox(client->BBox(), client->GetWorldTransform());
	m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
	m_radius = (m_bbox.m_max - m_centre).length();
	m_axis = m_bbox.split();
}

SG_Tree::~SG_Tree() 
{
	delete m_left;
	delete m_right;
}
	
MT_Scalar SG_Tree::volume() const
{
	return m_bbox.volume();
}
	
void SG_Tree::dump() const
{
	if (m_left)
		m_left->dump();
	if (m_client_object)
		std::cout << m_client_object << std::endl;
	else
		std::cout << this << " ";
	if (m_right)
		m_right->dump();
}

SG_Tree* SG_Tree::Left() const
{
	return m_left;
}

SG_Tree* SG_Tree::Right() const
{
	return m_right;
}

SG_Node* SG_Tree::Client() const
{
	return m_client_object;
}

void SG_Tree::Resize()
{
	bool setbbox = false;
	if (m_client_object)
	{
		m_bbox = SG_BBox(m_client_object->BBox(), m_client_object->GetWorldTransform());
		setbbox = true;
	}
	
	if (m_left)
	{
		if (setbbox)
			m_bbox += m_left->m_bbox;
		else
		{
			m_bbox = m_left->m_bbox;
			setbbox = true;
		}
	}
	
	if (m_right)
	{
		if (setbbox)
			m_bbox += m_right->m_bbox;
		else
			m_bbox = m_right->m_bbox;
	}
	
	m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
	m_radius = (m_bbox.m_max - m_centre).length();
	m_axis = m_bbox.split();
}

void SG_Tree::TreeResize()
{
	if (m_left)
		m_left->TreeResize();
	if (m_right)
		m_right->TreeResize();
	Resize();
}

void SG_Tree::RemoveLeft()
{
	if (m_left)
	{
		delete m_left;
		m_left = NULL;
	}
}

void SG_Tree::RemoveRight()
{
	if (m_right)
	{
		delete m_right;
		m_right = NULL;
	}
}

bool SG_Tree::RemoveNode(SG_Node *node)
{
	if (m_client_object == node)
	{
		m_client_object = NULL;
		if (!m_left && !m_right)
		{
			if (m_parent)
			{
				if (m_parent->m_left == this)
					m_parent->RemoveLeft();
				else
					m_parent->RemoveRight();
			}
		}
		return true;
	}
	
	SG_Tree *left = m_left, *right = m_right;
	
	if (left && right)
	{
		if (node->GetWorldPosition()[m_axis] > m_centre[m_axis])
			std::swap(left, right);
	}
	
	if (left)
	{
		if (left->RemoveNode(node))
			return true;
	}
	
	if (right)
	{
		if (right->RemoveNode(node))
			return true;
	}
	
	return false;
}

void SG_Tree::AddNode(SG_Node *node)
{
	if (node->GetWorldPosition()[m_axis] < m_centre[m_axis])
	{
		if (m_left)
		{
			m_left->AddNode(node);
			m_bbox += m_left->m_bbox;
			m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
			m_radius = (m_bbox.m_max - m_centre).length();
			m_axis = m_bbox.split();
		} else
			SetLeft(new SG_Tree(node));
	} else {
		if (m_right)
		{
			m_right->AddNode(node);
			m_bbox += m_right->m_bbox;
			m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
			m_radius = (m_bbox.m_max - m_centre).length();
			m_axis = m_bbox.split();
		}
		else
			SetRight(new SG_Tree(node));
	}
}

SG_Tree* SG_Tree::Find(SG_Node *node)
{
	if (m_client_object == node)
		return this;
	
	SG_Tree *left = m_left, *right = m_right;
	
	if (left && right)
	{
		if (right->m_bbox.intersects(node->BBox()))
			std::swap(left, right);
	}
		
	if (left)
	{
		SG_Tree* ret = left->Find(node);
		if (ret) return ret;
	}
	
	if (right)
	{
		SG_Tree* ret = right->Find(node);
		if (ret) return ret;
	}
	
	return NULL;
}

void SG_Tree::get(MT_Point3 *box) const
{
	MT_Transform identity;
	identity.setIdentity();
	m_bbox.get(box, identity);
}

bool SG_Tree::inside(const MT_Point3 &point) const
{
	return m_bbox.inside(point);
}

const SG_BBox& SG_Tree::BBox() const
{
	return m_bbox;
}

void SG_Tree::SetLeft(SG_Tree *left)
{
	m_left = left;
	if (left)
	{
		m_bbox += left->m_bbox;
		m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
		m_radius = (m_bbox.m_max - m_centre).length();
		m_axis = m_bbox.split();
		m_left->m_parent = this;
	}
}

void SG_Tree::SetRight(SG_Tree *right)
{
	m_right = right;
	if (right)
	{
		m_bbox += right->m_bbox;
		m_centre = (m_bbox.m_min + m_bbox.m_max)/2.0;
		m_radius = (m_bbox.m_max - m_centre).length();
		m_axis = m_bbox.split();
		m_right->m_parent = this;
	}
}

/**
 * A Half array is a square 2d array where cell(x, y) is undefined
 * if x < y.
 */
template<typename T>
class HalfArray
{
	std::vector<std::vector<T> > m_array;
public:
	HalfArray() {}
	~HalfArray() {}
	
	void resize(unsigned int size)
	{
		m_array.resize(size);
		for( unsigned int i = 0; i < size; i++)
		{
			m_array[i].resize(size - i);
		}
	}
	
	T& operator() (unsigned int x, unsigned int y)
	{
		assert(x >= y);
		return m_array[y][x - y];
	}
	
	void erase_column (unsigned int x)
	{
		for (unsigned int y = 0; y <= x; y++)
			m_array[y].erase(m_array[y].begin() + x - y);
	}

	void delete_column (unsigned int x)
	{
		for (unsigned int y = 0; y < x; y++)
		{
			delete m_array[y][x - y];
			m_array[y].erase(m_array[y].begin() + x - y);
		}
	}
	
	void erase_row (unsigned int y)
	{
		m_array.erase(m_array.begin() + y);
	}
};

SG_TreeFactory::SG_TreeFactory()
{
}

SG_TreeFactory::~SG_TreeFactory()
{
}
	
void SG_TreeFactory::Add(SG_Node* client)
{
	if (client)
		m_objects.push_back(new SG_Tree(client));
}

void SG_TreeFactory::Add(SG_Tree* tree)
{
	m_objects.push_back(tree);
}

static bool sortx(const SG_Tree *a, const SG_Tree *b)
{
	return a->Centre()[0] < b->Centre()[0];
}
static bool sorty(const SG_Tree *a, const SG_Tree *b)
{
	return a->Centre()[1] < b->Centre()[1];
}
static bool sortz(const SG_Tree *a, const SG_Tree *b)
{
	return a->Centre()[2] < b->Centre()[2];
}

static bool sortsize(const SG_Tree *a, const SG_Tree *b)
{
	return a->volume() > b->volume();
}

SG_Tree* SG_TreeFactory::MakeTreeDown()
{
	if (m_objects.size() == 0)
		return NULL;
	if (m_objects.size() == 1)
		return *m_objects.begin();
		
	TreeSet::iterator it = m_objects.begin();
	SG_BBox bbox((*it)->BBox());
	for (++it; it != m_objects.end(); ++it)
		bbox += (*it)->BBox();
		
	//std::sort(m_objects.begin(), m_objects.end(), sortsize);
	//SG_Tree *root = *m_objects.begin();
	
	switch (bbox.split())
	{
		case 0: // X axis
			std::sort(m_objects.begin(), m_objects.end(), sortx);
			break;
		case 1: // Y axis
			std::sort(m_objects.begin(), m_objects.end(), sorty);
			break;
		case 2: // Z axis
			std::sort(m_objects.begin(), m_objects.end(), sortz);
			break;
	}
	
	unsigned int half = m_objects.size() / 2;
	unsigned int count = 0;
	SG_TreeFactory lefttree, righttree;
	for (it = m_objects.begin(); it != m_objects.end(); ++it)
	{
		//if ((*it) == root)
		//	continue;
		if (count < half)
			lefttree.Add(*it);
		else
			righttree.Add(*it);
		count++;
	}
	
	//std::cout << "MakeTreeDown: Left: " << half << ", Right: " << m_objects.size() - half << std::endl;
	
	//root->SetLeft(lefttree.MakeTreeDown());
	//root->SetRight(righttree.MakeTreeDown());
	
	//return root;
	return new SG_Tree(lefttree.MakeTreeDown(), righttree.MakeTreeDown());
}

SG_Tree* SG_TreeFactory::MakeTree()
{
	return MakeTreeDown();
}

