/**
* $Id: OdeMesh.cpp 133 2003-10-05 02:56:47Z kester $
*
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): Kester.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
*/
/*
 * OdeMesh: Converts a Blender mesh to an ODE trimesh for collision detection.
 * This class makes a copy of the mesh data; ideally it should use the same
 * data as the renderer. However, the renderer splits the mesh by material, which
 * is not useful for collision detection.
 *
 * Therefore, the next step is to modify the renderer:
 *	Vertex array (all materials)
 *	Index array (per material)
 *	Index array (collision det.)
*/
#include <math.h>
#include <ode/ode.h>

#include "RAS_TexVert.h"
#include "RAS_MeshObject.h"

#include "OdeMesh.h"

ODEMesh::ODEMesh(dSpaceID space, ODEPhysicsController *ctrl, KX_ICollisionSensor *sens ) :
	ODESensor(space, ctrl, sens),
	m_world(NULL),
	m_tris(NULL)
{
	m_recurse = false;
	m_scaling[0] = m_scaling[1] = m_scaling[2] = 1.0;
}

ODEMesh::~ODEMesh()
{
	// TODO: cleanup
	if (m_geom)
	{
		dSpaceRemove(m_world, m_geom);
		dGeomDestroy (m_geom);
		m_geom = NULL;
	}
	if (m_tris)
	{
		dGeomTriMeshDataDestroy(m_tris);
		m_tris = NULL;
	}
	if (m_world) {
		dSpaceDestroy(m_world);
		m_world = NULL;
	}
}

void ODEMesh::Trimesh(RAS_MeshObject *mesh)
{
	int numvalidpolys = 0;
	std::vector<RAS_TexVert *> tempvarray;

	vertexarray.clear();
	indexarray.clear();
	for (std::vector<RAS_TriangleIndex>::iterator tri = mesh->m_triangle_indices.begin();tri != mesh->m_triangle_indices.end();++tri)
	{
		RAS_TriangleIndex  odeidx;

		// only add polygons that have the collisionflag set
		if ((*tri).m_collider)
		{
			for (int v=0;v<3;v++)
			{
				int num = mesh->m_xyz_index_to_vertex_index_mapping[(*tri).m_index[v]].size();
				if (num != 1)
				{
					int i=0;
				}
				RAS_MatArrayIndex& vertindex = mesh->m_xyz_index_to_vertex_index_mapping[(*tri).m_index[v]][0];

				numvalidpolys++;

				RAS_TexVert *texvert = mesh->GetVertex(vertindex.m_array,
					vertindex.m_index,
					(RAS_IPolyMaterial*)vertindex.m_matid);
				int vit;
				for (vit = 0; vit < tempvarray.size(); vit++)
				{
					if (tempvarray[vit] == texvert)
					{
						odeidx.m_index[v] = vit;
						break;
					}
				}

				if (vit == tempvarray.size())
				{
						// Not found
						tempvarray.push_back(texvert);
						odeidx.m_index[v] = vit;
				}
			}
			indexarray.push_back(odeidx);
		}
	}

	vertexarray.resize(tempvarray.size());
	for (int vit = 0; vit < tempvarray.size(); vit++)
	{
		MT_Point3 pt = tempvarray[vit]->xyz();

		vertexarray[vit].pt[0] = pt[0]*m_scaling[0];
		vertexarray[vit].pt[1] = pt[1]*m_scaling[1];
		vertexarray[vit].pt[2] = pt[2]*m_scaling[2];
	}
	if (!m_world)
		m_world = dSimpleSpaceCreate(m_space);
	dTriMeshDataID tris;
	dGeomID geom;

	int numindices = indexarray.size();
	int numverts = vertexarray.size();

	if (!numindices)
		return;

	tris = dGeomTriMeshDataCreate();
	dGeomTriMeshDataBuildSingle(tris, &(vertexarray[0].pt[0]), sizeof(struct vertex), numverts, &(indexarray[0].m_index[0]), numindices*3, sizeof(RAS_TriangleIndex));
	geom = dCreateTriMesh(m_world, tris, NULL, NULL, NULL);
	dGeomSetData(geom, this);

	if (m_geom)
	{
		dSpaceRemove(m_world, m_geom);
		dGeomDestroy (m_geom);
	}
	if (m_tris)
		dGeomTriMeshDataDestroy(m_tris);

	m_geom = geom;
	m_tris = tris;
#if 0
// Useful debugging code
	printf("--Vertices--\n");
	for (int vit = 0; vit < numverts; vit++)
	{
		printf("{ %0.02f, %0.02f, %0.02f }\n", vertexarray[vit].pt[0], vertexarray[vit].pt[1], vertexarray[vit].pt[2]);
	}
	printf("\n--Indices--\n");
	for (int iit = 0; iit < numindices; iit++)
	{
		printf("{ %2d, %2d, %2d }\n", indexarray[iit].m_index[0], indexarray[iit].m_index[1],indexarray[iit].m_index[2]);
	}
	printf("\n");
#endif
}


void ODEMesh::HandleCollision(ODESensor *obj2)
{
	// Can't sense other sensors
	if (!obj2->getBody())
		return;

	if (m_sens)
		m_sens->HandleCollision(this, obj2);
}

void ODEMesh::setScaling(dVector3 s)
{
	if (m_geom)
	{
		// Rescale
		dVector3 rescale = {s[0]/m_scaling[0], s[1]/m_scaling[1], s[2]/m_scaling[2], 0.0};
		for (std::vector<struct vertex>::iterator it = vertexarray.begin(); it != vertexarray.end(); ++it)
		{
			(*it).pt[0] *= rescale[0];
			(*it).pt[1] *= rescale[1];
			(*it).pt[2] *= rescale[2];
		}
		// Might have to notify ODE that the geometry changed here
	}

	m_scaling[0] = s[0];
	m_scaling[1] = s[1];
	m_scaling[2] = s[2];
}

void ODEMesh::setSpace(dSpaceID s)
{
	dSpaceRemove(m_space, (dGeomID) m_world);
	dSpaceAdd(m_space = s, (dGeomID) m_world);
}


