/**
* $Id: OdeSensor.cpp 133 2003-10-05 02:56:47Z kester $
*
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): kester.
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
*/
/*
 * ODE Sensor class.
 * 
 * This class comprises the code to pass a collision from ODE to a sensor
 * logic brick.
 *
*/
#include <math.h>

#include "OdeSensor.h"

ODESensor::ODESensor(dSpaceID space, ODEPhysicsController *ctrl, KX_ICollisionSensor *sens) :
	m_ctrl(ctrl),
	m_sens(sens),
	m_geom(NULL),
	m_space(space),
	m_body(false),
	m_recurse(false),
	m_debug(0)
{
}

ODESensor::~ODESensor()
{
	if (m_geom)
	{
		dSpaceRemove(m_space, m_geom);
		dGeomDestroy (m_geom);
		m_geom = NULL;
		m_debug--;
	}
	//printf("ODESensor::~ODESensor(%x): %d\n", (unsigned) this, m_debug);
}

dGeomID ODESensor::SphereSensor(float radius)
{
	if (m_geom)
	{
		dSpaceRemove(m_space, m_geom);
		dGeomDestroy (m_geom);
		m_debug--;
	}
	m_geom = dCreateSphere (m_space, radius);
	dGeomSetData(m_geom,this);
	m_debug++;
	return m_geom;
}

void ODESensor::setRadius(float radius)
{
	if (dGeomGetClass(m_geom) == dSphereClass)
		dGeomSphereSetRadius(m_geom, radius);
}

dGeomID ODESensor::CCylinderSensor(float radius, float length)
{
	if (m_geom)
	{
		dSpaceRemove(m_space, m_geom);
		dGeomDestroy (m_geom);
		m_debug--;
	}
	if (length < 0.)
		length = 0.;
	m_geom = dCreateCCylinder(m_space, radius, length);
	dGeomSetData(m_geom,this);
	m_debug++;
	//printf("ODESensor::BoxSensor: %x\n", (unsigned) m_geom);
	return m_geom;
}

dGeomID ODESensor::BoxSensor(float dx, float dy, float dz)
{
	if (m_geom)
	{
		dSpaceRemove(m_space, m_geom);
		dGeomDestroy (m_geom);
		m_debug--;
	}
	m_geom = dCreateBox(m_space, dx, dy, dz);
	dGeomSetData(m_geom,this);
	m_debug++;
	//printf("ODESensor::BoxSensor: %x\n", (unsigned) m_geom);
	return m_geom;
}

dGeomID ODESensor::RaySensor(int axis, float length)
{
	dVector3 dir = {0.f, 0.f, 0.f, 0.f};
	if (m_geom)
	{
		dSpaceRemove(m_space, m_geom);
		dGeomDestroy (m_geom);
		m_debug--;
	}
	//ODE INTERNAL ERROR 2: GeomTransform encapsulated object must not be in a space in dCollideTransform()
	dGeomID ray = dCreateRay(0, length);
	m_geom = dCreateGeomTransform(m_space);
	dGeomTransformSetGeom(m_geom, ray);
	dGeomTransformSetCleanup(m_geom, 1); // Delete ray when delete m_geom
	dGeomTransformSetInfo(m_geom, 1); // Transform object will be returned by collision functions
	switch( axis ){
		case 0:
			dir[1] = 1.f;
			break;
		case 1:
			dir[0] = 1.f;
			break;
		case 2:
			dir[2] = 1.f;
			break;
		case 3:
			dir[0] = -1.f;
			break;
		case 4:
			dir[1] = -1.f;
			break;
		case 5:
			dir[2] = -1.f;
			break;

		default:
			printf("ODESensor::RaySensor: Unknown axis %d [%f, %f, %f] - keep it 3d please\n", axis,
			 dir[0], dir[1], dir[2]);
			break;
	}
	dGeomRaySet(ray, 0.f, 0.f, 0.f, dir[0], dir[1], dir[2]);
	dGeomSetData(m_geom,this);
	dGeomSetData(ray,this);
	m_debug++;
	return m_geom;
}

void ODESensor::setBodyID(dBodyID body)
{
	dGeomSetBody(m_geom, body);
}

void ODESensor::HandleCollision(ODESensor *obj2)
{
	/*if (m_debug <= 0)
		printf("ODESensor::HandleCollision: m_debug = %d\n", m_debug);*/
	// Can't sense other sensors
	if (!obj2->getBody())
		return;


	if (m_sens) {
		/*if (m_debug)
		{
			const dReal*	pos = dGeomGetPosition(m_geom);
			printf("ODESensor::HandleCollision %x %x (%f, %f, %f)\n", (unsigned) this, (unsigned) obj2,
				pos[0], pos[1], pos[2]);
		}*/
		if (dGeomGetClass(m_geom) == dGeomTransformClass)
		{
			dGeomID ray = dGeomTransformGetGeom(m_geom);
			if (dGeomGetClass(ray) == dRayClass)
			{
				dContactGeom contactGeom[32];			// up to 4 contacts per box
				m_rayDepth = 10000000.0;
				int numc = dCollide (m_geom,obj2->m_geom,32,contactGeom,sizeof(dContactGeom));
				if (!numc)
					return;
				for (int i = 0; i < numc; i++)
				{
					if ( fabs(contactGeom[i].depth) < m_rayDepth)
						m_rayDepth = fabs(contactGeom[i].depth);
				}
			}
		}


		m_sens->HandleCollision(this, obj2);
	}

}

void ODESensor::setPosition(float x, float y, float z)
{
	//if (m_debug <= 0) printf("ODESensor(%x)::setPosition: m_debug = %d\n",(unsigned)this, m_debug);
	if (m_geom) dGeomSetPosition (m_geom, x, y, z);
}

void ODESensor::setRotation(dMatrix3 &R)
{
	if (m_geom) dGeomSetRotation (m_geom, R);
}

void ODESensor::setQuaternion(dQuaternion &q)
{
	if (m_geom) dGeomSetQuaternion(m_geom, q);
}

void ODESensor::setSpace(dSpaceID s)
{
	if (m_geom)
	{
		if (m_space) dSpaceRemove(m_space, m_geom);
		dSpaceAdd(s, m_geom);
	}
	m_space = s;
}
