#ifdef WIN32
#include "BLI_winstuff.h"
#endif

#include "MEM_guardedalloc.h"
#include "BIF_meshHE.h"
#include "BIF_gl.h"

#include "BKE_hemesh.h" 
#include "BKE_mesh.h"
#include "BKE_dmesh.h"
#include "BKE_global.h"
#include "BKE_main.h"
#include "BKE_utildefines.h"

#include "DNA_hemesh_types.h"
/*these next two are for #defines. . .*/
#include "DNA_meshdata_types.h"
#include "DNA_mesh_types.h"
#include "DNA_material_types.h"
#include "DNA_object_types.h"
#include "DNA_scene_types.h"

#include "BLI_parray.h"
#include "BLI_arithb.h"
#include "math.h"

static void draw_dmesh_solid(DMesh *dm);
static void draw_dmesh_solid_ext(DMesh *dm);
static void draw_dmesh_wire(DMesh *dm);
static void draw_dmesh_shaded(DMesh *dm);

DVertIndiceArray *DVertIndiceArray_New(void)
{
	return MEM_callocN(sizeof(DVertIndiceArray), "DVertIndiceArray");
}


DVertIndiceArray *DVertIndiceArray_Copy(DVertIndiceArray *array)
{
	DVertIndiceArray *arr = MEM_dupallocN(array);
	arr->array = MEM_dupallocN(array->array);
	return arr;
}


void DVertIndiceArray_Free(DVertIndiceArray *arr)
{
	if (arr->array) MEM_freeN(arr->array);
	MEM_freeN(arr);
}


void DVertIndiceArray_AddVerts(DVertIndiceArray *array, int *verts, int len)
{
	int i;
	if ((array->cursor+len)>=array->len)		  /*realloc*/
	{
		int *new_array = MEM_callocN(sizeof(int)*(array->len+DVERT_INCR), "increment");
		if (array->array)
		{
			for (i=0; i<array->cursor; i++) new_array[i] = array->array[i];
			MEM_freeN(array->array);
		}
		array->len += DVERT_INCR;
		array->array = new_array;
	}
	for (i=0; i<len; i++)
	{
		array->array[array->cursor] = verts[i];
		array->cursor++;
	}
}


DMesh *copy_DMesh(DMesh *olddm)
{
	DMesh *newm = MEM_dupallocN(olddm);
	if (olddm->verts) newm->verts = MEM_dupallocN(olddm->verts);
	if (olddm->edges) newm->edges = MEM_dupallocN(olddm->edges);
	if (olddm->faces) newm->faces = MEM_dupallocN(olddm->faces);
	if (olddm->array) newm->array = DVertIndiceArray_Copy(olddm->array);
	return newm;
}


void dface_AddVerts(DMesh *dm, DFace *df, int *verts, int vlen)
{
	if (!dm->array) dm->array = DVertIndiceArray_New();
	df->vlen = vlen;
	df->verts = dm->array->cursor;
	DVertIndiceArray_AddVerts(dm->array, verts, vlen);
}


void dface_AddCols(DMesh *dm, DFace *df, int *cols)
{
	if (!dm->array) dm->array = DVertIndiceArray_New();
	df->cols = dm->array->cursor;
	DVertIndiceArray_AddVerts(dm->array, cols, df->vlen);
}


void free_DMesh(DMesh *dm)
{
	if (!dm) return;
	if (dm->verts) MEM_freeN(dm->verts);
	if (dm->edges) MEM_freeN(dm->edges);
	if (dm->faces) MEM_freeN(dm->faces);
	if (dm->array) DVertIndiceArray_Free(dm->array);
	if (dm->gdl_solid) glDeleteLists(dm->gdl_solid, 1);
	MEM_freeN(dm);
}


static void dmesh_do_material(DMesh *dm, int mat)
{
	Material *ma;
	float diff[4], spec[4];
	if (dm->mat) ma = dm->mat[mat];
	else ma = G.main->mat.first;
	if (!ma) return;

	diff[0] = (ma->ref+ma->emit)*ma->r;
	diff[1] = (ma->ref+ma->emit)*ma->g;
	diff[2] = (ma->ref+ma->emit)*ma->b;
	diff[3] = 1.0;

	spec[0] = ma->spec*ma->specr;
	spec[1] = ma->spec*ma->specr;
	spec[2] = ma->spec*ma->specr;
	spec[3] = 1.0;

	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diff);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, spec);
}


static void draw_dmesh_shaded(DMesh *dm)
{
	DFace *df;
	int *vert, *cols;
	int i, v;

	if (!dm->cols)
	{
		draw_dmesh_solid(dm);
		return;
	}

	df = dm->faces;
	for (i=0; i<dm->totface; i++, df++)
	{
		if (df->f & ME_SMOOTH) glShadeModel (GL_SMOOTH);
		else glShadeModel (GL_FLAT);
		dmesh_do_material(dm, df->mat);

		glNormal3fv(df->no);
		glBegin(GL_POLYGON);
		vert = &dm->array->array[df->verts];
		cols = &dm->array->array[df->cols];
		for (v=0; v<df->vlen; v++, vert++)
		{
			glColor3fv(dm->cols[*cols].col);
			if (df->f & ME_SMOOTH) glNormal3fv(dm->verts[*vert].no);
			glVertex3fv(dm->verts[*vert].co);
		}
		glEnd();
	}
}


static void draw_dmesh_solid_ext(DMesh *dm)
{
	DFace *df;
	DVert *dv;
	int *vert;
	int i, v;
	
	glEnable(GL_LIGHTING);
	df = dm->faces;
	for (i=0; i<dm->totface; i++, df++)
	{
		if (df->f & ME_SMOOTH) glShadeModel (GL_SMOOTH);
		else glShadeModel (GL_FLAT);
		dmesh_do_material(dm, df->mat);

		glNormal3fv(df->no);
		glBegin(GL_POLYGON);
		vert = &dm->array->array[df->verts];
		for (v=0; v<df->vlen; v++, vert++)
		{
			if (df->f & ME_SMOOTH) glNormal3fv(dm->verts[*vert].no);
			dv = &dm->verts[*vert];
			glVertex3f(dv->co[0], dv->co[1], dv->co[2]);
		}
		glEnd();
	}
}


static void draw_dmesh_solid(DMesh *dm)
{
	if (!dm->gdl_solid)
	{
		glPushMatrix();
		glLoadIdentity();
		dm->gdl_solid = GDL_NewDispList();
		draw_dmesh_solid_ext(dm);
		GDL_EndDispList();
		glPopMatrix();
	}
	GDL_CallList(dm->gdl_solid);
//draw_dmesh_solid_ext(dm);
}


static void draw_dmesh_wire(DMesh *dm)
{
	DEdge *ded;
	int i;

	if (!dm->edges) return;
	ded = dm->edges;
	glBegin(GL_LINES);
	for (i=0; i<dm->totedge; i++, ded++)
	{
		glVertex3fv(dm->verts[ded->v1].co);
		glVertex3fv(dm->verts[ded->v2].co);
	}
	glEnd();
}


void draw_DMesh(DMesh *dm, int dt)
{
	if (dt >= OB_SHADED) draw_dmesh_shaded(dm);
	else if (dt >= OB_SOLID) draw_dmesh_solid(dm);
	else if (dt >= OB_WIRE) draw_dmesh_wire(dm);
}


void add_herefs_to_DMesh(DMesh *dm, HE_EditMesh *em)
{
	HE_EditEdgeInfo *ieed;
	HE_EditFace *efa;
	HE_EditVert *eve;
	DVert *dev;
	DFace *def;
	DEdge *ded;

	if (!em) return;

	dev = dm->verts;
	for (eve=em->verts.first; eve; dev++, eve=eve->next)
	{
		VECCOPY(eve->ssco, dev->co);
	}

	def = dm->faces;
	for (efa=em->faces.first; efa; def++, efa=efa->next)
	{
		def->ref = (void*) efa;
	}

	ded = dm->edges;
	for (ieed=em->einfo.first; ieed; ded++, ieed=ieed->next)
	{
		ded->ref = (void*) ieed;
	}
}


DMesh *make_DMesh_from_HEMesh(HE_Mesh *mesh)
{
	HE_Face *mef;
	HE_Vert *v1;
	HE_Edge *edge;
	DMesh *dmesh = MEM_callocN(sizeof(DMesh), "DMesh");
	int i, flist[6000], flen;

	dmesh->faces = MEM_mallocN(mesh->totfaces*sizeof(DFace), "dmesh->faces");
	dmesh->edges = MEM_mallocN(mesh->totiedges*sizeof(DEdge), "dmesh->edes");
	dmesh->verts = MEM_mallocN(mesh->totverts*sizeof(DVert), "dmesh->verts");
	dmesh->array = DVertIndiceArray_New();
	dmesh->mat = mesh->mat;
/*we just copy pointer, we don't copy memory for
							  material array.*/
	dmesh->totcol = mesh->totcol;

	dmesh->totface = mesh->totfaces;
	dmesh->totvert = mesh->totverts;
	dmesh->totedge = mesh->totiedges;

	for (i=0; i<mesh->totverts; i++)
	{
		v1 = &mesh->verts[i];
		VECCOPY(dmesh->verts[i].co, v1->co);
		VECCOPY(dmesh->verts[i].no, v1->no);
		dmesh->verts[i].f = v1->f;
	}

	for (i=0; i<mesh->totiedges; i++)
	{
		edge = &mesh->half_edges[mesh->iedges[i].e1];
		dmesh->edges[i].v1 = edge->v1;
		dmesh->edges[i].v2 = edge->v2;
		dmesh->edges[i].crease = mesh->iedges[i].crease;
	}

	for (i=0; i<mesh->totfaces; i++)
	{
		mef = &mesh->faces[i];
		edge = &mesh->half_edges[mef->edge];
		flen = 0;
		do
		{
			flist[flen] = edge->v1;
			flen++;
			edge = &mesh->half_edges[edge->next];
		} while (edge != &mesh->half_edges[mef->edge]);
		dmesh->faces[i].verts = dmesh->array->cursor;
		DVertIndiceArray_AddVerts(dmesh->array, flist, flen);
		dmesh->faces[i].vlen = flen;
		dmesh->faces[i].mat = mef->mat_index;
		if (mef->f & HEM_SMOOTH) dmesh->faces[i].f = ME_SMOOTH;
		else dmesh->faces[i].f = 0;
		VECCOPY(dmesh->faces[i].no, mef->nor);
	}

	return dmesh;
}


/* the rules of faces-to-MFaces for HE MEsh:
	1. Four-vert quads turn into one face.
	2. Five-vert ngons turn into three triangles.
	3. More then five-vert ngons form a center-vert
	   and makes triangles around that.*/
/*face_nors is pointer to pointer to face normal array*/
Mesh *DMesh_to_Mesh(DMesh *oldme, int do_libblock, float **face_nors)
{
	PointerArray varray = {NULL, 0};
	PointerArray farray = {NULL, 0};
	PointerArray nfarray = {NULL, 0};
	DVert *dev;
	DFace *def;
	DEdge *ded;
	Mesh *me;
	MFace *mef;
	MVert *v1;
	int i, j, f, totv = 0, totf = 0, len, mat_nr;
	int *verts;
	float cent[3], *nor;
	unsigned int vlist[6000];
#if 0
	if (do_libblock) me = add_mesh();
	else me = MEM_callocN(sizeof(Mesh), "Mesh from hemesh");

	totv = oldme->totvert;

	totv = 0;
	totf = 0;
	for (i=0; i<oldme->totvert; i++)
	{
		v1 = MEM_callocN(sizeof(MVert), "MVert from hevert");
		VECCOPY(v1->co, oldme->verts[i].co);
		VECCOPY(v1->no, oldme->verts[i].no);
		PA_AddToArray(&varray, v1, i);
		totv++;
	}

	for (i=0; i<oldme->totface; i++)
	{
		def = &oldme->faces[i];
		mat_nr = def->mat;
		f = def->f;
		nor = def->no;

		verts = &oldme->array->array[def->verts];
		len = 0;
		cent[0] = cent[1] = cent[2] = 0;
		for (j=0; j<def->vlen; j++, verts++)
		{
			dev = &oldme->verts[*verts];
			vlist[len] = *verts;
			VecAddf(cent, cent, dev->co);
			len++;
		}
		VecMulf(cent, 1.0f/((float)len));
		if (len==3)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			mef->flag = f;
			test_index_mface(mef, 3);
			totf++;
		}
		else if (len==4)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = vlist[3];
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 4);
			mef->flag = f;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

		}
		else if (len==5)
		{
			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[0];
			mef->v2 = vlist[1];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			mef->flag = f;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[2];
			mef->v2 = vlist[3];
			mef->v3 = vlist[4];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			mef->flag = f;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;

			mef = MEM_callocN(sizeof(MFace), "MFace from he face");
			mef->v1 = vlist[4];
			mef->v2 = vlist[0];
			mef->v3 = vlist[2];
			mef->v4 = 0;
			mef->mat_nr = mat_nr;

			test_index_mface(mef, 3);
			mef->flag = f;
			PA_AddToArray(&farray, mef, totf);
			PA_AddToArray(&nfarray, nor, totf);
			totf++;
		}										  /*make a tri fan*/
		else
		{
			v1 = MEM_callocN(sizeof(MVert), "MVertcenterHEf");
			VECCOPY(v1->co, cent);
			VECCOPY(v1->no, def->no);
			PA_AddToArray(&varray, v1, totv);
			totv++;

			for (j=0; j<len; j++)
			{
				mef = MEM_callocN(sizeof(MFace), "MFace from he face");
				mef->v1 = vlist[j];
				if (j != len-1) mef->v2 = vlist[j+1];
				else mef->v2 = vlist[0];
				mef->v3 = totv - 1;
				mef->v4 = 0;
				mef->mat_nr = mat_nr;

				test_index_mface(mef, 3);
				mef->flag = f;
				PA_AddToArray(&farray, mef, totf);
				PA_AddToArray(&nfarray, nor, totf);
				totf++;
			}
		}
	}

	me->mvert = MEM_callocN((totv+1)*sizeof(MVert), "MVert from verts");
	me->mface = MEM_callocN((totf+1)*sizeof(MFace), "MFace from faces");
	me->totvert = totv;
	me->totface = totf;
	if (face_nors) *face_nors = MEM_callocN(sizeof(float*)*(totf*3), "face_nors in hemesh.c");

	memset(me->mvert, 0, sizeof(me->mvert));
	for (i=0; i<totv; i++)
	{
		memcpy(&me->mvert[i], varray.array[i], sizeof(MVert));
		MEM_freeN(varray.array[i]);
	}

	mef = me->mface;
	for (i=0; i<totf; i++)
	{
		memcpy(&mef[i], farray.array[i], sizeof(MFace));
		if (face_nors)
		{
			float *fptr = *face_nors;
			VecCopyf(&fptr[i*3], ((float*)nfarray.array[i]));
		}
		MEM_freeN(farray.array[i]);
	}

	PA_FreeArray(&farray);
	PA_FreeArray(&varray);
	PA_FreeArray(&nfarray);

/*copy materials*/
	me->totcol = oldme->totcol;
	if (do_libblock)
	{
		me->mat = MEM_callocN(sizeof(Material*)*oldme->totcol, "material array");
		for (i=0; i<oldme->totcol; i++)
		{
			me->mat[i] = oldme->mat[i];
			id_us_plus((ID*)me->mat[i]);
		}
	} else me->mat = MEM_dupallocN(oldme->mat);

	tex_space_mesh(me);
#endif
	return me;
}


void dmesh_calcFaceNormals(DMesh *dm)
{
	DFace *df;
	DVert *v1, *v2;
	int i, j, *verts, count;
	float cent[3];

	df = dm->faces;
	for (i=0; i<dm->totface; i++, df++)
	{
		if (df->vlen < 3) continue;
		/*we need cent for normal calc of ngons*/
		cent[0] = cent[1] = cent[2] = 0;
		count = 0;
		verts = &dm->array->array[df->verts];
		for (j=0; j<df->vlen; j++, verts++)
		{
			v1 = &dm->verts[*verts];
			VecAddf(cent, cent, v1->co);
			count++;
		}
		VecMulf(cent, 1.0f/((float)count));
		verts = &dm->array->array[df->verts];
		v1 = &dm->verts[verts[0]];
		v2 = &dm->verts[verts[1]];
		CalcNormFloat(v1->co, v2->co, cent, df->no);
	}
}


void dmesh_calc_vertnormals(DMesh *dm)
{
	DFace *df;
	DVert *dv;
	int i, j, *verts;

	dv = dm->verts;
/*clear ->f1 counter. . .*/
	for (i=0; i<dm->totvert; i++, dv++)
	{
		dv->f1 = 0;
		dv->no[0] = dv->no[1] = dv->no[2] = 0;
	}

	df = dm->faces;
	for (i=0; i<dm->totface; i++, df++)
	{
		if (!(df->f & ME_SMOOTH)) continue;
		verts = &dm->array->array[df->verts];
		for (j=0; j<df->vlen; j++)
		{
			dv = &dm->verts[verts[j]];
			VecAddf(dv->no, dv->no, df->no);
			dv->f1++;
		}
	}

	dv = dm->verts;
	for (i=0; i<dm->totvert; i++, dv++)
	{
		if (dv->f1)
		{
			VecMulf(dv->no, 1.0f/((float)dv->f1));
		}
		else
		{
/*no bordering faces: well then, make it point up*/
			dv->no[0] = 0;
			dv->no[1] = 0;
			dv->no[2] = 1;
		}
	}
}


/****************opengl display list code*********/
/*code is here to prevent merging woes*/
/*module prefix: GDL_*/
typedef struct GDL_DList
{
	struct GDL_DList *next, *prev;
	int id;
} GDL_DList;

static ListBase opengl_displists = {NULL, NULL};

int GDL_NewDispList(void)
{
	int i=1;
/*GDL_DList *dl = opengl_displists.first;
while (dl) {
	if (dl->prev) {
		if (dl->id-dl->prev->id > 1) {
			i = dl->prev->id+1;
			break;
		}
	}
	i = dl->id+1;
	dl=dl->next;
}
dl = MEM_callocN(sizeof(GDL_DList), "GDL_DList");
dl->id = i;
BLI_addtail(&opengl_displists, dl);
glNewList(dl->id, GL_COMPILE);*/
	i = glGenLists(1);
	glNewList(i, GL_COMPILE);

	return i;
}


int GDL_EndDispList(void)
{
	glEndList();
}


void GDL_DeleteDispList(int dlid)
{
	GDL_DList *dl = opengl_displists.first, *next;
	while (dl)
	{
		if (dl->id==dlid)
		{
			BLI_remlink(&opengl_displists, dl);
			MEM_freeN(dl);
			return;
		}
		dl=dl->next;
	}
	printf("Unkown opengl displist!!!\n");
}


void GDL_CallList(int list)
{
	glCallList(list);
}
