/**
 * parameters.c
 */

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <fstream>
#include <string>
#include <sstream>

#include "error.h"
#include "parameters.h"



/**
 * Reads a parameter from a file.
 *
 * This method reads a parameter from a file, returning a newly created
 * parameter sub-class of the appropriate type.
 *
 * \param infile        Input file.
 *
 * \return      Newly created parameter sub-class or NULL if no parameter
 *               could be created (on either EOF or error).
 */
DY_Param* DY_Param::readParameter(FILE* infile)
{
        // check for EOF
        if (feof(infile) != 0)
                return NULL;

        // we expect a string of the following format:
        //  "group"(type)[size]"name":{v1,v2,v3,v4,...}

        // first, read in the group name
        string group;
        if (!readDelimitedString(infile, group))
        {
                DY_writeError("Could not read group name.");
                return NULL;
        }

        // now the type
        if (fgetc(infile) != '(')
        {
                DY_writeError("Missing '(' for type.");
                return NULL;
        }
        int t = fgetc(infile);
        DY_ParamType type;
        if (t == 'F')
                type = PARAMFLOAT;
        else if (t == 'T')
                type = PARAMTRIPLE;
        else if (t == 'M')
                type = PARAMMATRIX;
        else if (t == 'S')
                type = PARAMSTRING;
        else if (t == 'I')
                type = PARAMINT;
        else
        {
                stringstream msg;
                msg << "Unknown type '" << t << "'.";
                DY_writeError(msg.str());
                return NULL;
        }
        if (fgetc(infile) != ')')
        {
                DY_writeError("Missing ')' for type.");
                return NULL;
        }

        // now the size
        stringstream sizess;
        if (fgetc(infile) != '[') 
        {
                DY_writeError("Missing '[' for array size.");
                return NULL;
        }
        int c;
        while ((c = fgetc(infile)) != ']')
        {
                if (c == EOF)
                {
                        DY_writeError(
                                "EOF reached before ']' for array size.");
                        return NULL;
                }
                sizess << (char)c;
        }
        int sizei;
        if (sscanf(&(sizess.str()[0]), "%d", &sizei) != 1)
        {
                stringstream msg;
                msg << "Could not parse \"" << &sizess.str()[0] << "\" "
                    << "as an integer.";
                DY_writeError(msg.str());
                return NULL;
        }
        unsigned int size = static_cast<unsigned int>(sizei);

        // the name
        string name;
        if (!readDelimitedString(infile, name)) 
        {
                DY_writeError("Could not read parameter name.");
                return NULL;
        }

        // the delimiting colon
        if (fgetc(infile) != ':')
        {
                DY_writeError("Missing ':' in parameter definition.");
                return NULL;
        }

        // now call on the appropriate sub-class to construct itself
        switch (type)
        {
                case PARAMINT:
                        return DY_IntParam::readFile(group, name, size,
                                infile);
                case PARAMTRIPLE:
                        return DY_TripleParam::readFile(group, name, size, 
                                infile);
                case PARAMMATRIX:
                        return DY_MatrixParam::readFile(group, name, size, 
                                infile);
                case PARAMSTRING:
                        return DY_StringParam::readFile(group, name, size, 
                                infile);
                case PARAMFLOAT:
                default:
                        return DY_FloatParam::readFile(group, name, size, 
                                infile);
        }
}



/**
 * Reads a quote-delimited string from a file.
 *
 * \param infile        Input file.
 * \param s             String to store the result.
 *
 * \returns     true if successful, false otherwise
 */
bool DY_Param::readDelimitedString(FILE* infile, string &s)
{
        int c;
        stringstream ss;

        // first character should be a quote
        if ((c = fgetc(infile)) != '"')
        {
                DY_writeError("Missing '\"' at start of string.");
                return false;
        }

        // read until we hit the final quote
        while ((c = fgetc(infile)) != '"')
        {
                if (c == EOF)
                {
                        DY_writeError("EOF reached before '\"' at end of " \
                                "string.");
                        return false;
                }
                        
                ss << (char)c;
        }

        // save the string
        s = ss.str();

        return true;
}



/**
 * Reads in a floating-point value which must be terminated by the character
 * delim.
 *
 * \param infile        Input file.
 * \param f             Pointer to hold the floating-point value.
 * \param delim         Delimiter character, which must follow the floating-
 *                       point value.
 *
 * \returns     true if the method is successfull, false otherwise.
 */
bool DY_Param::readFloat(FILE* infile, float *f, char delim)
{
        stringstream fs;
        int c;
        while ((c = fgetc(infile)) != delim)
        {
                if (c == EOF)
                        return false;

                fs << (char)c;
        }
        if (sscanf(&fs.str()[0], "%f", f) != 1)
                return false;

        return true;
}



/**
 * Attempts to read an end-of-line character from infile.
 *
 * \param infile        Input file.
 *
 * \returns true if an end-of-line character could be read, and false
 *          otherwise.
 */
bool DY_Param::readEOL(FILE* infile)
{
        // NB: ALWAYS, ALWAYS, ALWAYS use the UNIX convention for newlines.
        //     Take good note if this is causing problems under Windoze.
        return (fgetc(infile) == '\n');
}



/**
 * Writes "header" information for this parameter to an output stream.
 *
 * The header information consists of:
 *      "group"(X)[S]"name":
 *
 * \param s     Stream.
 */
void DY_Param::writeHeader(ostream &s) const
{
        static char types[] = { 'I', 'F', 'T', 'M', 'S' };
        char typechar = types[getType()];

        s << '"' << getGroupName() << '"'
          << '(' << typechar << ')'
          << '[' << getSize() << ']'
          << '"' << getName() << '"'
          << ':';
}



/**
 * Writes this IntParam to an output stream.
 *
 * \param s     Stream to which the IntParam should be written.
 */
void DY_IntParam::write(ostream &s) const
{
        writeHeader(s);
        s << '{';
        unsigned int size = getSize();
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                s << getInt(i) << delim;
        }
        s << endl;
}



/**
 * Reads in a DY_IntParam from a file.
 *
 * \param group         Group name.
 * \param name          Parameter name.
 * \param size          Parameter array size.
 * \param infile        Input file.
 *
 * \returns Newly created DY_IntParam, or NULL if an error occurs.
 */
DY_IntParam* DY_IntParam::readFile(const string& group, const string& name,
        unsigned int size, FILE* infile)
{
        // read the first delimiting brace
        if (fgetc(infile) != '{') return NULL;

        // create the parameter
        DY_IntParam *p = new DY_IntParam(group, name, size);
        if (!p)
                return NULL;

        // read in the array of comma-delimited integers
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                if (!readInt(infile, &p->m_data[i], delim))
                {
                        delete p;
                        return NULL;
                }
        }

        // read newline
        if (!readEOL(infile))
        {
                delete p;
                return NULL;
        }

        // all done - return the newly-created parameter
        return p;
}



/**
 * Reads an integer value which must be terminated by the character delim.
 *
 * \param infile        Input file.
 * \param i             Pointer to hold the integer value.
 * \param delim         Delimiter character, which must follow the integer
 *                       value.
 *
 * \returns     true if the method is successful, false otherwise
 */
bool DY_IntParam::readInt(FILE* infile, int *i, char delim)
{
        stringstream is;
        int c;
        while ((c = fgetc(infile)) != delim)
        {
                if (c == EOF)
                        return false;

                is << (char)c;
        }
        if (sscanf(&is.str()[0], "%d", i) != 1)
                return false;

        return true;
}



/**
 * Writes this FloatParam to an output stream.
 *
 * \param s     Stream.
 */
void DY_FloatParam::write(ostream &s) const
{
        writeHeader(s);
        s << '{';
        unsigned int size = getSize();
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                s << getFloat(i) << delim;
        }
        s << endl;
}



/**
 * Reads in a FloatParam from a file.
 *
 * \param group         Group name.
 * \param name          Parameter name.
 * \param size          Parameter array size.
 * \param infile        Input file.
 *
 * \returns Newly created FloatParam, or NULL if an error occurs.
 */
DY_FloatParam* DY_FloatParam::readFile(const string& group, 
        const string& name, unsigned int size, FILE* infile)
{
        // read the first delimiting brace
        if (fgetc(infile) != '{') return NULL;

        // create the parameter
        DY_FloatParam *p = new DY_FloatParam(group, name, size);
        if (!p)
                return NULL;

        // read in the array of comma-delimited floats
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                if (!readFloat(infile, &p->m_data[i], delim))
                {
                        delete p;
                        return NULL;
                }
        }
        
        // read newline
        if (!readEOL(infile))
        {
                delete p;
                return NULL;
        }
        
        // all done - return the newly-created parameter
        return p;
}



/**
 * Writes this TripleParam to an output stream.
 *
 * \param s     Stream.
 */
void DY_TripleParam::write(ostream &s) const
{
        writeHeader(s);
        s << '{';
        float triple[3];
        unsigned int size = getSize();
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                getTriple(i, triple);
                s << '(' << triple[0] << ','
                         << triple[1] << ','
                         << triple[2] << ')' << delim;
        }
        s << endl;
}



/**
 * Reads a TripleParam from a file.
 *
 * \param group         Group name.
 * \param name          Parameter name.
 * \param size          Parameter array size.
 * \param infile        Input file.
 *
 * \returns Pointer to a new TripleParam if successful, and NULL otherwise.
 */
DY_TripleParam* DY_TripleParam::readFile(const string& group, 
        const string& name, unsigned int size, FILE* infile)
{
        // read in the first delimiting brace
        if (fgetc(infile) != '{') return NULL;

        // create the parameter
        DY_TripleParam *p = new DY_TripleParam(group, name, size);
        if (!p)
                return NULL;

        // read in the array of comma-delimited triples
        float triple[3];
        unsigned int dataidx = 0;
        for (unsigned int i=0; i < size; i++)
        {
                if (!readTriple(infile, triple))
                {
                        delete p;
                        return NULL;
                }
                p->m_data[dataidx++] = triple[0];
                p->m_data[dataidx++] = triple[1];
                p->m_data[dataidx++] = triple[2];

                char delim = (i == (size-1)) ? '}' : ',';
                if (fgetc(infile) != delim)
                {
                        delete p;
                        return NULL;
                }
        }

        // read end-of-line
        if (!readEOL(infile))
        {
                delete p;
                return NULL;
        }

        return p;
}



/**
 * Reads a triple from a file.
 *
 * A triple might look something like this: (1.2,2.3,3.4)
 *
 * \param infile        Input file.
 * \param triple        3-element array to hold the triple value.
 *
 * \returns true if successful, false otherwise.
 */
bool DY_TripleParam::readTriple(FILE* infile, float *triple)
{
        // read in the first delimiting bracket
        if (fgetc(infile) != '(') return false;

        // read in three floats
        float x, y, z;
        if (!readFloat(infile, &x, ',')) return false;
        if (!readFloat(infile, &y, ',')) return false;
        if (!readFloat(infile, &z, ')')) return false;
        
        // set the three floats
        triple[0] = x;
        triple[1] = y;
        triple[2] = z;

        // return success
        return true;
}



/**
 * Writes this MatrixParam to an output stream.
 *
 * \param s     Stream.
 */
void DY_MatrixParam::write(ostream &s) const
{
        writeHeader(s);
        s << '{';
        unsigned int size = getSize();
        float matrix[16];
        for (unsigned int i=0; i < size; i++)
        {
                getMatrix(i, matrix);
                char outerdelim = (i == (size-1)) ? '}' : ',';
                s << '(';
                for (int j=0; j < 16; j++)
                {
                        char innerdelim = (j == 15) ? ')' : ',';
                        s << matrix[j] << innerdelim;
                }
                s << outerdelim;
        }
        s << endl;
}



/**
 * Reads a single matrix from a file.
 *
 * A matrix might look something like this:
 *  (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
 *
 * \param infile        Input file.
 * \param matrix        16-element array to hold the matrix.
 *
 * \returns true if successful, false otherwise.
 */
bool DY_MatrixParam::readMatrix(FILE* infile, float *matrix)
{
        // read in the first delimiting bracket
        if (fgetc(infile) != '(') return false;

        // read in 16 floats
        float f[16];
        for (int i=0; i < 16; i++)
        {
                char delim = (i == 15) ? ')' : ',';
                if (!readFloat(infile, &f[i], delim))
                        return false;
        }

        // set the floats
        memcpy(matrix, f, 16*sizeof(float));

        return true;
}



/**
 * Reads a MatrixParam from file.
 *
 * \param group         Group name.
 * \param name          Parameter name.
 * \param size          Parameter array size.
 * \param infile        Input file.
 */
DY_MatrixParam* DY_MatrixParam::readFile(const string& group, 
        const string &name, unsigned int size, FILE* infile)
{
        // read in the first delimiting brace
        if (fgetc(infile) != '{') return NULL;

        // create the parameter
        DY_MatrixParam *p = new DY_MatrixParam(group, name, size);
        if (!p)
                return NULL;

        // read in the array of comma-delimited matrices
        for (unsigned int i=0; i < size; i++)
        {
                if (!readMatrix(infile, &p->m_data[16*i]))
                {
                        delete p;
                        return NULL;
                }

                char delim = (i == (size-1)) ? '}' : ',';
                if (fgetc(infile) != delim)
                {
                        delete p;
                        return NULL;
                }
        }

        // read end-of-line
        if (!readEOL(infile))
        {
                delete p;
                return NULL;
        }

        return p;
}



/**
 * Writes this StringParam to an output stream.
 *
 * \param s     Stream.
 */
void DY_StringParam::write(ostream &s) const
{
        writeHeader(s);
        s << '{';
        unsigned int size = getSize();
        for (unsigned int i=0; i < size; i++)
        {
                char delim = (i == (size-1)) ? '}' : ',';
                s << '"' << getString(i) << '"' << delim;
        }
        s << endl;
}



/**
 * Reads a string parameter from a file.
 *
 * \param group         Group name.
 * \param name          Parameter name.
 * \param size          Parameter array size.
 * \param infile        Input file name.
 *
 * \returns Pointer to a newly-created StringParam, or NULL if an error
 *          occurs.
 */
DY_StringParam* DY_StringParam::readFile(const string& group,
        const string& name, unsigned int size, FILE* infile)
{
        // read in the first delimiting brace
        if (fgetc(infile) != '{') return NULL;

        // create the parameter
        DY_StringParam *p = new DY_StringParam(group, name, size);
        if (!p)
                return NULL;

        // read in the array of comma-delimited strings
        for (unsigned int i=0; i < size; i++)
        {
                p->m_data.push_back(string());
                if (!readDelimitedString(infile, p->m_data[i]))
                {
                        delete p;
                        return NULL;
                }

                char delim = (i == (size-1)) ? '}' : ',';
                if (fgetc(infile) != delim)
                {
                        delete p;
                        return NULL;
                }
        }

        // read end-of-line
        if (!readEOL(infile))
        {
                delete p;
                return NULL;
        }

        return p;
}



/**
 * Returns all parameters with the specified group name.
 *
 * The group members are added to p.  If p already has parameters inside it 
 * then they are not removed.
 *
 * \param groupName     Name of the group of parameters to fetch.
 * \param p             Parameter set to add the group members to.
 */
void DY_ParamManager::getGroup(const string& groupName, DY_ParamSet &p) const
{
        unsigned int size = m_params.size();
        for (unsigned int i=0; i < size; i++)
        {
                if (m_params[i]->belongsToGroup(groupName))
                {
                        p.addParam(m_params[i]);
                }
        }
}



/**
 * Parses a parameter file.
 *
 * If the specified parameter file has not yet been parsed, then it is parsed.
 * If parsing has already taken place then the method returns and does
 * nothing.
 *
 * \param fileName      Name of file to parse.
 *
 * \returns true if parsing was successful, false otherwise.
 */
bool DY_ParamManager::addFile(const string& filename)
{
	unsigned int i;
        // check if the file has been parsed
        for (i=0; i < m_filenames.size(); i++)
        {
                if (m_filenames[i].compare(filename) == 0)
                        return true;
        }

        // open the file for reading
        FILE* infile = fopen(&filename[0], "r");
        if (!infile)
        {
                stringstream msg;
                msg << "Could not open file: \"" << filename << "\" "
                    << "for reading.";
                DY_writeError(msg.str());
                return false;
        }

        // read until EOF
        vector<DY_Param*> tp;
        int c;
        while (feof(infile) == 0)
        {
                c = fgetc(infile);
                if (feof(infile) == 0)
                        ungetc(c, infile);
                else
                        break;

                DY_Param* p = DY_Param::readParameter(infile);
                if (p)
                {
                        tp.push_back(p);
                }
                else
                {
                        fclose(infile);
                        return false;
                }
        }

        // close the file
        fclose(infile);

        // add the new parameters to the list
        for (i=0; i < tp.size(); i++)
        {
                m_params.push_back(tp[i]);
        }

        // add the filename to the list
        m_filenames.push_back(filename);

        // return success
        return true;
}



/**
 * Writes the current parameters to file.
 *
 * \param filename      Name of file to use.
 */
void DY_ParamManager::write(const string& filename)
{
        ofstream f(&filename[0]);
        
        // iterate over each parameter that the manager knows about and
        //  write them to file
        unsigned int size = m_params.size();
        for (unsigned int i=0; i < size; i++)
        {
                m_params[i]->write(f);
        }

        f.close();
}



/**
 * Adds a single integer parameter.
 *
 * \param groupName     Group name for the parameter.
 * \param name          Name for the parameter.
 * \param i             Integer value.
 */
void DY_ParamManager::addInt(const string& groupName, const string &name, int i)
{
        int vals[1] = { i };
        addIntV(groupName, name, 1, vals);
}



/**
 * Adds an array of integers parameter.
 *
 * \param groupName     Group name for the parameter.
 * \param name          Name of the parameter.
 * \param size          Number of integers.
 * \param vals          Array of integers.
 */
void DY_ParamManager::addIntV(const string& groupName, const string &name,
        int size, int vals[])
{
        DY_IntParam *p = new DY_IntParam(groupName, name, size);
        memcpy(p->m_data, vals, size*sizeof(int));
        m_params.push_back(p);
}



/**
 * Adds a floating point parameter.
 */
void DY_ParamManager::addFloat(const string& groupName, const string &name,
        float f)
{
        float vals[1] = { f };
        addFloatV(groupName, name, 1, vals);
}



/**
 * Adds an array of floating points parameter.
 */
void DY_ParamManager::addFloatV(const string& groupName, const string &name,
        int size, float vals[])
{
        DY_FloatParam *p = new DY_FloatParam(groupName, name, size);
        memcpy(p->m_data, vals, size*sizeof(float));
        m_params.push_back(p);
}


/**
 * Adds a triple parameter.
 */
void DY_ParamManager::addTriple(const string& groupName, const string &name,
        float t[3])
{
        addTripleV(groupName, name, 1, t);
}



/**
 * Adds an array of triples parameter.
 *
 * Note that the size of the array vals should be: 3 * size
 */
void DY_ParamManager::addTripleV(const string& groupName, const string &name,
        int size, float vals[])
{
        DY_TripleParam *p = new DY_TripleParam(groupName, name, size);
        memcpy(p->m_data, vals, 3*size*sizeof(float));
        m_params.push_back(p);
}



/**
 * Adds a matrix parameter.
 */
void DY_ParamManager::addMatrix(const string& groupName, const string& name,
        float m[16])
{
        DY_MatrixParam *p = new DY_MatrixParam(groupName, name, 1);
        memcpy(p->m_data, m, 16*sizeof(float));
        m_params.push_back(p);
}



/**
 * Adds a string parameter.
 */
void DY_ParamManager::addString(const string& groupName, const string& name,
        const string& s)
{
        DY_StringParam *p = new DY_StringParam(groupName, name, 1);
        p->m_data.push_back(s);
        m_params.push_back(p);
}



/**
 * Adds an array of strings parameter.
 */
void DY_ParamManager::addStringV(const string& groupName, const string& name,
        int size, string vals[])
{
        DY_StringParam *p = new DY_StringParam(groupName, name, size);
        for (int i=0; i < size; i++)
        {
                p->m_data.push_back(vals[i]);
        }
        m_params.push_back(p);
}

