/**
 * parameters.h
 */
#ifndef PARAMETERS_H
#define PARAMETERS_H

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <vector>
#include <iostream>

using namespace std;

typedef enum {
        PARAMINT,
        PARAMFLOAT,
        PARAMTRIPLE,
        PARAMMATRIX,
        PARAMSTRING
} DY_ParamType;


/**
 * \class DY_Param
 *
 * Base class for parameters used by Dobby.  Dobby parameter list files store
 * one parameter per line.  Dobby parameters are always arrays of values, but
 * can have lengths of 1 to represent scalars.  Each line has the following
 * format:
 *   "group"(X)[S]"name":{v1,v2,v3,...,vn}
 * where:
 *   group - name of the group to which this parameter belongs
 *   X     - parameter type, one of 'I', 'F', 'T', 'M', 'S'
 *   S     - integer size of parameter array
 *   name  - name of parameter
 *   vi    - array of parameter (format is type-dependent)
 *
 * Note that the exact format of Dobby parameter files is very important.
 * In particular, whitespace should not be included except as UNIX newlines
 * at the end of each parameter definition.
 *
 * The different types permitted reflect those available in the RenderMan
 * shading language, with the addition of integer types.
 *   'I' - PARAMINT    - integer
 *   'F' - PARAMFLOAT  - floating point
 *   'T' - PARAMTRIPLE - triple (point, vector, normal, color)
 *   'M' - PARAMMATRIX - 4x4 matrices (16 elements)
 *   'S' - PARAMSTRING - string
 */
class DY_Param
{
        //------------------------------------------------------ Public Methods
        public:
                static DY_Param* readParameter(FILE* infile);
                virtual ~DY_Param() {}
                virtual DY_ParamType getType() const
                { 
                        assert(0);  // should never be called
                        return PARAMFLOAT;
                }
                const string &getGroupName() const
                {
                        return m_group;
                }
                virtual unsigned int getSize() const
                {
                        assert(0);  // should never be called
                        return 0;
                }
                const string &getName() const
                {
                        return m_name;
                }
                bool belongsToGroup(const string& groupName) const
                {
                        return (m_group.compare(groupName) == 0);
                }
                virtual void write(ostream &s) const
                {
                        s << "ERROR";
                        assert(0);
                }
        //--------------------------------------------------- Protected Methods
        protected:
                /**
                 * Param constructor.
                 *
                 * Do not use this constructor to create a parameter.  The
                 * expected usage scenario is to call the static 
                 * readParameter(), which will create a Param subclass of the
                 * correct type.
                 */
                DY_Param(string groupName, string name) :
                        m_group(groupName),
                        m_name(name)
                {}
                static bool readDelimitedString(FILE* infile, string &s);
                static bool readFloat(FILE* infile, float *f, char delim);
                static bool readEOL(FILE* infile);
                void writeHeader(ostream &s) const;
        //--------------------------------------------------- Protected Members
        protected:
                string    m_group;      ///< name of the parameter's group
                string    m_name;       ///< name of the parameter
};


/**
 * \class DY_IntParam
 *
 * Integer Dobby parameter.
 *
 * An example of an integer parameter stored in a file is the following:
 *    "group"(I)[3]"myint":{3,2,1}
 */
class DY_IntParam : public DY_Param
{
        friend class DY_Param;
        friend class DY_ParamManager;
        //------------------------------------------------------ Public Methods
        public:
                virtual ~DY_IntParam()
                {
                        delete [] m_data;
                }
                int getInt(unsigned int index) const
                {
                        assert(index < m_size);
                        assert(m_data);
                        return m_data[index];
                }
                virtual DY_ParamType getType() const
                {
                        return PARAMINT;
                }
                virtual unsigned int getSize() const
                {
                        return m_size;
                }
                virtual void write(ostream &s) const;
        //--------------------------------------------------- Protected Methods
        protected:
                DY_IntParam(const string& group, const string &name,
                        unsigned int size) :
                        DY_Param(group, name),
                        m_size(size)
                {
                        m_data = new int[m_size];
                }
                static DY_IntParam* readFile(const string& group,
                        const string& name, unsigned int size, FILE* infile);
                static bool readInt(FILE* infile, int *i, char delim);
        //--------------------------------------------------- Protected Members
        protected:
                unsigned int m_size;    ///< array size
                int*         m_data;    ///< array
                
};


/**
 * \class DY_FloatParam
 *
 * Floating point Dobby parameter.
 *
 * An example of a floating point parameter stored in a file is the following:
 *   "group"(F)[2]"myfloat":{0.033,23.44}
 */
class DY_FloatParam : public DY_Param
{
        friend class DY_Param;
        friend class DY_ParamManager;
        //------------------------------------------------------ Public Methods
        public:
                virtual ~DY_FloatParam()
                {
                        delete [] m_data;
                }
                float getFloat(unsigned int index) const
                {
                        assert(index < m_size);
                        assert(m_data);
                        return m_data[index];
                }
                virtual DY_ParamType getType() const
                {
                        return PARAMFLOAT;
                }
                virtual unsigned int getSize() const
                {
                        return m_size;
                }
                virtual void write(ostream &s) const;
        //--------------------------------------------------- Protected Methods
        protected:
                DY_FloatParam(const string& group, const string &name,
                        unsigned int size) :
                        DY_Param(group, name),
                        m_size(size)
                {
                        m_data = new float[m_size];
                }
                static DY_FloatParam* readFile(const string& group, 
                        const string& name, unsigned int size, FILE* infile);
        //--------------------------------------------------- Protected Members
        protected:
                unsigned int m_size;    ///< array size
                float*       m_data;    ///< array
};


/**
 * \class DY_TripleParam
 *
 * Triple Dobby parameter.
 *
 * An example of a triple stored in a file is the following:
 *    "group"(T)[1]"mytriple":{(0,0,0),(1,3.2,2),(5.44,3.2,1)}
 */
class DY_TripleParam : public DY_Param
{
        friend class DY_Param;
        friend class DY_ParamManager;
        //------------------------------------------------------ Public Methods
        public:
                virtual ~DY_TripleParam()
                {
                        delete [] m_data;
                }
                void getTriple(unsigned int index, float *triple) const
                {
                        assert(index < m_size);
                        assert(m_data);
                        index *= 3;
                        triple[0] = m_data[index++];
                        triple[1] = m_data[index++];
                        triple[2] = m_data[index];
                }
                virtual DY_ParamType getType() const
                {
                        return PARAMTRIPLE;
                }
                virtual unsigned int getSize() const
                {
                        return m_size;
                }
                virtual void write(ostream &s) const;
        //--------------------------------------------------- Protected Methods
        protected:
                DY_TripleParam(const string& group, const string& name,
                        unsigned int size) :
                        DY_Param(group, name),
                        m_size(size)
                {
                        m_data = new float[m_size * 3];
                }
                static DY_TripleParam* readFile(const string& group,
                        const string& name, unsigned int size, FILE* infile);
                static bool readTriple(FILE* infile, float *triple);
        //--------------------------------------------------- Protected Members
        protected:
                unsigned int m_size;    ///< array size
                float*       m_data;    ///< array
};


/**
 * \class DY_MatrixParam
 *
 * A Dobby matrix parameter.
 *
 * An example of a matrix parameter stored in a file is the following:
 *    "group"(M)[1]"mymatrix":{(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)}
 */
class DY_MatrixParam : public DY_Param
{
        friend class DY_Param;
        friend class DY_ParamManager;
        //------------------------------------------------------ Public Methods
        public:
                virtual ~DY_MatrixParam()
                {
                        delete [] m_data;
                }
                void getMatrix(unsigned int index, float *matrix) const
                {
                        assert(m_data);
                        assert(index < m_size);
                        index *= 16;
                        memcpy(matrix, &m_data[index], 16*sizeof(float));
                }
                virtual DY_ParamType getType() const
                {
                        return PARAMMATRIX;
                }
                virtual unsigned int getSize() const
                {
                        return m_size;
                }
                virtual void write(ostream &s) const;
        //--------------------------------------------------- Protected Methods
        protected:
                DY_MatrixParam(const string& group, const string& name,
                        unsigned int size) :
                        DY_Param(group, name),
                        m_size(size)
                {
                        m_data = new float[m_size * 16];
                }
                static DY_MatrixParam* readFile(const string& group,
                        const string &name, unsigned int size, FILE* infile);
                static bool readMatrix(FILE* infile, float *matrix);
        //--------------------------------------------------- Protected Members
        protected:
                unsigned int m_size;    ///< array size
                float*       m_data;    ///< array
};


/**
 * \class DY_StringParam
 *
 * A Dobby string parameter.
 *
 * An example of a string parameter stored in a file is the following:
 *    "group"(S)[2]"mystring":{"Hello","World !"}
 */
class DY_StringParam : public DY_Param
{
        friend class DY_Param;
        friend class DY_ParamManager;
        //------------------------------------------------------ Public Methods
        public:
                virtual ~DY_StringParam() {}
                const string& getString(unsigned int index) const
                {
                        assert(index < m_data.size());
                        return m_data[index];
                }
                virtual DY_ParamType getType() const
                {
                        return PARAMSTRING;
                }
                virtual unsigned int getSize() const
                {
                        return m_data.size();
                }
                virtual void write(ostream &s) const;
        //--------------------------------------------------- Protected Methods
        protected:
                DY_StringParam(const string& group, const string& name,
                        unsigned int size) :
                        DY_Param(group, name)
                {
                        m_data.reserve(size);
                }
                static DY_StringParam* readFile(const string& group,
                        const string &name, unsigned int size, FILE* infile);
        //--------------------------------------------------- Protected Members
        protected:
                vector<string> m_data;
};


/**
 * \class DY_ParamSet
 *
 * Represents a set of parameters.  The set may be related in some way (eg:
 * possibly by all coming from the same group), but this is not an enforced
 * requirement.
 */
class DY_ParamSet
{
        //------------------------------------------------------ Public Methods
        public:
                DY_ParamSet() {}
                ~DY_ParamSet() {}
                /**
                 * Adds another parameter to the set.
                 *
                 * \param p     Pointer to the parameter to add.
                 */
                void addParam(DY_Param *p)
                {
                        assert(p);
                        // add the parameter
                        m_params.push_back(p);
                }
                /**
                 * Returns a parameter of the specified name and type, or NULL
                 * if no such parameter can be found.
                 *
                 * \param name  Name of the parameter.
                 * \param type  Type of the parameter.
                 *
                 * \returns Requested parameter, or NULL if it cannot be 
                 *          found.
                 */
                DY_Param* getParam(const string& name, DY_ParamType type)
                {
                        for (unsigned int i=0; i < m_params.size(); i++)
                        {
                                if ((m_params[i]->getType() == type) &&
                                    (m_params[i]->getName().compare(name)==0))
                                        return m_params[i];
                        }
                        return NULL;
                }
                /**
                 * Returns an integer parameter.
                 *
                 * \param name  Parameter name.
                 */
                DY_IntParam* getIntParam(const string& name)
                {
                        return static_cast<DY_IntParam*>(getParam(name, PARAMINT));
                }
                /**
                 * Returns a float parameter.
                 *
                 * \param name  Parameter name.
                 */
                DY_FloatParam* getFloatParam(const string& name)
                {
                        return static_cast<DY_FloatParam*>(
                                getParam(name, PARAMFLOAT));
                }
                /**
                 * Returns a triple parameter.
                 *
                 * \param name  Parameter name.
                 */
                DY_TripleParam* getTripleParam(const string& name)
                {
                        return static_cast<DY_TripleParam*>(
                                getParam(name, PARAMTRIPLE));
                }
        //--------------------------------------------------- Protected Members
        protected:
                vector<DY_Param*> m_params;        ///< stored parameters
};


/**
 * \class DY_ParamManager
 * 
 * Manages all parameters used by a particular context of the Dobby engine.
 */
class DY_ParamManager
{
        //------------------------------------------------------ Public Methods
        public:
                DY_ParamManager() {}
                ~DY_ParamManager() 
                {
                        for (unsigned int i=0; i < m_params.size(); i++)
                                delete m_params[i];
                }
                void getGroup(const string& groupName, DY_ParamSet &p) const;
                bool addFile(const string& fileName);
                void write(const string& fileName);
                void addInt(const string& groupName, const string& name, int i);
                void addIntV(const string& groupName, const string& name,
                        int size, int vals[]);
                void addFloat(const string& groupName, const string& name,
                        float f);
                void addFloatV(const string& groupName, const string& name,
                        int size, float vals[]);
                void addTriple(const string& groupName, const string& name,
                        float t[3]);
                void addTripleV(const string& groupName, const string& name,
                        int size, float vals[]);
                void addMatrix(const string& groupName, const string& name,
                        float m[16]);
                void addString(const string& groupName, const string& name,
                        const string &s);
                void addStringV(const string& groupName, const string& name,
                        int size, string vals[]);
        //--------------------------------------------------- Protected Members
        protected:
                vector<DY_Param*> m_params;        ///< stored parameters
                vector<string>    m_filenames;     ///< files already parsed
};


#endif // PARAMETERS_H

